/**
	$Id: JobQueueChecker.java 5384 2010-08-13 08:50:27Z nicklas $

	Copyright (C) 2006 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.jobagent;

import java.util.List;
import java.util.TimerTask;

import net.sf.basedb.core.Application;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.Job;
import net.sf.basedb.core.JobAgent;
import net.sf.basedb.core.JobAgentSettings;
import net.sf.basedb.core.SessionControl;

/**
	This class is given the responsibility to check the job queue for 
 	jobs that are awaiting execution. Each agent has one instance of this
 	class which is registered with the BASE core scheduler {@link Application#getScheduler()}.
 	<p>
 	This object should be thread-safe since the scheduler creates a new thread each 
 	time the {@link #run()} method is called. 

	@author nicklas
	@version 2.0
	@base.modified $Date: 2010-08-13 10:50:27 +0200 (Fri, 13 Aug 2010) $
*/
public class JobQueueChecker
	extends TimerTask
{

	/**
		Log job agent events.
	*/
	private static final org.apache.log4j.Logger log = 
		org.apache.log4j.LogManager.getLogger("net.sf.basedb.clients.jobagent.JobQueueChecker");
	
	private final Agent agent;
	
	/**
		Cached query for waiting jobs. Reloaded only when the 
		job agent's version has changed. We cache the query since loading
		all users that are allowed to use the job agent is an expensive operation.
	*/ 
	private ItemQuery<Job> jobQuery = null;
	/**
		The job agent's version when the query was cached.
	*/
	private int agentVersionForJobQuery = -1;
	
	public JobQueueChecker(Agent agent)
	{
		this.agent = agent;
		
		// Verify that jobagent is registered
		DbControl dc = null;
		try
		{
			SessionControl sc = agent.getSessionControl();
			dc = sc.newDbControl();
			agent.getJobAgent(dc);
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
	/*
		From the TimerTask class
		-------------------------------------------
	*/
	public void run()
	{
		log.info("Checking for jobs to execute");
		checkForJobs();
	}
	public boolean cancel()
	{
		log.info("Cancelling job queue checker");
		return super.cancel();
	}
	// -------------------------------------------
	
	private void checkForJobs()
	{
		DbControl dc = null;
		try
		{
			SessionControl sc = agent.getSessionControl();
			dc = sc.newDbControl();
			JobAgent jobAgent = agent.getJobAgent(dc);
			
			// Do we need to create a new query for jobs or not?
			// We create a new query only if none exists or if the cached
			// version is different from the current version since that may
			// be because permissions, etc have changed.
			if (jobQuery == null || jobAgent.getVersion() != agentVersionForJobQuery)
			{
				jobQuery = jobAgent.getWaitingJobs();
				jobQuery.setMaxResults(1);
				agentVersionForJobQuery = jobAgent.getVersion();
			}
			
			List<Job> jobs = jobQuery.list(dc);
			if (jobs.size() == 0)
			{
				log.info("No jobs waiting");
			}
			else
			{
				Job job = jobs.get(0);
				JobAgentSettings settings = jobAgent.getSettings(job);
				agent.startJob(job, settings);
			}
		}
		catch (Throwable t)
		{
			log.error(t.getMessage(), t);
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
				
}
