/*
	$Id: ReporterTransfer.java 4889 2009-04-06 12:52:39Z nicklas $

	Copyright (C) 2005 Jari Häkkinen, Nicklas Nordborg, Gregory Vincic
	Copyright (C) 2006 Johan Enell, Jari Häkkinen, Nicklas Nordborg, Gregory Vincic
	Copyright (C) 2007 Jari Häkkinen, Nicklas Nordborg, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	BASE is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/

package net.sf.basedb.clients.migrate;

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.DatabaseException;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.DynamicReporterQuery;
import net.sf.basedb.core.DynamicResultIterator;
import net.sf.basedb.core.Reporter;
import net.sf.basedb.core.ReporterBatcher;
import net.sf.basedb.core.data.ReporterData;
import net.sf.basedb.core.query.Dynamic;
import net.sf.basedb.core.query.Orders;
import net.sf.basedb.core.query.SqlResult;
import net.sf.basedb.util.timer.PerformanceTimer;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.HashMap;
import java.util.Map;

/**
This class provides methods for transferring reporter data from base 1 to
BASE 2.

@author Gregory, Martin, Nicklas
@version 2.0
@base.modified $Date: 2009-04-06 14:52:39 +0200 (Mon, 06 Apr 2009) $
*/

public class ReporterTransfer 
	extends Transfer
{
	
	/**
	Logger for this class
	 */
	private static final org.apache.log4j.Logger log = 
		org.apache.log4j.Logger.getLogger("net.sf.basedb.clients.migrate.ReporterTransfer");
	
	/**
		Maps Base 1 ID to ReporterData object.
	*/
	private Map<Integer, ReporterData> reporterMap = null;
	
	/**
		Maps Reporter external ID to Base 1 ID. Only used temporary until the
		reporterMap is generated.
	*/
	private Map<String, Integer> externalIdMap = null;
	
	/**
		Maps BASE 1 reporters that are considered duplicate by BASE 2 after
		whitespace has been trimmed from the external ID. Only used temporary until the
		reporterMap is generated. Map contain BASE 1 ID --> Base 1 ID where the value
		is the "real" reporter that is migrated to BASE 2.
	*/
	private Map<Integer, Integer> duplicates = null;
	
	/**
		Used for inserting reporter into Base 2.
	*/
	private ReporterBatcher batcher = null;
	
	/**
		Flag that indicates if we should run in RESUME mode 
		(ie. NOT create new reporter items, only map ID:s)
	*/
	private boolean RESUME = false;
	
	/**
	Set string and text properties, order must match the select statement
	except for the id and reporterId in the beginning Extended properties are
	defined in the extended-properties.xml file for class ReporterData.
	Performance is increased by using the column index instead of name.
	 */
	private String[] stringProperties = { "species", "clusterId", "accession",
			"nid", "chromosome", "cytoband", "markers", "antibiotics",
			"locusLink", "omim", "sequence", "vector", "tissue", "library" };

	/**
	Reporters are independant from other transfers.
	 */
	ReporterTransfer(boolean resume)
	{
		RESUME = resume;
	}

	public void start()
	{
		if (RESUME) QUICK_TRANSFER = true;
		if (RESUME)
		{
			log.info("Loading reporters...");
		}
		else if (!QUICK_TRANSFER)
		{
			log.info("Transferring reporters...");
		}
		else 
		{
			log.info("Transferring reporters with minimum properties...");
		}
		
		int count = getInt("countReporters");
		// First pass creates reporters...
		startProgress(count, "Reporters"); 
		PreparedStatement ps = QUICK_TRANSFER ? prepareStatementFromFile("selectMinimumReporters")
				: prepareStatementFromFile("selectReporters");

		externalIdMap = new HashMap<String, Integer>(count);
		duplicates = new HashMap<Integer, Integer>();
		DbControl dc = null;
		try
		{
			dc = newDbControl();
			batcher = ReporterBatcher.getNew(dc);
			batcher.setBatchSize(getBatchSize());
			run(ps, 0, getBatchSize());
			batcher.close();
			dc.commit();
		}
		finally
		{
			if (dc != null) dc.close();
			progress.stop();
		}
			
		// ... second pass reads in the new ID:s
		progress = null;
		startProgress(count, "Mapping reporter ID:s"); 
		try
		{
			reporterMap = new HashMap<Integer, ReporterData>(count);
			dc = newDbControl();
			log.info("Mapping reporters in BASE 2...");
			int start = 0;
			int batchSize = getBatchSize();

			DynamicReporterQuery query = Reporter.getDynamicQuery();
			query.select(Dynamic.selectReporter("id"));
			query.select(Dynamic.selectReporter("externalId"));
			query.order(Orders.asc(Dynamic.reporter("id")));
			query.setMaxResults(batchSize);
			while (start < count)
			{
				query.setFirstResult(start);
				cleanUpMemory();
				log.info("Mapping reporter #" + start + " to #"+(start+batchSize));
				mapMigratedReporters(dc, query);
				start += batchSize;
			}
			cleanUpMemory();
			log.info("Mapping duplicate reporters");
			mapDuplicateReporters();
		}
		finally
		{
			if (dc != null) dc.close();
			externalIdMap.clear();
			duplicates.clear();
			progress.stop();
		}
		done();
		log.info("Transferring Reporters...DONE");
	}

	/**
	Recursive run over a SQL query defined by a prepared statement and LIMIT
	values from and size.
	
	@param ps <code>PreparedStatement</code> which includes a "LIMIT ?,?"
	       at the end.
	@param from <code>int</code> indicating start position of the
	       ResultSet.
	@param size <code>int</code> indicating number of rows to include.
	@see Transfer#prepareStatementFromFile(String)
	 */
	protected void run(PreparedStatement ps, int from, int size)
	{
		String base1Properties = null;
		cleanUpMemory();
		try
		{
			int rowCount = 0;
			ps.setInt(1, from);
			ps.setInt(2, size);
			ResultSet rs = ps.executeQuery();
			while (rs.next())
			{
				base1Properties = rs.getInt(1) + ":" + rs.getString(2);
				rowCount++;
				ReporterData item = createDataItem(rs);
				if (item != null && !RESUME) batcher.insert(item);
				progress.increase();
			}
			if (rowCount == size)
			{
				// Fetch next batch
				log.debug("fetching next batch");
				run(ps, from + size, size);
			}
		}
		catch (SQLException e)
		{
			log.error("FAILED when executing SQL statement", e);
			stop();
		}
		catch (BaseException e)
		{
			log.error("FAILED when transferring reporter[" + base1Properties + "]", e);
			stop();
		}
	}

	private void mapMigratedReporters(DbControl dc, DynamicReporterQuery query)
	{
//		DataResultIterator<ReporterData> result = null;
		DynamicResultIterator result = null;
		PerformanceTimer timer = new PerformanceTimer(
			"iterate()", "hasNext()", "next()", "HashMap.get()", "HashMap.put()", "other");
		try
		{
			timer.start(0);
			result = query.iterate(dc);
			timer.stopStart(0, 1);
			while (result.hasNext())
			{
				timer.stopStart(1, 2);
				SqlResult reporter = result.next();
				String externalId = reporter.getString(2);
				timer.stopStart(2, 3);
				Integer base1Id = externalIdMap.get(externalId);
				timer.stopStart(3, 4);
				reporterMap.put(base1Id, new ReporterProxy(reporter.getInt(1)));
				timer.stopStart(4, 5);
				progress.increase();
				if (log.isDebugEnabled())
				{
					log.debug("Reporter: " + externalId + 
						" has been mapped to BASE 1 ID="+base1Id);
				}
				timer.stopStart(5, 1);
			}
			timer.stopAll();
			if (log.isInfoEnabled())
			{
				log.info("\n"+timer.toString());
			}
		}
		catch (SQLException ex)
		{
			throw new DatabaseException(ex);
		}
		finally
		{
			if (result != null) result.close();
		}
	}

	private void mapDuplicateReporters()
	{
		// The map contains base1Id of duplicate -> base1Id of transfered reporter
		for (Map.Entry<Integer, Integer> entry : duplicates.entrySet())
		{
			Integer duplicateId = entry.getKey();
			Integer migratedId = entry.getValue();
			ReporterData migratedReporter = reporterMap.get(migratedId);
			reporterMap.put(duplicateId, migratedReporter);
			progress.increase();
			if (log.isDebugEnabled())
			{
				log.debug("Reporter: " + migratedReporter.getExternalId() + 
					" has been mapped to BASE 1 ID=" + duplicateId);
			}
		}
	}
	
	

	/**
	Creates a {@link ReporterData} from the current position in
	<code>ResultSet</code>.
	
	@param rs <code>ResultSet</code> containing the data for this item.
	@return <code>ReporterData</code> or null if another reporter with same ID has already been created
	@throws BaseException
	@throws SQLException If current ResultSet position is invalid or number
	        of columns in the select statement is to low.
	 */
	private ReporterData createDataItem(ResultSet rs) throws BaseException,
			SQLException
	{
		// Populate a new ReporterData object		
		String externalId = rs.getString(2).trim();
		int base1Id = rs.getInt(1);
		ReporterData rd = null;
		if (externalIdMap.containsKey(externalId))
		{
			Integer migratedBase1Id = externalIdMap.get(externalId);
			log.warn("Duplicate reporter found in BASE 1 after trimming whitespace: id=" + 
				base1Id + "; externald ID=" + externalId + "; Merging this reporter with reporter id="+ 
				migratedBase1Id);
			duplicates.put(base1Id, migratedBase1Id);
		}
		else
		{
			externalIdMap.put(externalId, base1Id);
			if (log.isDebugEnabled())
			{
				log.debug("Transferring Reporter BASE1 id(" + base1Id + ")\texternal Id(" + externalId + ")");
			}
			rd = Reporter.getNew(externalId);
			String name = rs.getString(3);
			if (name == null || "".equals(name)) name = externalId;
			rd.setName(name);
			rd.setSymbol(rs.getString(4));
			if (!QUICK_TRANSFER)
			{
				int columnIndex = 5;
				for (String property : stringProperties)
				{
					rd.setExtended(property, rs.getString(columnIndex));
					columnIndex++;
				}
				// Set integer properties, integer properties should be listed
				// last(after the string properties) in the select statement
				String[] intProperties = { "length" };
				for (String property : intProperties)
				{
					rd.setExtended(property, rs.getInt(columnIndex));
					columnIndex++;
				}
			}
		}
		return rd;
	}

	public ReporterData getReporterById(DbControl dc, int base1ReporterId)
	{
		return reporterMap.get(base1ReporterId);
	}
	
	public Integer getBase2Id(DbControl dc, int base1ReporterId)
	{
		ReporterData reporter = reporterMap.get(base1ReporterId);
		return reporter == null ? null : reporter.getId();
	}
	
}
