/**
	$Id: Settings.java 5384 2010-08-13 08:50:27Z nicklas $

	Copyright (C) Authors contributing to this file.

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.extensions;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import net.sf.basedb.core.Presets;
import net.sf.basedb.core.Presets.Preset;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.extensions.DefaultFilter;
import net.sf.basedb.util.extensions.Extension;
import net.sf.basedb.util.extensions.ExtensionPoint;
import net.sf.basedb.util.extensions.ExtensionsFilter;
import net.sf.basedb.util.extensions.Registry;

/**
	Class for keeping settings for the extensions system. Settings
	are kept in the file <code>WEB-INF/extensions/settings.xml</code>.
	<p>

	Extension points and extensions can be disabled/enabled by
	{@link #enableExtensionPoint(String, boolean)} and 
	{@link #enableExtension(String, boolean)}.
	<p>

	This class also implements the {@link ExtensionsFilter}
	interface, and can be used on the 
	{@link Registry#useExtensions(net.sf.basedb.util.extensions.ClientContext, ExtensionsFilter, String...)}
	method. This means that disabling/enabling extensions and extension points 
	will immediately be visible in the web interface. Sorting is delegated
	to the {@link DefaultFilter}.

	@author nicklas
	@version 2.7
	@base.modified $Date:2008-03-20 12:15:25 +0100 (Thu, 20 Mar 2008) $
*/
public class Settings 
	extends DefaultFilter
	implements ExtensionsFilter 
{
	
	private static final String AUTO_INSTALL = "auto-install";

	private final ExtensionsDirectory directory;
	private final File file;
	private boolean hasChanged;
	private Presets presets;
	private Preset settings;
	private Preset disabledExtensions;
	private Preset disabledExtensionPoints;
	
	/**
		Create a new Settings object. If the given file doesn't exists
		it will be created if {@link #save()} is called. The settings file
		must be an XML file that can be read by the {@link Presets} class.
		@param file The file to load/save settings in
	*/
	public Settings(ExtensionsDirectory directory, File file)
	{
		this.directory = directory;
		this.file = file;
		this.presets = new Presets();
		if (file.exists())
		{
			try
			{
				InputStream in = new BufferedInputStream(new FileInputStream(file));
				presets.loadFrom(in, file.getAbsolutePath());
				in.close();
			}
			catch (IOException ex)
			{
				throw new RuntimeException(ex);
			}
		}
		this.settings = presets.getDefault();
		this.disabledExtensionPoints = presets.getPreset("disabled-extension-points");
		this.disabledExtensions = presets.getPreset("disabled-extensions");
		hasChanged = false;
	}
	
	/*
		From the ExtensionsFilter interface
		-----------------------------------
	*/
	@Override
	public boolean isEnabled(ExtensionPoint extensionPoint) 
	{
		if (extensionPoint == null) return false;
		ExtensionsFile extFile = directory.getFileByExtensionId(extensionPoint.getId());
		if (extFile != null && (extFile.hasError() || !extFile.isValid())) return false;
		return disabledExtensionPoints.getSetting(extensionPoint.getId()) == null;
	}

	@Override
	public boolean isEnabled(Extension extension) 
	{
		if (extension == null) return false;
		ExtensionsFile extFile = directory.getFileByExtensionId(extension.getId());
		if (extFile != null && (extFile.hasError() || !extFile.isValid())) return false;
		return disabledExtensions.getSetting(extension.getId()) == null;
	}
	// --------------------------------------------------

	/**
		Enable/disable an extension.
		@param extensionId The ID of the extension to enable/disable
		@param enable TRUE to enable the extension, FALSE to disable it
	*/
	public synchronized void enableExtension(String extensionId, boolean enable)
	{
		disabledExtensions.setSetting(extensionId, enable ? null : "1");
		hasChanged = true;
	}
	
	/**
		Enable/disable an extension point.
		@param extensionPointId The ID of the extension point to enable/disable
		@param enable TRUE to enable the extension point, FALSE to disable it
	*/
	public synchronized void enableExtensionPoint(String extensionPointId, boolean enable)
	{
		disabledExtensionPoints.setSetting(extensionPointId, enable ? null : "1");
		hasChanged = true;
	}
	
	/**
		Get the interval, in seconds, between automatic checks 
		for new/modified/deleted extensions. 
		@return The interval in seconds or 0 if the auto-install feature
		 is disabled (the default)
	*/
	public int getAutoInstall()
	{
		int autoInstall = Values.getInt(settings.getSetting(AUTO_INSTALL));
		return autoInstall > 0 ? autoInstall : 0;
	}
	
	/**
		Change the interval, in seconds, between automatic checks 
		for new/modified/deleted extensions.
		@param autoInstall The interval in seconds, or 0 to disable the
			feature
	*/
	public synchronized void setAutoInstall(int autoInstall)
	{
		if (autoInstall < 0) autoInstall = 0;
		settings.setSetting(AUTO_INSTALL, Integer.toString(autoInstall));
		hasChanged = true;
	}
	
	/**
		Save the settings to disk. If the settings has not been
		changed the file is not written.
	*/
	public synchronized void save()
	{
		if (!hasChanged) return;
		try
		{
			OutputStream out = new FileOutputStream(file);
			presets.writeTo(out);
			out.close();
		}
		catch (IOException ex)
		{
			throw new RuntimeException(ex);
		}
		hasChanged = true;
	}
}
