/*
	$Id: SimpleExport.java 5433 2010-09-29 12:32:32Z nicklas $

	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Martin Svensson
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.plugins;

import net.sf.basedb.core.Annotatable;
import net.sf.basedb.core.Annotation;
import net.sf.basedb.core.AnnotationSet;
import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.BooleanParameterType;
import net.sf.basedb.core.Config;
import net.sf.basedb.core.DataQuery;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.DynamicQuery;
import net.sf.basedb.core.InvalidUseOfNullException;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemContext;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.Metadata;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.RawBioAssay;
import net.sf.basedb.core.ReporterScore;
import net.sf.basedb.core.ReporterScoreQuery;
import net.sf.basedb.core.RequestInformation;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.ProgressReporter;
import net.sf.basedb.core.StringParameterType;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.Unit;
import net.sf.basedb.core.Version;

import net.sf.basedb.core.Job.ExecutionTime;
import net.sf.basedb.core.data.BasicData;
import net.sf.basedb.core.data.RawData;
import net.sf.basedb.core.data.ReporterListScoreData;
import net.sf.basedb.core.plugin.About;
import net.sf.basedb.core.plugin.AboutImpl;
import net.sf.basedb.core.plugin.AbstractExporterPlugin;
import net.sf.basedb.core.plugin.ExportOutputStream;
import net.sf.basedb.core.plugin.Plugin;
import net.sf.basedb.core.plugin.InteractivePlugin;
import net.sf.basedb.core.plugin.NonRestartable;
import net.sf.basedb.core.plugin.Request;
import net.sf.basedb.core.plugin.Response;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Query;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.query.ResultIterator;
import net.sf.basedb.core.signal.SignalHandler;
import net.sf.basedb.core.signal.SignalTarget;
import net.sf.basedb.core.signal.ThreadSignalHandler;
import net.sf.basedb.util.Enumeration;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.formatter.Formatter;

import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.Set;
import java.util.HashSet;

/**
	Export plugin for exporting data from table listings in the web client.
	This plugin only works from within the web client, since it depends 
	on context and table information. It must have access to the current query
	which means that it always must be executed immediately, since queries can't 
	be saved as parameters.

   @author Nicklas
	@version 2.0
	@base.modified $Date: 2010-09-29 14:32:32 +0200 (Wed, 29 Sep 2010) $
*/
public class SimpleExport
	extends AbstractExporterPlugin
	implements InteractivePlugin, SignalTarget, NonRestartable
{

	/**
	   Plugin description
	*/
	private static final About about =
		new AboutImpl
		(
			"Table exporter",
			"Export all table listings in the web interface as tab-separated text files " +
			"or as XML. This plugin only works from within the web client, since it depends " +
			"on context and table information.",
			Version.getMajor() + "." + Version.getMinor() + "." + Version.getMaintenance(),
			"2006, Base 2 development team",
			null,
			null,
			"http://base.thep.lu.se"
		);

	private static final Set<GuiContext> guiContexts = new HashSet<GuiContext>();
	private RequestInformation configureJob;

	private ThreadSignalHandler signalHandler;
	
	/*
		From the Plugin interface
		-------------------------------------------
	*/
	public About getAbout()
	{
		return about;
	}
	public Plugin.MainType getMainType()
	{
		return Plugin.MainType.EXPORT;
	}
	public boolean supportsConfigurations()
	{
		return false;
	}
	public boolean requiresConfiguration()
	{
		return false;
	}
	// -------------------------------------------
	/*
		From the InteractivePlugin interface
		-------------------------------------------
	*/
	public Set<GuiContext> getGuiContexts()
	{
		if (guiContexts.size() == 0) 
		{
			initGuiContexts();
		}
		return guiContexts;
	}
	public String isInContext(GuiContext context, Object item)
	{
		return null;
	}
	public RequestInformation getRequestInformation(GuiContext context, String command) 
		throws BaseException
	{
		RequestInformation requestInformation = null;
		if (Request.COMMAND_CONFIGURE_JOB.equals(command))
		{
			requestInformation = getConfigureJobParameters(context);
		}
		return requestInformation;
	}
	public void configure(GuiContext context, Request request, Response response)
	{
		String command = request.getCommand();		
		try
		{
			if (command.equals(Request.COMMAND_CONFIGURE_JOB))
			{
				RequestInformation ri = getConfigureJobParameters(context);
				List<Throwable> errors = validateRequestParameters(ri.getParameters(), request);
				if (errors != null)
				{
					response.setError(errors.size() + " invalid parameters were found in the request",errors);
					return;
				}
				
				job.setValue("itemType", new StringParameterType(), context.getItem().name());
				job.setValue("subcontext", new StringParameterType(), context.getSubContext());

				storeValue(job, request, ri.getParameter(FORMAT));
				storeValue(job, request, ri.getParameter(WHICH_ITEMS));
				storeValues(job, request, ri.getParameter(WHICH_PROPERTIES));
				storeValue(job, request, ri.getParameter(COLUMN_PREFIX));
				storeValue(job, request, ri.getParameter(SAME_UNITS));
				
				if (request.getParameterValue(SAVE_AS) == null)
				{
					response.setDownloadImmediately("Export items", ExecutionTime.SHORTEST, true);
				}
				else
				{					
					Object parameterValue = request.getParameterValue(OVERWRITE); 
					boolean overwrite = parameterValue != null ? (Boolean)parameterValue : false;
					if (!pathCanBeUsed((String)request.getParameterValue(ri.getParameter(SAVE_AS).getName()), overwrite))
					{
						response.setError("File exists: " + (String)request.getParameterValue(ri.getParameter(SAVE_AS).getName()), null);
						return;
					}
					storeValue(job, request, ri.getParameter(SAVE_AS));
					storeValue(job, request, ri.getParameter(OVERWRITE));
					response.setExecuteImmediately("The job configuration is complete", ExecutionTime.SHORTEST, true);
				}
			}
		}
		catch (Throwable ex)
		{
			response.setError(ex.getMessage(), Arrays.asList(ex));
		}		
	}
	// -------------------------------------------

	/*
	 	From the AbstractExporterPlugin class
	 	--------------------------------------------
	*/
	private DbControl dc;
	private String successMessage;
	protected void begin(DbControl dc)
	{
		this.dc = dc;
		this.successMessage = "";
	}
	
	@SuppressWarnings("unchecked")
	protected void performExport(ExportOutputStream out, ProgressReporter progress)
		throws IOException
	{
		if (signalHandler != null) signalHandler.setWorkerThread(null);
		
		// Input parameters for the export
		Item itemType = Item.valueOf((String)job.getValue("itemType"));
		String subContext = Values.getString((String)job.getValue("subcontext"), "");
		String whichItems = (String)job.getValue(WHICH_ITEMS);
		String format = (String)job.getValue(FORMAT);
		boolean sameUnits = !Boolean.FALSE.equals(job.getValue(SAME_UNITS));
		List<String> columns = (List<String>)job.getValues(WHICH_PROPERTIES);
		String columnPrefix = (String)job.getValue(COLUMN_PREFIX);
		ItemContext cc = sc.getCurrentContext(itemType, subContext);
		
		//
		String entityName = null;
		Class entityClass = null;
		Metadata metadata = null;
		if (itemType == Item.RAWDATA)
		{
			RawBioAssay rba = RawBioAssay.getById(dc, sc.getCurrentContext(Item.RAWBIOASSAY).getId());
			entityName = rba.getRawDataType().getEntityName();
			entityClass = RawData.class;
		}
		else if (itemType == Item.REPORTERSCORE)
		{
			entityClass = ReporterScore.class;
			entityName = ReporterListScoreData.class.getName();
		}
		else if (itemType.getItemClass() != null)
		{
			entityClass = itemType.getItemClass();
		}
		else
		{
			entityClass = itemType.getDataClass();
		}
		if (entityName != null)
		{
			metadata = Metadata.getInstance(entityClass, entityName);
		}
		else if (entityClass != null)
		{
			metadata = Metadata.getInstance(entityClass);
		}

		// Parse the WHICH_PROPERTIES parameter. It is a list of property:type:title entries
		List<ExportedProperty> exportedProperties = 
			new ArrayList<ExportedProperty>(columns.size());
		List<String> properties = new ArrayList<String>(columns.size());
		boolean annotatable = itemType.getItemClass() != null && 
			Annotatable.class.isAssignableFrom(itemType.getItemClass());
		for (String column : columns)
		{
			Formatter<?> formatter = null;
			DataLoader dataloader = null;
			int splitAt = column.indexOf(":");
			if (splitAt >= 0)
			{
				String property = column.substring(0, splitAt);
				// Get a registered formatter and clear it to allow GC to reclaim it later
				Object possibleFormatter = cc.setObject("export.formatter." + property, null);
				if (possibleFormatter instanceof Formatter) 
				{
					formatter = (Formatter<?>)possibleFormatter;
				}
				// Get the registered dataloader for current column, if any, and clear it. 
				Object possibleDataloader = cc.setObject("export.dataloader." + property, null);				
				if (possibleDataloader instanceof DataLoader)
				{
					dataloader = (DataLoader<?>)possibleDataloader;
				}
			}
			ExportedProperty ep = ExportedProperty.parse(dc, metadata, column, columnPrefix, 
				annotatable, formatter, sameUnits, dataloader);
			exportedProperties.add(ep);
			properties.add(ep.name);
		}
		
		// Get the query from the current context
		Query query = cc.getQuery();
		Set<Integer>selectedItems = cc.getSelected();
		if (query instanceof DynamicQuery)
		{			
			cc.configureQuery(dc, (DynamicQuery)query, properties);
		}
		
		// Configure the query with additional filters depending on which items should be exported
		query.setReturnTotalCount(false);
		//boolean needToStep = false;
		int itemsPerStep = Config.getInt("export.max.items", 0);
		// Which of "all", "current page", or "selected" is selected?
		if ("all".equals(whichItems))
		{	
			query.setFirstResult(0);
			query.setMaxResults(itemsPerStep);
		}
		else if ("selected".equals(whichItems))
		{
			query.setFirstResult(0);
			query.setMaxResults(0);
			String propertyName = query instanceof ReporterScoreQuery ? "reporter.id" : "id";
			query.restrict(
				Restrictions.in(
					Hql.property(propertyName),
					Expressions.parameter("_selected_", selectedItems, Type.INT)
				)
			);
		}
		else
		{
			// current page, doesn't need additional configuration of the query
		}
		final String exclude = (String)cc.getObject("exclude");
		if (exclude != null)
		{
			query.restrict(
				Restrictions.not(
					Restrictions.in(
						Hql.property("id"),
						Expressions.parameter("_excludes_")
					)
				)
			);
			query.setParameter("_excludes_", Arrays.asList(Values.getInt(exclude.split(","))), Type.INT);
		}
		ThreadSignalHandler.checkInterrupted();
		exportProperties(dc, out, format, exportedProperties, itemType, query, progress);
	}
	
	protected void end(boolean success)
	{
		this.dc = null;
	}
	protected String getSuccessMessage()
	{
		return successMessage;
	}
	// -------------------------------------------
	/*
		From the SignalTarget interface
		-------------------------------------------
	*/
	public SignalHandler getSignalHandler()
	{
		signalHandler = new ThreadSignalHandler();
		return signalHandler;
	}
	// -------------------------------------------

	private void initGuiContexts()
	{
		for (Item i : Item.values())
		{
			if (i.getDataClass() != null)
			{
				guiContexts.add(new GuiContext(i, GuiContext.Type.LIST));
			}
		}
		guiContexts.add(new GuiContext(Item.REPORTERSCORE, GuiContext.Type.LIST));
		guiContexts.add(new GuiContext(Item.SPOTDATA, GuiContext.Type.LIST));
	}
	
	private RequestInformation getConfigureJobParameters(GuiContext context)
	{
		if (configureJob == null)
		{
			List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();

			parameters.add(getFormatParameter());
			parameters.add(getWhichItemsParameter());
			parameters.add(getWhichPropertiesParameter());
			parameters.add(getColumnPrefixParameter());
			parameters.add(getSameUnitsParameter());
			parameters.add(getSaveAsParameter(null, null, null, false));
			parameters.add(getOverwriteParameter(null, null));
			
			configureJob = new RequestInformation
			(
				Request.COMMAND_CONFIGURE_JOB,
				"Export " + context.getItem(),
				"Set options for which items and which columns that should be exported.",
				parameters, "simple_export.jsp"
			);
		}
		return configureJob;
	}
	
	private static final String FORMAT = "format";
	private PluginParameter<String> getFormatParameter()
	{
		String label = "Format";
		String description = "You may select to export in text or XML format.";
		Enumeration<String, String> options = new Enumeration<String, String>();
		options.add("text", "Tab-separated text file");
		options.add("xml", "XML");
		PluginParameter<String> formatParameter = new PluginParameter<String>
			(
				FORMAT, label, description, "text",
				new StringParameterType(255, "text", true, 1, 0, 0, options)
			);
		return formatParameter;
	}

	private static final String WHICH_ITEMS = "whichItems";
	private PluginParameter<String> getWhichItemsParameter()
	{
		String label = "Which items?";
		String description = "Select which items you want to export";
		Enumeration<String, String> options = new Enumeration<String, String>();
		options.add("selected", "Selected items");
		options.add("page", "Current page");
		options.add("all", "All pages");
		PluginParameter<String> whichItemsParameter = new PluginParameter<String>
			(
				WHICH_ITEMS, label, description, "all",
				new StringParameterType(255, "all", true, 1, 0, 0, options)
			);
		return whichItemsParameter;
	}

	private static final String WHICH_PROPERTIES = "whichProperties";
	private PluginParameter<String> getWhichPropertiesParameter()
	{
		String label = "Which properties?";
		String description = "Enter the names of the properties you want to export. " +
			"Use a colon to separate an optional title from the property name. For example: " +
			"name:Name of sample";
		PluginParameter<String> whichPropertiesParameter = new PluginParameter<String>
			(
				WHICH_PROPERTIES, label, description, null,
				new StringParameterType(255, null, true, 0, 0, 0)
			);
		return whichPropertiesParameter;
	}
	
	private static final String COLUMN_PREFIX = "columnPrefix";
	private PluginParameter<String> getColumnPrefixParameter()
	{
		String label = "Column prefix";
		String description = "Enter an optional prefix that will be used in the column header " +
			" names. For example: Sample.";
		PluginParameter<String> columnPrefixParameter = new PluginParameter<String>
			(
					COLUMN_PREFIX, label, description, null,
				new StringParameterType(255, null, true, 0, 0, 0)
			);
		return columnPrefixParameter;
	}

	private static final String SAME_UNITS = "sameUnits";
	private PluginParameter<Boolean> getSameUnitsParameter()
	{
		String label = "Same units";
		String description = "If this option is selected all annotation values in a " +
				"column that has a unit are exported in the same unit.";
		return new PluginParameter<Boolean>
			(
				SAME_UNITS, label, description, 
				new BooleanParameterType(true, true)
			);
	}
	
	/**
	 	Exports properties from a query to an output stream.  
	 	The export can be either tab-separated text or XML. 
	  	@param dc The DbControl to use when accessing the database
	  	@param out Output stream to export to.
	  	@param format Use 'xml' to export in XML-format. All other inputs will give a tab-separated format.
	  	@param exportedProperties A list of which properties to include in the export.
	  	@param itemType Kind of item the export is about 
	  	@param query This query is used to get the items to export.
	  	@param progress Optional progress reporter to get overview of how the export is proceeding.
	  	@throws IOException If anything goes wrong with export to the output stream.
	 */
	@SuppressWarnings("unchecked")
	public void exportProperties(DbControl dc, ExportOutputStream out, String format, List<ExportedProperty> exportedProperties, 
			Item itemType, Query query, ProgressReporter progress)
		throws IOException
	{
		if (dc == null) throw new InvalidUseOfNullException("dc");
		if (out == null) throw new InvalidUseOfNullException("out");
		if (exportedProperties == null) throw new InvalidUseOfNullException("exportedProperties");
		if (itemType == null) throw new InvalidUseOfNullException("itemType");
		if (query == null) throw new InvalidUseOfNullException("query");
		
		this.dc = dc;
		boolean hasAnnotations = false;
		int itemsPerStep = Config.getInt("export.max.items", 0);
		boolean needToStep = query.getMaxResults() > 0 ? (itemsPerStep > 0) : false;
		long totalItems = query.getMaxResults() == 0 ? query.count(dc) : query.getMaxResults();
		
		for (ExportedProperty ep : exportedProperties)
		{
			hasAnnotations |= ep.annotationType != null;
		}
		QueryWrapper queryWrapper = null;
		// The query wrapper allows us to get results from the query in a simplier way
		if (query instanceof ItemQuery)
		{
			queryWrapper = new ItemQueryWrapper(dc, (ItemQuery<BasicItem>)query);
		}
		else if (query instanceof DataQuery)
		{
			queryWrapper = new DataQueryWrapper(dc, (DataQuery<BasicData>)query);
		}
		else if (query instanceof ReporterScoreQuery)
		{
			queryWrapper = new ReporterScoreQueryWrapper(dc, (ReporterScoreQuery)query);
		}
		else if (query instanceof DynamicQuery)
		{
			DynamicQuery dynamicQuery = (DynamicQuery)query;			
			queryWrapper = new DynamicQueryWrapper(dc, dynamicQuery);
		}
		
		// Open stream for output and set file name and MIME type
		Writer exportStream = new OutputStreamWriter(out, "UTF-8");
		// Create export template for outputting the data
		ExportTemplate template = "xml".equals(format) ? 
			new XMLTemplate(exportStream, exportedProperties, itemType) : 
			new PlainTextTemplate(exportStream, exportedProperties, itemType);
		out.setFilename(itemType.name().toLowerCase()+".export.txt");
		out.setMimeType(template.getMimeType());
		out.setCharacterSet("UTF-8");
		
		// Write start of document and headers
		template.beginDocument();
		template.writeHeaders();
		
		long numExported = 0;
		ResultIterator result = null;
		try
		{
			do
			{
				ThreadSignalHandler.checkInterrupted();
				result = queryWrapper.iterate();
				while (result.hasNext())
				{
					ThreadSignalHandler.checkInterrupted();
					numExported++;
					Object item = result.next();
					
					int id = queryWrapper.getId(item);
					AnnotationSet as = null;
					if (hasAnnotations)
					{
						Annotatable a = (Annotatable)item;
						if (a.isAnnotated()) as = a.getAnnotationSet();
					}
					
					template.beginItem(id);
					for (ExportedProperty ep : exportedProperties)
					{
						if (ep.annotationType != null)
						{
							List<?> values = null;
							String unitSymbol = null;
							if (ep.dataloader == null)
							{
								Annotation a = as != null && as.hasAnnotation(ep.annotationType) ?
										as.getAnnotation(ep.annotationType) : null;
								if (a != null)
								{
									values = a.getValues(ep.unit);
									if (ep.unit == null) unitSymbol = a.getUnitSymbol(null);
								}
							}
							else
							{
								Object data = ep.dataloader.getData(ep, item);
								if (data instanceof Collection)
								{
									values = new ArrayList<Object>((Collection)data);
								}
								else
								{
									values = Arrays.asList(data);
								}
								if (ep.unit == null && ep.annotationType.supportUnits())
								{
									Unit unit = ep.annotationType.getDefaultUnit();
									if (unit != null) unitSymbol = unit.getDisplaySymbol();
								}
							}
							template.writeAnnotations(ep, values, unitSymbol);
						}
						else
						{
							Object data = ep.dataloader == null ? 
									queryWrapper.getData(ep, item) : ep.dataloader.getData(ep, item);
							
							if (data instanceof Collection)
							{
								template.writeCollection(ep, (Collection)data);
							}
							else
							{
								template.writeProperty(ep, data);
							}
						}
					}
					template.endItem();
					if (progress != null)
					{
						int percent = (int)(100L * numExported / totalItems);
						progress.display(percent, numExported + " of " + totalItems + " done.");
					}
				}
				if (needToStep)
				{
					query.setFirstResult(query.getFirstResult() + itemsPerStep);
				}
			}
			while (needToStep && (numExported % itemsPerStep == 0));
			template.endDocument();
			successMessage = numExported + " items were exported successfully";
		}
		catch (Exception ex)
		{
			throw new IOException(ex.getMessage(), ex);
		}
		finally
		{
			exportStream.flush();
		}
	}
}
