/*
	$Id: AnnotationBetweenRestriction.java 5320 2010-04-21 09:48:29Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.query.Query;
import net.sf.basedb.util.EqualsHelper;

import org.hibernate.SQLQuery;

/**
	Restricts a query using annotation values between a low and high
	value.
	
	@author Nicklas
	@version 2.0
	@see net.sf.basedb.core.query.Annotations
	@base.modified $Date: 2010-04-21 11:48:29 +0200 (Wed, 21 Apr 2010) $
*/
public class AnnotationBetweenRestriction
	extends AnnotationRestriction
{
	private final Object lowValue;
	private final Object highValue;

	/**
		Create a new annotation restriction.
		
		@deprecated Use {@link #AnnotationBetweenRestriction(String, AnnotationType, Object, Object, boolean)}
			instead with alias=null
	*/
	public AnnotationBetweenRestriction(AnnotationType annotationType, Object lowValue, Object highValue, boolean includeInheriting)
		throws InvalidDataException
	{
		this(null, annotationType, lowValue, highValue, includeInheriting);
	}
	
	/**
		Create a new annotation restriction.
		
		@param alias The alias of a joined item where the annotations are
			located or null to use the root entity of the query
		@param annotationType The annotation type to use in the query
		@param lowValue The low value to use in the query, it must be of the 
			correct value type for the annotation as defined by the 
			{@link AnnotationType#getValueType()} property
		@param highValue The high value to use in the query, it must be of the 
			correct value type for the annotation as defined by the 
			{@link AnnotationType#getValueType()} property
		@param includeInheriting If items inherting the annotation should be returned
			by the query or not
		@throws InvalidDataException If any of the parameters are null
			or not follow the rules above.
		@since 2.4
	*/
	public AnnotationBetweenRestriction(String alias, AnnotationType annotationType, Object lowValue, Object highValue, boolean includeInheriting)
		throws InvalidDataException
	{
		super(alias, annotationType, includeInheriting);
		if (lowValue == null) throw new InvalidUseOfNullException("lowValue");
		if (highValue == null) throw new InvalidUseOfNullException("highValue");
		if (!valueType.isCorrectType(lowValue))
		{
			throw new InvalidDataException("Low value '"+lowValue+"' is a "+
				lowValue.getClass()+", not a "+valueType);
		}
		if (!valueType.isCorrectType(highValue))
		{
			throw new InvalidDataException("Hig value '"+highValue+"' is a "+
				highValue.getClass()+", not a "+valueType);
		}
		this.lowValue = lowValue;
		this.highValue = highValue;
	}


	/*
		From the AnnotationRestriction class
		-------------------------------------------
	*/
	String getRestrictionSql(Query query, DbControl dc)
		throws BaseException
	{
		org.hibernate.dialect.Dialect dialect = HibernateUtil.getDialect();
		return "v."+dialect.openQuote()+"value"+dialect.closeQuote()+" between :lowValue and :highValue";
	}
	
	void setRestrictionParameters(SQLQuery query)
	{
		query.setParameter("lowValue", lowValue);
		query.setParameter("highValue", highValue);
	}
	// -----------------------------------------

	/*
		From the Object class
		-------------------------------------------
	*/
	/**
		This restriction is equal to another AnnotationBetweenRestriction if they 
		have the same annotation type, upper and lower bound and inheritance setting.
	*/
	@Override
	public boolean equals(Object other)
	{
		Boolean eq = maybeEquals(other);
		if (eq != null) return eq;
		AnnotationBetweenRestriction o = (AnnotationBetweenRestriction)other;
		return EqualsHelper.equals(lowValue, o.lowValue) && 
			EqualsHelper.equals(highValue, o.highValue);
	}
	@Override
	public int hashCode()
	{
		return super.hashCode() + EqualsHelper.hashCode(lowValue, highValue);
	}
	// -------------------------------------------

	

}

