/*
	$Id: AnnotationSet.java 5319 2010-04-20 08:02:16Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Johan Enell, Jari Häkkinen, Nicklas Nordborg, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import net.sf.basedb.core.query.EntityQuery;
import net.sf.basedb.core.query.Expression;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.signal.ThreadSignalHandler;
import net.sf.basedb.core.snapshot.AnnotationSetSnapshot;
import net.sf.basedb.core.snapshot.AnnotationSnapshot;
import net.sf.basedb.core.snapshot.AnnotationTypeFilter;
import net.sf.basedb.core.snapshot.SnapshotManager;

import net.sf.basedb.core.data.AnnotationData;
import net.sf.basedb.core.data.AnnotationSetData;
import net.sf.basedb.core.data.AnnotationTypeData;
import net.sf.basedb.util.AnnotationUtil;
import net.sf.basedb.util.filter.Filter;

/**
	An annotation set is the container for all annotations on an item.
	An item can have one annotation of each annotation type that is defined
	for it.
	<p>
	An item can also inherit annotation from it's parents. It is possible
	to inherit all annotations or only specific ones. In the first case this means
	that if the parent gets new annotations those are automatically inherited. 
	In the second case the inherticane is lost if the annotation is deleted. If a
	new annotation of the same annotation type is creted later, it is NOT automatically
	inherited again.

	@author Nicklas
	@version 2.0
	@see AnnotationType
	@see Annotation
	@base.modified $Date: 2010-04-20 10:02:16 +0200 (Tue, 20 Apr 2010) $
*/
public class AnnotationSet
	extends BasicItem<AnnotationSetData>
	implements Transactional
{
	/**
		The type of item represented by this class.
		@see Item#ANNOTATIONSET
		@see #getType()
	*/
	public static final Item TYPE = Item.ANNOTATIONSET;
	
	/**
		This filter gives everybody read permission to annotations.
	*/
	private static final QueryRuntimeFilter RUNTIME_FILTER = new QueryRuntimeFilterImpl();

	/**
		Create a new annotation set for the specified item.
		@see AnnotatedItem#getAnnotationSet()
	*/
	static AnnotationSet getNew(DbControl dc, Annotatable item)
		throws BaseException
	{	
		AnnotationSet as = dc.newItem(AnnotationSet.class, item);		
		as.getData().setItemType(item.getType().getValue());		
		return as;
	}
	/**
		Get an <code>AnnotationSet</code> item when you know the id.
	
		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The id of the item to load
		@return The <code>AnnotationSet</code> item
		@throws ItemNotFoundException If an item with the specified 
			id is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have read permission to the item
		@throws BaseException If there is another error
	*/
	public static AnnotationSet getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		AnnotationSet as = dc.loadItem(AnnotationSet.class, id);
		if (as == null) throw new ItemNotFoundException("AnnotationSet[id="+id+"]");
		return as;
	}	
	/**
		Get a {@link ItemQuery} object configured to retrieve <code>AnnotationSet</code> 
		items. This query may return items which the logged in user doesn't have 
		read permission to. At the moment there is no way to solve this problem.
		@return An {@link ItemQuery} object
	*/
	static ItemQuery<AnnotationSet> getQuery()
	{
		return new ItemQuery<AnnotationSet>(AnnotationSet.class, RUNTIME_FILTER);
	}

	/**
		Load the id of the item that this annotation set belongs to.
		@since 2.14
	*/
	static int loadItemId(org.hibernate.Session session, Item itemType, int annotationSetId)
	{
		String hql = "SELECT item.id"+
			" FROM "+itemType.getDataClass().getName()+" item"+
			" WHERE item.annotationSet = :annotationSet";
		org.hibernate.Query query = HibernateUtil.createQuery(session, hql);
		query.setInteger("annotationSet", annotationSetId);
		return HibernateUtil.loadData(Integer.class, query);
	}
	
	/**
		Create a new annotation snapshot for the given annotation set.
		NOTE! This method will always create a new snapshot from the 
		database. The snapshot is not saved in the cache.
		<p>
		It is recommended that snapshots are cached as files in the static cache
		since they are much quiker to load. Use {@link SnapshotManager#getSnapshot(DbControl, int)} 
		is instead of this method.
		
		@param dc A DbControl to use for database access
		@param annotationSetId The ID of the annotation set
		@return A snapshot, or null if the annotation set with the given
			id doesn't exist
		@since 2.14
	*/
	public static AnnotationSetSnapshot createSnapshot(DbControl dc, int annotationSetId)
	{
		AnnotationSetData set = HibernateUtil.loadData(dc.getHibernateSession(), AnnotationSetData.class, annotationSetId);
		if (set == null) return null;
		
		AnnotationSetSnapshot snapshot = new AnnotationSetSnapshot();
		snapshot.init(set);
		return snapshot;
	}
	
	/**
		The item this annotation set belongs to.
	*/
	private Annotatable item;
	
	/**
		Flag to indicate if this annotation set has been modified 
		and the snapshot must be removed.
	*/
	private boolean snapshotIsInvalid;
	
	/**
		Regular constructor if the item is not known beforehand.
	*/
	AnnotationSet(AnnotationSetData annotationSetData)
	{
		super(annotationSetData);		
	}

	/**
		Special constructor to avoid {@link #getItem()} to hit the database
		when the item is known beforehand.
		@see AnnotatedItem#getAnnotationSet()
	*/
	AnnotationSet(AnnotationSetData annotationSetData, Annotatable item)
	{
		super(annotationSetData);
		setItem(item);
	}

	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------

	/*
		From the BasicItem class
		-------------------------------------------
	*/
	/**
		Always return TRUE. An annotation set always has a one-to-one relation
		to some item. The annotation set is automatically deleted if that
		item is deleted.
	*/
	@Override
	public boolean isUsed()
		throws BaseException
	{
		return true;
	}
	/**
		Get the item this annotation set belongs to.
		@since 2.2
	*/
	@Override
	public Set<ItemProxy> getUsingItems()
		throws BaseException
	{
		Set<ItemProxy> using = super.getUsingItems();
		using.add(new ItemProxy(getItemId(), getItemType()));
		return using;
	}
	
	/**
		READ permission is granted to all users. CREATE, WRITE and DELETE permissions are granted
		if the logged in user has WRITE permission on the associated item.
	*/
	void initPermissions(int granted, int denied)
		throws BaseException
	{	
		Annotatable item = null;
		try
		{
			item = getItem();			
		}
		catch (PermissionDeniedException ex)
		{}
		if (item != null && item.hasPermission(Permission.WRITE))
		{
			granted |= Permission.grant(Permission.CREATE, Permission.READ, Permission.WRITE, Permission.DELETE);				
		}
		else if (item != null && item.hasPermission(Permission.USE))
		{
			granted |= Permission.grant(Permission.USE);
		}
		else
		{
			granted |= Permission.grant(Permission.READ);				
		}	
		super.initPermissions(granted, denied);
	}
	PluginPermission getPluginPermissions()
	{
		return PluginPermission.NO_PLUGIN;
	}
	/**
		Remove the cached snapshot if needed.
	*/
	@Override
	void onBeforeCommit(Action action) 
		throws BaseException
	{
		super.onBeforeCommit(action);
		if ((snapshotIsInvalid && action == Action.UPDATE) || action == Action.DELETE) 
		{
			if (!SnapshotManager.removeSnapshot(getId()))
			{
				throw new BaseException("Could not remove cached snapshot for annotation set[id=" + getId() + "]");
			}
		}
	}
	// -------------------------------------------

	/**
		Get the item this annotation set belongs to.
		@return The <code>Annotatable</code> item
		@throws PermissionDeniedException If the logged in user doesn't have 
			read permission to the item
		@throws BaseException If there is another error
	*/
	public Annotatable getItem()
		throws PermissionDeniedException, BaseException
	{
		return getItem(null);
	}
	
	/**
		Get the item this annotation set belongs to using a specific 
		DbControl for database access. If the item has already been loaded
		it is re-loaded by the given DbControl.
		
		@param dc An open DbControl, or null to use the current DbControl associated
			with this annotation set
		@return The <code>Annotatable</code> item
		@since 2.9
	*/
	public Annotatable getItem(DbControl dc)
		throws PermissionDeniedException, BaseException
	{
		if (item == null)
		{	
			Integer itemId = getData().getItemId();
			if (dc == null) dc = getDbControl();
			if (itemId == null) 
			{
				itemId = loadItemId(dc.getHibernateSession(), getItemType(), getId());
				getData().setItemId(itemId);
			}
			item = (Annotatable)getItemType().getById(dc, itemId);			
		}
		else if (dc != null)
		{
			item = (Annotatable)getItemType().getById(dc, item.getId());
		}
		return item;
	}
	
	/**
		Get the ID of the item this annotation set belongs to. This may
		need to hit the database unless the id is already known.
		@since 2.14
	*/
	public int getItemId()
	{
		Integer itemId = getData().getItemId();
		if (itemId == null)
		{
			itemId = loadItemId(getDbControl().getHibernateSession(), getItemType(), getId());
			getData().setItemId(itemId);
		}
		return itemId;
	}
	
	/**
		Set the item this annotation set belongs to.
	*/
	void setItem(Annotatable item)
	{
		this.item = item;
		if (item.getId() != 0) getData().setItemId(item.getId());
	}
	
	/**
		Get the {@link Item} type of the item this annotation set belongs to.
	*/
	public Item getItemType()
	{
		return Item.fromValue(getData().getItemType());
	}
	
	/**
		Get the {@link Annotation} object of the specified annotation type.
		If the annotation set doesn't contain an annotation of the specified type
		a new object is created and automatically saved to the database when
		{@link DbControl#commit()} is called. To check if an annotation of a certain
		annotation type exists without creating a new annotation use the
		{@link #hasAnnotation(AnnotationType)} method.
		
		@param annotationType The annotation type
		@return An <code>Annotation</code> object
		@throws PermissionDeniedException If a new annotation must be created
			and the logged in user doesn't have write permission
		@throws InvalidDataException If a new annotation must be created
			and the annotation type isn't valid for this type of item
		@throws BaseException If there is another error
		@see #findAnnotations(DbControl, AnnotationType, Boolean)
	*/
	public Annotation getAnnotation(AnnotationType annotationType)
		throws PermissionDeniedException, InvalidDataException, BaseException
	{
		if (annotationType == null) throw new InvalidUseOfNullException("annotationType");
		DbControl dc = getDbControl();
		Annotation a = dc.getItem(Annotation.class, getData().getAnnotations().get(annotationType.getData()));
		if (a == null)
		{
			checkPermission(Permission.WRITE);
			// AnnotatableProxy items allow annotations of all types
			boolean isAnnotatableProxy = AnnotatableProxy.class.isAssignableFrom(getItemType().getItemClass());
			if (!isAnnotatableProxy && !annotationType.isEnabledForItem(getItemType()) &&  
				!annotationType.isProtocolParameter(getItem().getProtocol()))
			{
				throw new InvalidAnnotationException(annotationType, getItemType());
			}
			a = Annotation.getNew(dc, annotationType, this);
			getData().getAnnotations().put(annotationType.getData(), a.getData());
			dc.saveItemIf(this, a, false);
			setSnapshotInvalid();
		}
		return a;
	}
	
	/**
		Check if the annotation set contains an annotation of the specified type.
		@param annotationType An <code>AnnotationType</code> object
		@return TRUE if the annotation set contains an annotation of 
			the specified type, FALSE otherwise
	*/
	public boolean hasAnnotation(AnnotationType annotationType)
	{
		if (annotationType == null) return false;
		return getData().getAnnotations().containsKey(annotationType.getData());
	}

	/**
	 	Find an annotation of the specified annotation type among all annotations
	 	including the inherited annotations. This method will first search for the
	 	annotation in the following order:
	 	<ol>
	 	<li>The primary annotations
	 	<li>The inherited annotations
	 	<li>The inherited annotation sets
	 	</ol>
	 	If no annotation of the specified type is found, null is returned.
	 	If more than one annotation of the specified type exists only the first
	 	one found is returned.
	 	
		@param annotationType The annotation type to look for
		@return An <code>Annotation</code> object or null
		@deprecated Use {@link #findAnnotations(DbControl, AnnotationType, Boolean)} instead
	*/
	public Annotation findAnnotation(AnnotationType annotationType)
		throws PermissionDeniedException
	{
		List<Annotation> annotations = findAnnotations(getDbControl(), annotationType, null);
		return annotations.size() > 0 ? annotations.get(0) : null;
	}
	
	/**
	 	Find annotations of the specified annotation type among all annotations
	 	including the inherited annotations. This method will search for 
	 	annotations in the following order:
	 	<ol>
	 	<li>The primary annotations
	 	<li>The inherited annotations
	 	<li>The inherited annotation sets
	 	</ol>
	 	If no annotation of the specified type is found, an empty list is returned.
	 	<p>
	 	Note! This method only looks in the database. Changes that have been made in
	 	the current transaction and not yet committed to the database are not included
	 	in the result.
	 	<p>
	 	NOTE!!!! Do not use this method if you need to find many annotation in many items.
	 	It has bad performance. It is better to use the cached snapshots. 
	 	See {@link SnapshotManager}.
	 	
	 	@param dc The DbControl to use for database access
		@param annotationType The annotation type to look for
		@param findInherited TRUE if the method always should try to find inherited annotations,
			FALSE if it never should try, or null if it only should try if it hasn't found any
			annotation so far
		@return A list <code>Annotation</code> objects (may be empty)
	 	@throws PermissionDeniedException If the logged in user don't have the 
	 		proper permissions to access the database. 
		@since 2.5
	*/
	public List<Annotation> findAnnotations(DbControl dc, AnnotationType annotationType, Boolean findInherited)
		throws PermissionDeniedException
	{
		if (annotationType == null) throw new InvalidUseOfNullException("annotationType");
		if (!isInDatabase()) return Collections.emptyList();
		// List for storing the result
		List<Annotation> annotations = new LinkedList<Annotation>();
		
		org.hibernate.Session session = dc.getHibernateSession();
		// Find primary annotation
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "GET_PRIMARY_ANNOTATION");
		/*
			SELECT ad
			FROM AnnotationData ad
			WHERE ad.annotationType = :annotationType
			AND ad.annotationSet = :annotationSet
		*/
		query.setEntity("annotationType", annotationType.getData());
		query.setEntity("annotationSet", this.getData());
		AnnotationData ad = HibernateUtil.loadData(AnnotationData.class, query);
		if (ad != null) annotations.add(dc.getItem(Annotation.class, ad));
		
		if (Boolean.TRUE.equals(findInherited) || (annotations.isEmpty() && findInherited == null))
		{
			// Find among inherited annotations
			query = HibernateUtil.getPredefinedQuery(session, "GET_DIRECTLY_INHERITED_ANNOTATION");
			/*
				SELECT ad 
				FROM AnnotationData ad
				JOIN ad.inheritingSets ans
				WHERE ad.annotationType = :annotationType
				AND ans = :annotationSet
			*/
			query.setEntity("annotationType", annotationType.getData());
			query.setEntity("annotationSet", this.getData());
			List<AnnotationData> inherited = HibernateUtil.loadList(AnnotationData.class, query, dc.getSessionControl());
			for (AnnotationData a : inherited)
			{
				annotations.add(dc.getItem(Annotation.class, a));
			}
		}
	
		if (Boolean.TRUE.equals(findInherited) || (annotations.isEmpty() && findInherited == null))
		{
			// Find among inherited annotation sets
			query = HibernateUtil.getPredefinedQuery(session, "GET_INDIRECTLY_INHERITED_ANNOTATION");
			/*
				SELECT ad
				FROM AnnotationData ad
				JOIN ad.annotationSet.inheritingSets ians
				WHERE ad.annotationType = :annotationType
				AND ians = :annotationSet
			*/
			query.setEntity("annotationType", annotationType.getData());
			query.setEntity("annotationSet", this.getData());
			List<AnnotationData> inherited = HibernateUtil.loadList(AnnotationData.class, query, dc.getSessionControl());
			for (AnnotationData a : inherited)
			{
				annotations.add(dc.getItem(Annotation.class, a));
			}
		}
		return annotations;
	}	
	
	/**
		Delete the annotation of the specified annotation type.
		@param annotationType The type of the annotation to delete
		@throws PermissionDeniedException If the logged in user
			doesn't have write permission
		@throws InvalidDataException If the annotation type is null
		@throws BaseException If there is another error
	*/
	public void removeAnnotation(AnnotationType annotationType)
		throws PermissionDeniedException, InvalidDataException, BaseException
	{
		checkPermission(Permission.WRITE);
		if (annotationType == null) throw new InvalidUseOfNullException("annotationType");
		DbControl dc = getDbControl();
		if (getData().getAnnotations().containsKey(annotationType.getData()))
		{
			AnnotationData ad = getData().getAnnotations().remove(annotationType.getData());
			if (ad != null && ad.getId() != 0)
			{
				Annotation a = dc.getItem(Annotation.class, ad);
				dc.deleteItem(a);
				setSnapshotInvalid();
			}
		}
	}
	
	/**
		Get a query that reurns all primary annotations in this annotation
		set.
		@return An <code>ItemQuery</code> object
	*/
	public ItemQuery<Annotation> getAnnotations()
	{
		ItemQuery<Annotation> query = Annotation.getQuery();
		query.restrictPermanent(
			Restrictions.eq(
				Hql.property("annotationSet"),
				Hql.entity(this)
			)
		);
		return query;
	}

	/**
		Inherit an annotation.
		@param annotation The annotation to inherit
		@throws PermissionDeniedException If the logged in user
			doesn't have write permission for this annotation set or
			use permission for the annotation
		@throws InvalidDataException If the annotation is null
	*/
	public void inheritAnnotation(Annotation annotation)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		if (annotation == null) throw new InvalidUseOfNullException("annotation");
		annotation.checkPermission(Permission.USE);
		if (getData().getInherited().add(annotation.getData()))
		{
			setSnapshotInvalid();
		}
	}
	
	/**
		Remove an inherited an annotation.
		@param annotation The annotation to remove
		@throws PermissionDeniedException If the logged in user
			doesn't have write permission for this annotation set
		@throws InvalidDataException If the annotation is null
	*/
	public void removeInheritedAnnotation(Annotation annotation)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		if (annotation == null) throw new InvalidUseOfNullException("annotation");
		if (getData().getInherited().remove(annotation.getData()))
		{
			setSnapshotInvalid();
		}
	}
	
	/**
		Check if an annotation is inherited or not by this annotation set.
		@param annotation The annotation to check
	  	@return TRUE if inherited, FALSE otherwise. 
		@throws InvalidDataException If the annotation is null
	*/
	public boolean isInherited(Annotation annotation)
		throws InvalidDataException
	{
		if (annotation == null) throw new InvalidUseOfNullException("annotation");
		return getData().getInherited().contains(annotation.getData());
	}
	
	/**
		Get a query that returns directly inherited annotations in this annotation set.
		Directly inherited annotations are annotations that are linked to this annotation
		set.
		@see #inheritAnnotation(Annotation)
		@return An <code>ItemQuery</code> object
	 	@throws BaseException If getting the query fails.
	*/
	public ItemQuery<Annotation> getInheritedAnnotations()
		throws BaseException
	{
		ItemQuery<Annotation> query = Annotation.getQuery();
		query.joinPermanent(Hql.innerJoin("inheritingSets", Item.ANNOTATIONSET.getAlias()));
		query.restrictPermanent(
			Restrictions.eq(
				Hql.alias(Item.ANNOTATIONSET.getAlias()),
				Hql.entity(this)
			)
		);
		return query;
	}
	
	/**
	  	Get a query that returns all inherited (directly and indirectly) annotations in
	  	this annotation set. The directy inherited annotations include annotations added
	  	by {@link #inheritAnnotation(Annotation)}, the indirectly inherited annotations
	  	include annotations in annotation sets added by 
	  	{@link #inheritAnnotationSet(AnnotationSet)}
	  	
	  	@see #getInheritedAnnotations()
		@return An <code>ItemQuery</code> object
	*/
	public ItemQuery<Annotation> getAllInheritedAnnotations()
	{
		ItemQuery<Annotation> query = Annotation.getQuery();
		query.joinPermanent(Hql.leftJoin("inheritingSets", "direct"));
		query.joinPermanent(Hql.leftJoin("annotationSet.inheritingSets", "indirect"));
		Expression thisEntity = Expressions.parameter("annotationSet", this.getId(), Type.INT);
		query.restrictPermanent(
			Restrictions.or(
				Restrictions.eq(Hql.alias("direct"), thisEntity),
				Restrictions.eq(Hql.alias("indirect"), thisEntity)
			)
		);
		query.setDistinct(true);
		return query;
	}
	
	/**
		Inherit an annotation set.
		@param annotationSet The annotation set to inherit
		@throws PermissionDeniedException If the logged in user
			doesn't have write permission for this annotation set or
			use permission for the annotation set to inherit
		@throws InvalidDataException If the annotation set is null
	*/
	public void inheritAnnotationSet(AnnotationSet annotationSet)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		if (annotationSet == null) throw new InvalidUseOfNullException("annotationSet");
		annotationSet.checkPermission(Permission.USE);
		if (getData().getInheritedSets().add(annotationSet.getData()))
		{
			setSnapshotInvalid();
		}

	}
	
	/**
		Remove an inherited an annotation set.
		@param annotationSet The annotation set to remove
		@throws PermissionDeniedException If the logged in user
			doesn't have write permission for this annotation set
		@throws InvalidDataException If the annotation set is null
	*/
	public void removeInheritedAnnotationSet(AnnotationSet annotationSet)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		if (annotationSet == null) throw new InvalidUseOfNullException("annotationSet");
		if (getData().getInheritedSets().remove(annotationSet.getData()))
		{
			setSnapshotInvalid();
		}
	}

	/**
		Check if an annotation set is inherited or not by this annotation set.
		@param annotationSet The annotation set to check
	 	@return TRUE if inherited, FALSE otherwise. 
		@throws InvalidDataException If the annotation set is null
	*/
	public boolean isInherited(AnnotationSet annotationSet)
		throws InvalidDataException
	{
		if (annotationSet == null) throw new InvalidUseOfNullException("annotationSet");
		return getData().getInheritedSets().contains(annotationSet.getData());
	}

	/**
		Get a query that returns all inherited annotation sets in this annotation set.
		@return An <code>ItemQuery</code> object
		@throws BaseException If there is an error
	*/
	public ItemQuery<AnnotationSet> getInheritedAnnotationSets()
	{
		ItemQuery<AnnotationSet> query = AnnotationSet.getQuery();
		query.joinPermanent(Hql.innerJoin("inheritingSets", "i"));
		query.restrictPermanent(
			Restrictions.eq(
				Hql.alias("i"),
				Hql.entity(this)
			)
		);
		return query;
	}

	/**
		Automatically inherit all annotations of the specified annotation
		type from parent items unless this annotation set already contains
		an annotation of the specified type.
		<p>
		This method will only consider parent items that the logged in user
		has access to with USE permission.
		<p>
		This method may result in inheriting multiple annotations (from different
		parent items) of the specified annotation type.
		
		@param dc The DbControl to use for database access
		@param annotationType The annotation type to look for
		@param preferAnnotationSet If TRUE, the entire annotation set containing the
			found annotation is inherited, otherwise only the found annotation is inherited
		@since 2.9
		@deprecated In 2.15, use {@link #autoInherit(DbControl, Collection, SnapshotManager, boolean)}
			instead since it has better performance
	*/
	public int autoInherit(DbControl dc, final AnnotationType annotationType, boolean preferAnnotationSet)
	{
		return autoInherit(dc, Collections.singletonList(annotationType), new SnapshotManager(), preferAnnotationSet);
	}
	
	/**
		Automatically inherit all annotations of one or more specified annotation
		types from parent items that are currntly missing from this annotation
		set
		<p>
		This method will only consider parent items that the logged in user
		has access to with USE permission.
		<p>
		This method may result in inheriting multiple annotations (from different
		parent items) of the specified annotation type.
		
		@param dc The DbControl to use for database access
		@param annotationTypes The annotation type to look for
		@param manager A snapshot manager that is used to look up annotations
		@param preferAnnotationSet If TRUE, the entire annotation set containing the
			found annotation is inherited, otherwise only the found annotation is inherited
		@since 2.15
	*/
	public int autoInherit(DbControl dc, Collection<AnnotationType> annotationTypes, 
		SnapshotManager manager, boolean preferAnnotationSet)
	{
		if (annotationTypes == null || annotationTypes.isEmpty()) return 0;
		// Find the annotation types that are currently missing from this annotation set
		Set<AnnotationType> missing = new HashSet<AnnotationType>();
		AnnotationSetSnapshot snapshot = manager.getSnapshot(dc, getId());
		AnnotationTypeFilter atFilter = new AnnotationTypeFilter();
		if (snapshot == null)
		{
			// If snapshot==null this annotation set is new and unsaved 
			// and doesn't have any annotations
			missing.addAll(annotationTypes);
		}
		else
		{
			for (AnnotationType at : annotationTypes)
			{
				atFilter.setAnnotationType(at);
				if (manager.findAnnotations(dc, snapshot, atFilter, true).size() == 0)
				{
					missing.add(at);
				}
			}
		}
		if (missing.isEmpty()) return 0;
		ThreadSignalHandler.checkInterrupted();

		// Find all annotated parent items for 
		// which the logged in user has USE permission
		int numInherited = 0;
		Annotatable item = getItem(dc);
		Filter<Annotatable> parentFilter = new Filter<Annotatable>()
		{
			@Override
			public boolean evaluate(Annotatable a)
			{
				return a.isAnnotated() && a.hasPermission(Permission.USE);
			}
		};
		Set<Annotatable> allParents = AnnotationUtil.getAllAnnotatableParentItems(dc, item, parentFilter);

		// For each parent -- inherit any annotation that match the given annotation types
		for (Annotatable parent : allParents)
		{
			ThreadSignalHandler.checkInterrupted();
			AnnotationSet pas = parent.getAnnotationSet();
			snapshot = manager.getSnapshot(dc, pas.getId());
			
			for (AnnotationType at : missing)
			{
				atFilter.setAnnotationType(at);
				List<AnnotationSnapshot> tmp = manager.findAnnotations(dc, snapshot, atFilter, false);
				if (tmp.size() > 0)
				{
					numInherited++;
					if (preferAnnotationSet)
					{
						inheritAnnotationSet(pas);
						break;	// No need to check more annotation types since the complete set is inherited
					}
					else
					{
						inheritAnnotation(tmp.get(0).getAnnotation(dc));
					}
				}
			}
		}
		return numInherited;
	}

	
	/**
		Copy annotations from another annotatable item to this annotation set.
		This method will copy all annotations with an annotation type that is 
		valid for this item type. The logged in user doesn't need to have access
		to all annotation types. The copy will still work. The logged in user must
		have write permission on this annotation set.
		
		@param item The item to copy annotation from
		@param overwrite TRUE to overwrite existing values
		@since 2.5
	*/
	public void copyFrom(Annotatable item, boolean overwrite)
	{
		checkPermission(Permission.WRITE);
		if (!item.isAnnotated()) return;
		AnnotationSet fromAs = item.getAnnotationSet();
		Item toType = this.getItemType();
		DbControl dc = getDbControl();
		
		Map<AnnotationTypeData, AnnotationData> toAnnotations = this.getData().getAnnotations();
		Map<AnnotationTypeData, AnnotationData> fromAnnotations = fromAs.getData().getAnnotations();
		
		for (Map.Entry<AnnotationTypeData, AnnotationData> entry : fromAnnotations.entrySet())
		{
			AnnotationTypeData at = entry.getKey();
			// Check if the target item can hold annotation of this type
			if (at.getItemTypes().contains(toType.getValue()))
			{
				AnnotationData to = null;
				if (toAnnotations.containsKey(at))
				{					
					// Use existing entry unless overwrite == false
					if (!overwrite)
					{
						/* #### CONTINUE-STATEMENT #### */
						continue;	
					}
					to = toAnnotations.get(at);
				}
				else
				{
					// Create new annotation
					Type valueType = Type.fromValue(at.getValueType());
					to = new AnnotationData();
					to.setAnnotationSet(this.getData());
					to.setAnnotationType(at);
					to.setValues(valueType.newParameterValueData());
					toAnnotations.put(at, to);
					Annotation newAnnotation = new Annotation(to, this);
					dc.saveItemIf(this, newAnnotation, false);
				}
				// Copy the values
				AnnotationData from = entry.getValue();
				to.getValues().replaceValues(from.getValues().getValues());
				to.setLastUpdate(new Date());
				setSnapshotInvalid();
			}
		}
	}
	
	
	/**
		Get a query that returns all annotation sets inheriting from this annotation set.
		@return An <code>ItemQuery</code> object
	*/
	public ItemQuery<AnnotationSet> getInheritingAnnotationSets()
	{
		ItemQuery<AnnotationSet> query = AnnotationSet.getQuery();
		query.joinPermanent(Hql.innerJoin("inheritedSets", "i"));
		query.restrictPermanent(
			Restrictions.eq(
				Hql.alias("i"),
				Hql.entity(this)
			)
		);
		return query;
	}
	
	/**
		Mark the snapshot (if it exists) of this annotation set as
		invalid. The snapshot will be removed when the transaction
		is committed. Once the snapshot has been marked as invalid
		it can't be unmarked.		
		@since 2.14
	*/
	public void setSnapshotInvalid()
	{
		this.snapshotIsInvalid = true;
	}
	
	/**
		Get the status of the 'invalid snapshot' flag.
		@return TRUE if the snapshot has been marked as invalid, 
			FALSE if not
		@since 2.14
	*/
	public boolean isSnapshotInvalid()
	{
		return snapshotIsInvalid;
	}
	
	private static class QueryRuntimeFilterImpl
		implements QueryRuntimeFilter
	{
		public void enableFilters(QueryRuntimeFilterManager manager, EntityQuery query, DbControl dc)
		{
			// Do not add any filters.
		}
	}
}
