/*
	$Id: ExtendedProperties.java 5431 2010-09-29 06:53:13Z nicklas $

	Copyright (C) 2005 Samuel Andersson, Nicklas Nordborg
	Copyright (C) 2006 Johan Enell, Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.dbengine.DbEngine;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.XMLUtil;

import java.util.HashSet;
import java.util.List;
import java.util.ArrayList;
import java.util.Map;
import java.util.HashMap;
import java.util.Set;
import java.util.regex.PatternSyntaxException;
import java.net.URL;

import org.jdom.Document;
import org.jdom.Element;

/**
	This class is used for reading XML files with information
	about extended properties. All files returned by {@link Application#getExtendedPropertyFiles()}
	are parsed.

	@see net.sf.basedb.core.data.ExtendableData
	@see ExtendedProperty
	@author Nicklas
	@version 2.0
*/
public class ExtendedProperties
{

	/**
		Log core events.
	*/
	private static final org.apache.log4j.Logger log = 
		org.apache.log4j.LogManager.getLogger("net.sf.basedb.core.ExtendedProperties");


	/**
		A list with the <code>name</code> attribute of all classes found in
		the xml file.
	*/
	private static List<String> classes = null;

	/**
		A map from class name --> a list of extended properties of for
		that class. The list holds {@link ExtendedProperty} objects.
	*/
	private static Map<String, List<ExtendedProperty>> allProperties = null;

	/**
		The DTD which is used to validate the XML file.
	*/
	private static final URL dtdFile = ExtendedProperties.class.getResource("/net/sf/basedb/core/dtd/extended-properties.dtd");

	private static boolean isInitialised = false;

	/**
		Initialize the class.
		@throws BaseException This exception is thrown if there is an error
	*/
	static synchronized void init()
		throws BaseException
	{
		if (isInitialised) return;
		classes = new ArrayList<String>();
		allProperties = new HashMap<String, List<ExtendedProperty>>();
		loadExtendedPropertiesFile();
		isInitialised = true;
	}

	/**
		Unload all settings.
	*/
	static synchronized void unload()
	{
		if (classes != null) classes.clear();
		classes = null;
		if (allProperties != null) allProperties.clear();
		allProperties = null;
		isInitialised = false;
	}
	
	/**
		Load and parse the file with predefined queries.
		This method will populate the {@link #classes} and
		{@link #allProperties} variables.
	*/
	private static synchronized void loadExtendedPropertiesFile()
		throws BaseException
	{
		List<String> files = Application.getExtendedPropertyFiles();
		try
		{
			Set<String> usedNames = new HashSet<String>();
			for (String xmlFile : files)
			{
				log.debug("Loading extended properties from file: " + xmlFile);
				Document dom = XMLUtil.getValidatedXml(ExtendedProperties.class.getResource(xmlFile), dtdFile);
				loadClasses(usedNames, dom, xmlFile);
				log.info("Loaded extended properties from file: " + xmlFile);
			}
		}
		catch (Exception ex)
		{
			throw new BaseException(ex);
		}
	}

	/**
		Get a list with the name of all classes. The names do not
		include a package name.
		@return A <code>List</code> containing the names of the classes.
	*/
	public static List<String> getClasses()
	{
		return classes;
	}

	/**
		Check if extended properties has been defined for the
		specified class.
		@param className The name of the class with or without
			the package name
		@return TRUE if the class is extendable, FALSE otherwise
	*/
	public static boolean isExtendable(String className)
	{
		int index = className.lastIndexOf(".");
		if (index >= 0)
		{
			className = className.substring(index+1);
		}
		return allProperties.containsKey(className);
	}

	/**
		Get a list of extended properties for the specified class.
		@param className The name of the class with or without
			the package name
		@return A <code>List</code> object containing {@link ExtendedProperty}
			objects, or null if the specified class is not extendable or
			no extended properties has been defined
	*/
	public static List<ExtendedProperty> getProperties(String className)
	{
		int index = className.lastIndexOf(".");
		if (index >= 0)
		{
			className = className.substring(index+1);
		}
		List<ExtendedProperty> l = (List<ExtendedProperty>)allProperties.get(className);
		return l;
	}

	/**
		Find an extended property in a given class.
		@param className The name of an extendable class with or without
			the package name 
		@param property The name of the property
		@return The extended property or null if not found
		@since 2.15
	*/
	public static ExtendedProperty getProperty(String className, String property)
	{
		List<ExtendedProperty> l = getProperties(className);
		if (l == null) return null;
		for (ExtendedProperty e : l)
		{
			if (e.getName().equals(property)) return e;
		}
		return null;
	}
	
	/**
		Load the classes and put their names in the {@link #classes}
		variable.
	*/
	@SuppressWarnings("unchecked")
	private static void loadClasses(Set<String> usedNames, Document dom, String xmlFile)
	{
		List<Element> classTags = (List<Element>)dom.getRootElement().getChildren("class");
		for (Element el : classTags)
		{
			String className = el.getAttributeValue("name");
			if (!classes.contains(className)) 
			{
				classes.add(className);
				allProperties.put(className, new ArrayList<ExtendedProperty>());
			}
			loadProperties(usedNames, el, xmlFile);
		}
	}

	/**
		Load the properties for the specified class node and
		put the list in the {@link #allProperties} variable.
	*/
	@SuppressWarnings("unchecked")
	private static void loadProperties(Set<String> usedNames, Element classElement, String xmlFile)
	{
		String className = classElement.getAttributeValue("name");
		List<ExtendedProperty> properties = allProperties.get(className);
		List<Element> children = (List<Element>)classElement.getChildren("property");
		DbEngine engine = HibernateUtil.getDbEngine();
		for (Element property : children)
		{
			String name = Values.getStringOrNull(property.getAttributeValue("name"));
			String idName = className + ":name:" + name;
			String title = Values.getStringOrNull(property.getAttributeValue("title"));
			if (title == null) title = name;
			String column = Values.getStringOrNull(property.getAttributeValue("column"));
			String idColumn = className + ":column:" + column;
			if (!ExtendedProperty.isValidName(name))
			{
				throw new InvalidDataException("Invalid property for class '" + 
					className + "' in file '" + xmlFile + "': name=" + name);
			}
			if (usedNames.contains(idName))
			{
				throw new InvalidDataException("Duplicate property for class '" + 
					className + "' in file '" + xmlFile + "': name=" + name);
			}
			usedNames.add(idName);
			if (!engine.isValidColumnName(column))
			{
				throw new InvalidDataException("Invalid column for property '" + 
					className + "[" + name + "]' in file '" + xmlFile + "': column=" + column);
			}
			if (usedNames.contains(idColumn))
			{
				throw new InvalidDataException("Duplicate column for property '" + 
					className + "[" + name + "]' in file '" + xmlFile + "': column=" + column);
			}
			usedNames.add(idColumn);
			String description = Values.getStringOrNull(property.getAttributeValue("description"));
			int length = XMLUtil.getIntAttribute(property, "length", 255);
			Type type = Type.fromValue(property.getAttributeValue("type"));
			if (type == Type.STRING && length > 255) type = Type.TEXT;
			boolean nullable = XMLUtil.getBooleanAttribute(property, "null", true);
			boolean insertable = XMLUtil.getBooleanAttribute(property, "insert", true);
			boolean updateable = XMLUtil.getBooleanAttribute(property, "update", true);
			boolean averagable = XMLUtil.getBooleanAttribute(property, "averagable", type.isNumerical());
			String avgMethod = property.getAttributeValue("averagemethod");
			Formula.AverageMethod averageMethod = null;
			if (avgMethod != null) 
			{
				averageMethod = Formula.AverageMethod.valueOf(avgMethod.toUpperCase());
			}
			else if (averagable) 
			{
				averageMethod = Formula.AverageMethod.ARITHMETIC_MEAN;
			}
			else
			{
				averageMethod = Formula.AverageMethod.NONE;				
			}
			
			List<Element> links = (List<Element>)property.getChildren("link");
			List<ExtendedPropertyLinker> epLinks = null;
			if (links != null && links.size() > 0)
			{
				epLinks = new ArrayList<ExtendedPropertyLinker>(links.size());
				for (Element link : links)
				{
					String regexp = Values.getStringOrNull(link.getAttributeValue("regexp"));
					String url = Values.getStringOrNull(link.getAttributeValue("url"));
					if (url == null)
					{
						throw new InvalidDataException("Missing url for property link '" + 
							className + "[" + name + "]' in file '" + xmlFile + "': regexp=" + regexp);
					}
					try
					{
						epLinks.add(new ExtendedPropertyLinker(regexp, url));
					}
					catch (PatternSyntaxException ex)
					{
						throw new InvalidDataException("Invalid regexp for property link '" + 
							className + "[" + name + "]' in file '" + xmlFile + "': regexp=" + regexp, ex);
					}
				}
			}
			properties.add(new ExtendedProperty(name, title, description, column, type, length, nullable, insertable, updateable, averageMethod, epLinks));
		}
	}
}
