/*
	$Id$

	Copyright (C) 2005 Johan Enell, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

/**
 	Represents the coordinate of a feature on an array design. One coordinate
 	is equal to another if the row and column match and if either the block number
 	(if it is non-zero) or the meta row and meta column (if the block is zero) match.

	@base.modified $Date$
	@author Enell, Nicklas
	@version 2.0
*/
public class FeatureCoordinate
{
	
	private final int block;
	
	private final int metaRow;
	
	private final int metaColumn;
	
	/**
		The row of this feature.
	*/
	private final int row;
	/**
		The column of this feature.
	*/
	private final int column;
	
	/**
		Create new object with the specified row and column.
	 	@param block Block number the feature belongs to. 
	 	@param metaRow Meta row number
	 	@param metaColumn Meta column number
	 	@param row Row number of the feature
	 	@param column Column number of the feature.
	*/
	public FeatureCoordinate(Integer block, Integer metaRow, Integer metaColumn, Integer row, Integer column)
	{
		this.block = block == null ? 0 : block.intValue();
		this.metaRow = metaRow == null ? 0 : metaRow.intValue();
		this.metaColumn = metaColumn == null ? 0 : metaColumn.intValue();
		this.row = row == null ? 0 : row.intValue();
		this.column = column == null ? 0 : column.intValue();
	}

	/*
		From the Object class
		-------------------------------------------
	*/
	/**
		Check if this object is equal to another <code>FeatureCoordinate</code>
		object. For non-zero block number they are equal if the block, row and column
		match, otherwise they are equal if the metarow, metacolumn, row and column
		match.
	*/
	public boolean equals(Object o)
	{
		if ((o == null) || (getClass() != o.getClass())) return false;
		FeatureCoordinate fc = (FeatureCoordinate)o;
		return (this.row == fc.row) && (this.column == fc.column) && this.block == fc.block && 
			(this.block != 0 || (this.metaRow == fc.metaRow && this.metaColumn == fc.metaColumn));
	}
	
	/**
		Calculate the hash code for the object.
	*/
	public int hashCode()
	{
		return 89 * row + 53 * column + (block != 0 ? 37 * block : 17 * metaRow + 7 * metaColumn);
	}
	
	/**
		Get the coordinate as [block, row, column] or
		[metarow, metacolumn, row, column]
	*/
	public String toString()
	{
		if (block == 0)
		{
			return "[" + metaRow + ", " + metaColumn + ", " + row + ", " + column + "]";
		}
		else
		{
			return "[" + block + ", " + row + ", " + column + "]";
		}
	}
	// -------------------------------------------
	
	public int getBlock()
	{
		return block;
	}
	
	public int getMetaRow()
	{
		return metaRow;
	}
	
	public int getMetaColumn()
	{
		return metaColumn;
	}
	
	public int getColumn()
	{
		return column;
	}

	public int getRow()
	{
		return row;
	}

}
