/*
  $Id: HardwareType.java 5338 2010-05-07 10:30:10Z nicklas $

	Copyright (C) 2005 Johan Enell, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

  BASE is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 3
  of the License, or (at your option) any later version.

  BASE is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.Date;
import java.util.Set;

import net.sf.basedb.core.data.HardwareTypeData;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;

/**
	This class is used to represent the type of
	{@link Hardware Hardware} items in BASE. It is 
	not possible for client applications to create
	new hardware types or modify existing ones.

	@author enell
	@version 2.0
	@base.modified $Date: 2010-05-07 12:30:10 +0200 (Fri, 07 May 2010) $
	@see Hardware Hardware
*/
public class HardwareType
	extends BasicItem<HardwareTypeData>
	implements Nameable, SystemItem, Removable, Registered
{

	/**
		The type of item represented by this class.
		@see Item#FILETYPE ITEM.FILETYPE
		@see #getType() getType()
	*/
	public static final Item TYPE = Item.HARDWARETYPE;
	
	/**
		The id for the <code>HardwareType</code> object representing a
		scanner. 
	*/
	public static final String SCANNER = "net.sf.basedb.core.HardwareType.SCANNER";

	/**
		The id for the <code>HardwareType</code> object representing a
		print robot.
	*/
	public static final String PRINT_ROBOT = "net.sf.basedb.core.HardwareType.PRINT_ROBOT";
	
	/**
	 	The id for the <code>HardwareType</code> object representing a 
	 	hybridization station.
	*/
	public static final String HYBRIDIZATION_STATION = "net.sf.basedb.core.HardwareType.HYBRIDIZATION_STATION";
	
	/**
	 	The id for the <code>HardwareType</code> object representing a 
	 	freezer.
	 */
	public static final String FREEZER = "net.sf.basedb.core.HardwareType.FREEZER";

	/**
	 	The id for the <code>HardwareType</code> object representing something
	 	else.
	 	@since 2.16
	 */
	public static final String OTHER = "net.sf.basedb.core.HardwareType.OTHER";

	
	/**
		Create a new <code>HardwareType</code> item.
	
		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access
		@return The new <code>HardwareType</code> item
		@throws BaseException If there is an error
	*/
	public static HardwareType getNew(DbControl dc)
		throws BaseException
	{
		HardwareType hwt = dc.newItem(HardwareType.class);
		hwt.setName("New hardware type");
		return hwt;
	}
	
	/**
		Get a <code>HardwareType</code> object when you know the ID.

		@param dc The {@link DbControl} which will be used for
			permission checking and database access.
		@param id The system ID of the item to load.
		@return The <code>HardwareType</code> item
		@throws ItemNotFoundException This exception is thrown if an
			item with the specified ID is not found
		@throws PermissionDeniedException This exception is thrown if
			the logged in user doesn't have {@link Permission#READ READ}
			permission to the item
		@throws BaseException This exception is thrown if there is another
			error
	*/
	public static HardwareType getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		HardwareType ht = dc.loadItem(HardwareType.class, id);
		if (ht == null) throw new ItemNotFoundException("HardwareType[ID="+id+"]");
		return ht;
	}
	
	/**
		Get a query configured to retrieve hardware types.
		
		@return An {@link ItemQuery} object
	*/
	public static ItemQuery<HardwareType> getQuery()
	{
		return new ItemQuery<HardwareType>(HardwareType.class);
	}
	
	/**
		Creates a hardwaretype item.
		
		@param data the data
	*/
	HardwareType(HardwareTypeData data)
	{
		super(data);
	}
	/*
		From the Registered interface
		-------------------------------------------
	*/
	public Date getEntryDate()
	{
		return DateUtil.copy(getData().getEntryDate());
	}
	// -------------------------------------------

	/*
		From the BasicItem class
		-------------------------------------------
	*/
	/**
		Check if:
		<ul>
		<li>A Hardware of this type exists
		</ul>
		
		@return TRUE if the hardwaretype is used.
		@throws BaseException If there is an error.
	*/
	@Override
	public boolean isUsed()
		throws BaseException
	{
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"GET_HARDWARE_FOR_TYPE", "count(*)");
		/*
			SELECT {1}
			FROM Hardware hw
			WHERE hw.hardwareType = :hardwaretype
		*/
		query.setEntity("hardwaretype", this.getData());
		return HibernateUtil.loadData(Long.class, query) > 0 || super.isUsed();
	}
	/**
		Get all:
		<ul>
		<li>{@link Hardware}:s if this type
		<ul>
		@since 2.2
	*/
	@Override
	public Set<ItemProxy> getUsingItems()
	{
		Set<ItemProxy> using = super.getUsingItems();
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
				"GET_HARDWARE_FOR_TYPE", "hw.id");
		/*
			SELECT {1}
			FROM Hardware hw
			WHERE hw.hardwareType = :hardwaretype
		*/
		query.setEntity("hardwaretype", this.getData());
		addUsingItems(using, Item.HARDWARE, query);
		return using;
	}

	/**
		If this is a system hardware type, delete and create permissions are revoked.
	*/
	void initPermissions(int granted, int denied)
		throws BaseException
	{
		if (isSystemItem()) 
		{
			denied |= Permission.deny(Permission.DELETE, Permission.CREATE);
		}
		super.initPermissions(granted, denied);
	}
	// -------------------------------------------
	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	public Item getType()
	{
		return TYPE;
	}
	//  -------------------------------------------
	/*
		From the Removable interface
		-------------------------------------------
	*/
	public boolean isRemoved()
	{
		return getData().isRemoved();
	}
	public void setRemoved(boolean removed)
		throws PermissionDeniedException
	{
		checkPermission(removed ? Permission.DELETE : Permission.WRITE);
		getData().setRemoved(removed);
	}
	// -------------------------------------------
	/*
		From the SystemItem interface
		-------------------------------------------
	*/
	public String getSystemId()
	{
		return getData().getSystemId();
	}

	public boolean isSystemItem()
	{
		return getSystemId() != null;
	}
	// -------------------------------------------
	/*
		From the Nameable interface
		-------------------------------------------
	*/
	public String getName()
	{
		return getData().getName();
	}

	public void setName(String name)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setName(getData(), name);
	}

	public String getDescription()
	{
		return getData().getDescription();
	}

	public void setDescription(String description)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setDescription(getData(), description);
	}
	// -------------------------------------------

	/**
		Create a new {@link Hardware} item of this type.
	 
		@return The new <code>Hardware</code> item
		@throws DetachedItemException If this HardwareType is detached from DbControl.
		@throws PermissionDeniedException This exception is thrown if
			the logged in user doesn't have {@link Permission#CREATE CREATE}
			permission for hardware items or {@link Permission#USE USE}
			permission for this hardware type
		@throws BaseException This exception is thrown if there is another
			error
	*/
	public Hardware newHardware() 
		throws PermissionDeniedException, BaseException
	{
		checkPermission(Permission.USE);
		Hardware hw = Hardware.getNew(getDbControl(), this);
		return hw;
	}
	
	/**
		Get a query that returns all hardware of this type
		@return An {@link ItemQuery} object
	*/
	public ItemQuery<Hardware> getHardware()
	{
		ItemQuery<Hardware> query = Hardware.getQuery();
		query.restrictPermanent(
			Restrictions.eq(
				Hql.property("hardwareType"), 
				Hql.entity(this)
			)
		);
		return query;
	}
	
}
