/*
	$Id: Image.java 4889 2009-04-06 12:52:39Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.ImageData;
import net.sf.basedb.core.data.SharedData;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.query.Hql;

import java.util.Map;
import java.util.HashMap;

/**
	An image represents a single image produced by a {@link Scan}.
	The images are usually analysed by some software to produce
	a {@link RawBioAssay}.

	@author Nicklas
	@base.modified $Date: 2009-04-06 14:52:39 +0200 (Mon, 06 Apr 2009) $
	@version 2.0
*/
public class Image
	extends ChildItem<ImageData>
	implements Nameable, Removable, FileAttachable
{
	/**
		The type of item represented by this class.
		@see Item#IMAGE
		@see #getType()
	*/
	public static final Item TYPE = Item.IMAGE;

	/**
		This filter will only return items if the logged in user has
		generic read permission to scans.
	*/
	private static final QueryRuntimeFilter RUNTIME_FILTER = 
		new QueryRuntimeFilterFactory.ChildFilter(TYPE, Item.SCAN);

	/**
		Create a new <code>Image</code> item.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access
		@param scan The <code>Scan</code> which the image is related to
		@return The new <code>Image</code> item
		@throws BaseException If there is an error
		@see Scan#newImage()
	*/
	public static Image getNew(DbControl dc, Scan scan)
		throws BaseException
	{
		Image i = dc.newItem(Image.class);
		i.setScan(scan);
		i.setName("New image");
		return i;
	}

	/**
		Get an <code>Image</code> item when you know the id.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The id of the item to load
		@return The <code>Image</code> item
		@throws ItemNotFoundException If an item with the specified 
			id is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have read permission to the item
		@throws BaseException If there is another error
	*/
	public static Image getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		Image i = dc.loadItem(Image.class, id);
		if (i == null) throw new ItemNotFoundException("Image[id="+id+"]");
		return i;
	}

	/**
		Get a query configured to retrieve 
		images for the specified scan.
		@param scan The scan to retreive images for, null is allowed if
			the logged in user has generic READ permission for scans in which case
			all images will be returned
		@return An {@link ItemQuery} object
	 	@throws BaseException If getting the query fails. 
		@see Scan#getImages()
	*/
	public static ItemQuery<Image> getQuery(Scan scan)
		throws BaseException
	{
		ItemQuery<Image> query = null;
		if (scan != null)
		{
			query = new ItemQuery<Image>(Image.class, null);
			query.restrictPermanent(
				Restrictions.eq(
					Hql.property("scan"), 
					Hql.entity(scan)
				)
			);
		}
		else
		{
			query = new ItemQuery<Image>(Image.class, RUNTIME_FILTER);
		}
		return query;
	}

	Image(ImageData data)
	{
		super(data);
	}

	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/*
		From the Nameable interface
		-------------------------------------------
	*/
	public String getName()
	{
		return getData().getName();
	}
	public void setName(String name)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setName(getData(), name);
	}
	public String getDescription()
	{
		return getData().getDescription();
	}
	public void setDescription(String description)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setDescription(getData(), description);
	}
	// -------------------------------------------
	/*
		From the Removable interface
		-------------------------------------------
	*/
	public boolean isRemoved()
	{
		return getData().isRemoved();
	}
	public void setRemoved(boolean removed)
		throws PermissionDeniedException
	{
		checkPermission(removed ? Permission.DELETE : Permission.WRITE);
		getData().setRemoved(removed);
	}
	// -------------------------------------------
	/*
		From the FileAttachable interface
		-------------------------------------------
	*/
	public File getFile()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(File.class, getData().getFile());
	}

	public void setFile(File file)
		throws PermissionDeniedException
	{
		checkPermission(Permission.WRITE);
		if (file != null) file.checkPermission(Permission.USE);
		getData().setFile(file == null ? null : file.getData());
	}
	// -------------------------------------------
	/*
		From the ChildItem class
		-------------------------------------------
	*/
	SharedData getSharedParent()
	{
		return getData().getScan();
	}
	// -------------------------------------------

	/**
		Get the scan this image belongs to.
		@return The <code>Scan</code> item
		@throws PermissionDeniedException If the logged in user doesn't have 
			{@link Permission#READ} permission
		@throws BaseException If there is another error
	*/
	public Scan getScan()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(Scan.class, getData().getScan());
	}
	
	/**
		Set the scan this image belongs to. Can only be set 
		on a new item.
		@throws InvalidDataException If the scan is null
	*/
	void setScan(Scan scan)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		if (scan == null) throw new InvalidUseOfNullException("scan");
		getData().setScan(scan.getData());
	}
	
	/**
		A boolean value indicating if this image is a preview image
		or the full-sized image.
		@return TRUE if the image is a preview image, FALSE otherwise
	*/
	public boolean isPreview()
	{
		return getData().isPreview();
	}
	
	/**
		Set a flag indicating if this image is a preview image
		or the full-sized image.
		@param preview TRUE if the image is a preview image, FALSE otherwise
		@throws PermissionDeniedException If the logged in user doesn't have 
			write permission
	*/
	public void setPreview(boolean preview)
		throws PermissionDeniedException
	{
		checkPermission(Permission.WRITE);
		getData().setPreview(preview);
	}

	/**
		Get the format of this image.
		@return A <code>Format</code> object
	*/
	public Format getFormat()
	{
		return Format.fromValue(getData().getFormat());
	}
	
	/**
		Set the format of the image.
		@param format The format
		@throws PermissionDeniedException If the logged in user doesn't have 
			write permission
	*/
	public void setFormat(Format format)
		throws PermissionDeniedException
	{
		checkPermission(Permission.WRITE);
		getData().setFormat(format == null ? Format.UNKNOWN.getValue() : format.getValue());
	}

	
	/**
		This enumeration is used to indicate the type of image.
		
		@author Nicklas
	*/
	public enum Format
	{
		/**
			The image is in an unknown format.
		*/
		UNKNOWN(0, "Unkown"),
		/**
			The image is a TIFF image.
		*/
		TIFF(1, "TIFF"),
		/**
			The image is a JPEG image.
		*/
		JPEG(2, "JPEG");

		/**
			Maps the integer that is stored in the database with an fileaction.
		*/
		private static final Map<Integer, Image.Format> valueMapping = new HashMap<Integer, Image.Format>();
		
		static
		{
			for(Format fmt : Format.values())
			{
				Format f = valueMapping.put(fmt.getValue(), fmt);
				assert f == null : "Another Image.Format with the value "+fmt.getValue()+" already exists";
			}
		}
		
		/**
			The integer value of this format. 
		*/
		private final int value;
		/**
			Returned by the toString method.
		*/
		private final String displayValue;

		private Format(int value, String displayValue)
		{
			this.value = value;
			this.displayValue = displayValue;
		}
		
		public String toString()
		{
			return displayValue;
		}
		
		/**
			Get the integer value that is used when storing an Image.Format to the database.
			@return The intger value of this action. 
		*/
		public int getValue()
		{
			return value;
		}
		/**
			Get the <code>Image.Format</code> object when you know the integer code.
			
			@param value The integer value.
			@return The <code>Format</code> for the integer value.
		*/
		public static Format fromValue(int value)
		{
			Format fmt = valueMapping.get(value);
			assert fmt != null : "format == null for value "+value;
			return fmt;
		}
		

	}
	

}

