/*
	$Id: PlateGeometry.java 5339 2010-05-07 11:31:02Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.Date;
import java.util.Set;

import net.sf.basedb.core.data.PlateGeometryData;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.query.Hql;

/**
	This class defines a plate geometry, which is the number of rows
	and columns of wells on a {@link Plate}. The number of rows and columns
	must be specified when the item is created, and the values can't
	be changed after that.
	<p>
	A {@link PlateMapping} is a way to specify how one geometry can be
	transformed to another. It holds information for how to map all source
	wells to destination wells. For example it can
	map four 96-well plates to one 384-well plate.

	@author Nicklas
	@version 2.0
	@see Plate
	@see PlateMapping
	@base.modified $Date: 2010-05-07 13:31:02 +0200 (Fri, 07 May 2010) $
*/
public class PlateGeometry
	extends BasicItem<PlateGeometryData>
	implements Nameable, Removable, Registered
{
	/**
		The type of item represented by this class.
		@see Item#PLATEGEOMETRY
		@see #getType()
	*/
	public static final Item TYPE = Item.PLATEGEOMETRY;
	
	/**
		Create a new <code>PlateGeometry</code> item.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access
		@param rows The number of rows
		@param columns The number of columns
		@return The new <code>PlateGeometry</code> item
		@throws InvalidDataException If the number of rows or columns is
			less than or equal to zero
		@throws BaseException If there is an error
	*/
	public static PlateGeometry getNew(DbControl dc, int rows, int columns)
		throws InvalidDataException, BaseException
	{
		PlateGeometry pg = dc.newItem(PlateGeometry.class);
		pg.setName("New plate geometry");
		pg.setRows(rows);
		pg.setColumns(columns);
		pg.getData().setEntryDate(new Date());
		return pg;
	}

	/**
		Get a <code>PlateGeometry</code> item when you know the id.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The id of the item to load
		@return The <code>PlateGeometry</code> item
		@throws ItemNotFoundException If an item with the specified 
			id is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have {@link Permission#READ} permission to the item
		@throws BaseException If there is another error
	*/
	public static PlateGeometry getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		PlateGeometry pg = dc.loadItem(PlateGeometry.class, id);
		if (pg == null) throw new ItemNotFoundException("PlateGeometry[id="+id+"]");
		return pg;
	}

	/**
		Get a query configured to retrieve plate geometries.
		@return An {@link ItemQuery} object
	*/
	public static ItemQuery<PlateGeometry> getQuery()
	{
		return new ItemQuery<PlateGeometry>(PlateGeometry.class);
	}

	PlateGeometry(PlateGeometryData plateGeometryData)
	{
		super(plateGeometryData);
	}

	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/*
		From the Nameable interface
		-------------------------------------------
	*/
	public String getName()
	{
		return getData().getName();
	}
	public void setName(String name)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setName(getData(), name);
	}
	public String getDescription()
	{
		return getData().getDescription();
	}
	public void setDescription(String description)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setDescription(getData(), description);
	}
	// -------------------------------------------
	/*
		From the Removable interface
		-------------------------------------------
	*/
	public boolean isRemoved()
	{
		return getData().isRemoved();
	}
	public void setRemoved(boolean removed)
		throws PermissionDeniedException
	{
		checkPermission(removed ? Permission.DELETE : Permission.WRITE);
		getData().setRemoved(removed);
	}
	// -------------------------------------------
	/*
		From the Registered interface
		-------------------------------------------
	*/
	public Date getEntryDate()
	{
		return DateUtil.copy(getData().getEntryDate());
	}
	// -------------------------------------------

	/*
		From the BasicItem class
		-------------------------------------------
	*/
	/**
		Check if:
		<ul>
		<li>A {@link PlateType} with this geometry exists
		<li>A {@link PlateMapping} to/from this geometry exists
		<li>A {@link BioPlate} that uses this geometry exists
		</ul>
	*/
	public boolean isUsed()
		throws BaseException
	{
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"GET_PLATETYPES_FOR_PLATEGEOMETRY", "count(*)");
		/*
			SELECT {1}
			FROM PlateTypeData pt
			WHERE pt.plateGeometry = :plateGeometry
		*/
		query.setEntity("plateGeometry", this.getData());
		boolean used = HibernateUtil.loadData(Long.class, query) > 0;
		if (!used)
		{
			query = HibernateUtil.getPredefinedQuery(session, 
				"GET_PLATEMAPPINGS_FOR_PLATEGEOMETRY", "count(*)");
			/*
				SELECT {1}
				FROM PlateMappingData pm
				WHERE pm.sourceGeometry = :plateGeometry
				OR pm.destinationGeometry = :plateGeometry
			*/
			query.setEntity("plateGeometry", this.getData());
			used = HibernateUtil.loadData(Long.class, query) > 0;
		}
		if (!used)
		{
			query = HibernateUtil.getPredefinedQuery(session, 
				"GET_BIOPLATES_FOR_PLATEGEOMETRY", "count(*)");
			/*
				SELECT {1}
				FROM BioPlateData bp
				WHERE bp.plateGeometry = :plateGeometry
			*/
			query.setEntity("plateGeometry", this.getData());
			used = HibernateUtil.loadData(Long.class, query) > 0;
		}
		return used || super.isUsed();
	}
	/**
		Get all:
		<ul>
		<li>{@link PlateType}:s of this geometry
		<li>{@link PlateMapping}:s using this geometry 
		<li>{@link BioPlate}:s using this geometry
		<ul>
		@since 2.2
	*/
	@Override
	public Set<ItemProxy> getUsingItems()
	{
		Set<ItemProxy> using = super.getUsingItems();
		org.hibernate.Session session = getDbControl().getHibernateSession();
		
		// Plate types
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"GET_PLATETYPES_FOR_PLATEGEOMETRY", "pt.id");
			/*
				SELECT {1}
				FROM PlateTypeData pt
				WHERE pt.plateGeometry = :plateGeometry
			*/
		query.setEntity("plateGeometry", this.getData());
		addUsingItems(using, Item.PLATETYPE, query);

		// Plate mappings
		query = HibernateUtil.getPredefinedQuery(session, 
			"GET_PLATEMAPPINGS_FOR_PLATEGEOMETRY", "pm.id");
			/*
				SELECT {1}
				FROM PlateMappingData pm
				WHERE pm.sourceGeometry = :plateGeometry
				OR pm.destinationGeometry = :plateGeometry
			*/
		query.setEntity("plateGeometry", this.getData());
		addUsingItems(using, Item.PLATEMAPPING, query);

		// Bioplates
		query = HibernateUtil.getPredefinedQuery(session, 
			"GET_BIOPLATES_FOR_PLATEGEOMETRY", "bp.id");
			/*
				SELECT {1}
				FROM BioPlateData bp
				WHERE bp.plateGeometry = :plateGeometry
			*/
		query.setEntity("plateGeometry", this.getData());
		addUsingItems(using, Item.BIOPLATE, query);
		return using;
	}
// -------------------------------------------

	/**
		Get the number of rows in this geometry.
		@return The number of rows
	*/
	public int getRows()
	{
		return getData().getRows();
	}
	/**
		The number of rows cannot be changed, used at object creation only.
		@throws InvalidDataException If the row count is less than 1
	*/
	private void setRows(int rows)
		throws InvalidDataException
	{
		if (rows < 1) throw new NumberOutOfRangeException("rows", rows, 1, false);
		getData().setRows(rows);
	}

	/**
		Get the number of columns in this geometry.
		@return The number of columns
	*/
	public int getColumns()
	{
		return getData().getColumns();
	}
	/**
		The number of columns cannot be changed, used at object creation only.
		@throws InvalidDataException If the column count is less than 1
	*/
	private void setColumns(int columns)
		throws InvalidDataException
	{
		if (columns < 1) throw new NumberOutOfRangeException("columns", columns, 1, false);
		getData().setColumns(columns);
	}
	
	/**
		Create a new {@link PlateType} of this geometry.
		@return The new plate type
		@throws PermissionDeniedException If the logged in user
			doesn't have use permission for this plate geometry
		@throws BaseException If there is another error
	*/
	public PlateType newPlateType()
		throws PermissionDeniedException, BaseException
	{
		PlateType pt = PlateType.getNew(getDbControl(), this);
		return pt;
	}
	
	/**
		Get a query returning all plate types using this geometry.
		@return An <code>ItemQuery</code>
	*/
	public ItemQuery<PlateType> getPlateTypes()
	{
		ItemQuery<PlateType> query = PlateType.getQuery();
		query.restrictPermanent(
			Restrictions.eq(
				Hql.property("plateGeometry"),
				Hql.entity(this)
			)
		);
		return query;
	}

}
