/*
	$Id: Platform.java 4517 2008-09-11 20:11:02Z jari $

	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.Set;

import net.sf.basedb.core.data.FileTypeIndex;
import net.sf.basedb.core.data.PlatformData;
import net.sf.basedb.core.data.PlatformFileTypeData;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;

/**
	This class is used to represent an experimental platform in BASE.
	An experimental platform is used to define the types of files
	that are used as data files. For example, the {@link Platform#AFFYMETRIX}
	platform defines {@link DataFileType#AFFYMETRIX_CDF CDF} and 
	{@link DataFileType#AFFYMETRIX_CEL CEL} files.
	<p>
	A platform may have variants. A {@link PlatformVariant} has independent
	settings but may inherit file types from the main platform.
	<p>
	Some platforms are file-only platforms, and has data that can't be imported
	into the database, for example the Affymetrix platform. Platforms that
	are not file-only can also import data into the database. Such platforms
	may be locked to a single {@link RawDataType} or they may be {@link Platform#GENERIC}.

	@author Nicklas
	@version 2.5
	@base.modified $Date: 2008-09-11 22:11:02 +0200 (Thu, 11 Sep 2008) $
*/
public class Platform
	extends BasicItem<PlatformData>
	implements Nameable, Removable
{
	
	/**
		The type of item represented by this class.
		@see Item#PLATFORM
		@see #getType()
	*/
	public static final Item TYPE = Item.PLATFORM;

	/**
		The external id for the <code>Generic</code> platform. It allows
		importing raw data into the database for all raw data
		type.
		@see Platform#getByExternalId(DbControl, String)
	*/
	public static final String GENERIC = "generic";
	
	/**
		The external id for the <code>Affymetrix</code> platform. It 
		is a file-only platform.
		@see Platform#getByExternalId(DbControl, String)
	*/
	public static final String AFFYMETRIX = "affymetrix";
	
	/**
		Create a new file-only platform. Data for this platform can't be imported
		to the database.
		
		@param dc The DbControl to use for database access
		@param externalId The external ID of the platform, must be 
			unique for all platforms
		@param channels Number of channels in this platform
		@return The new platform object
	*/
	public static Platform getNew(DbControl dc, String externalId, int channels)
	{
		Platform p = dc.newItem(Platform.class);
		p.getData().setFileOnly(true);
		p.getData().setRawDataType("platform." + externalId);
		p.setExternalId(externalId);
		p.setChannels(channels);
		return p;
	}

	/**
		Create a new platform which may optionally import data into the database.
		
		@param dc The DbControl to use for database access
		@param externalId The external ID of the platform, must be 
			unique for all platforms
		@param rawDataType A specific raw data type this platform will be locked
			to or null if it is not locked to any specific raw data type
		@return The new platform object
	*/
	public static Platform getNew(DbControl dc, String externalId, RawDataType rawDataType)
	{
		Platform p = dc.newItem(Platform.class);
		p.getData().setFileOnly(false);
		p.setExternalId(externalId);
		p.setRawDataType(rawDataType);
		return p;
	}
	
	/**
		Get a <code>Platform</code> object when you know the ID.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The ID of the item to load
		@return The <code>Platform</code> item
		@throws ItemNotFoundException If an item with the specified ID is not found
		@throws PermissionDeniedException If the logged in user doesn't have 
			{@link Permission#READ} permission to the items
		@throws BaseException If there is another error
	*/
	public static Platform getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		Platform p = dc.loadItem(Platform.class, id);
		if (p == null) throw new ItemNotFoundException("Platform[id="+id+"]");
		return p;
	}
	
	/**
		Get a <code>Platform</code> item when you know the external ID.
	
		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param externalId The external ID of the item to load
		@return The <code>Platform</code> item.
		@throws ItemNotFoundException If an item with the specified ID is not found
		@throws PermissionDeniedException If the logged in user doesn't have
			read permission for the item
		@throws BaseException If there is another error
	*/
	public static Platform getByExternalId(DbControl dc, String externalId)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		if (externalId != null) externalId = externalId.trim();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(dc.getHibernateSession(), 
			"GET_PLATFORM_FOR_EXTERNAL_ID");
		/*
			SELECT plf
			FROM PlatformData plf
			WHERE plf.externalId = :externalId
		*/
		query.setString("externalId", externalId);
		Platform p = dc.getItem(Platform.class, HibernateUtil.loadData(PlatformData.class, query));
		if (p == null) throw new ItemNotFoundException("Platform[externalId="+externalId+"]");
		return p;
	}
	
	/**
		Get a query configured to retrieve <code>Platform</code> items.
		@return An {@link ItemQuery} object
	 	@throws BaseException If the query could not be created. 
	*/
	public static ItemQuery<Platform> getQuery()
		throws BaseException
	{
		return new ItemQuery<Platform>(Platform.class);
	}

	/**
		Creates a new filetype item from the given data.
	
		@param data the data.
	*/
	Platform(PlatformData data)
	{
		super(data);
	}
	
	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/*
		From the Nameable interface
		-------------------------------------------
	*/
	public String getName()
	{
		return getData().getName();
	}
	public void setName(String name)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setName(getData(), name);
	}
	public String getDescription()
	{
		return getData().getDescription();
	}
	public void setDescription(String description)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setDescription(getData(), description);
	}
	// -------------------------------------------
	/*
		From the Removable interface
		-------------------------------------------
	*/
	public boolean isRemoved()
	{
		return getData().isRemoved();
	}
	public void setRemoved(boolean removed)
		throws PermissionDeniedException
	{
		checkPermission(removed ? Permission.DELETE : Permission.WRITE);
		getData().setRemoved(removed);
	}
	// -------------------------------------------
	/*
	From the BasicItem class
	-------------------------------------------
	*/
	/**
		Check if:
		<ul>
		<li>any raw bioassays or array design are using this platform
		</ul>
	*/
	@Override
	public boolean isUsed()
		throws BaseException
	{
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"GET_RAWBIOASSAYS_FOR_PLATFORM", "count(*)");
		/*
			SELECT {1}
			FROM RawBioAssayData rba
			WHERE rba.platform = :platform
		*/
		query.setEntity("platform", this.getData());
		long count = HibernateUtil.loadData(Long.class, query);
		if (count == 0)
		{
			query = HibernateUtil.getPredefinedQuery(session, 
				"GET_ARRAYDESIGNS_FOR_PLATFORM", "count(*)");
				/*
					SELECT {1}
					FROM ArrayDesignData ad
					WHERE ad.platform = :platform
				*/
			query.setEntity("platform", this.getData());
			count = HibernateUtil.loadData(Long.class, query);			
		}
		return count > 0 || super.isUsed();
	}
	/**
		Get all:
		<ul>
		<li>{@link ArrayDesign}:s and {@link RawBioAssay}:s using
		this platform
		<ul>
	*/
	@Override
	public Set<ItemProxy> getUsingItems()
	{
		Set<ItemProxy> using = super.getUsingItems();
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"GET_RAWBIOASSAYS_FOR_PLATFORM", "rba.id");
			/*
				SELECT {1}
				FROM RawBioAssayData rba
				WHERE rba.platform = :platform
			*/
		query.setEntity("platform", this.getData());
		addUsingItems(using, Item.RAWBIOASSAY, query);
		query = HibernateUtil.getPredefinedQuery(session, 
			"GET_ARRAYDESIGNS_FOR_PLATFORM", "ad.id");
			/*
				SELECT {1}
				FROM ArrayDesignData ad
				WHERE ad.platform = :platform
			*/
		query.setEntity("platform", this.getData());
		addUsingItems(using, Item.ARRAYDESIGN, query);		
		return using;
	}
	/**
		Deny deletion of the GENERIC and Affymetrix platform. The GENERIC
		platform is required by the system and the AFFYMETRIX platform is
		required for backwards compatibility reasons.
	*/
	@Override
	protected void initPermissions(int granted, int denied)
	{
		if (GENERIC.equals(getExternalId()) || AFFYMETRIX.equals(getExternalId())) 
		{
			denied |= Permission.deny(Permission.DELETE);
		}
		super.initPermissions(granted, denied);
	}
	// -------------------------------------------
	
	/**
		The maximum length of the external ID that can be stored in the database.
		@see #setExternalId(String)
	*/
	public static final int MAX_EXTERNAL_ID_LENGTH = PlatformData.MAX_EXTERNAL_ID_LENGTH;
	
	/**
		Get the external id for the platform.
	*/
	public String getExternalId()
	{
		return getData().getExternalId();
	}
	
	private void setExternalId(String externalId)
		throws PermissionDeniedException, InvalidDataException, BaseException
	{
		getData().setExternalId(StringUtil.setNullableString(
			externalId, "externalId", MAX_EXTERNAL_ID_LENGTH));
	}
	
	/**
		Set the number of channels for a file-only platform.
	*/
	private void setChannels(int channels)
	{
		getData().setChannels(IntegerUtil.checkMin(channels, "channels", 1));
	}
	
	/**
		Set the raw data type. The raw data type must allow
		data to be imported into the database.
		@param rawDataType The raw data type, null is allowed
	*/
	private void setRawDataType(RawDataType rawDataType)
	{
		String rdt = null;
		if (rawDataType != null)
		{
			if (!rawDataType.isStoredInDb())
			{
				throw new InvalidDataException("Raw data type '" + rawDataType.getName() + 
					"' can't store data in the database");
			}
			rdt = rawDataType.getId();
			getData().setChannels(rawDataType.getChannels());
		}
		getData().setRawDataType(rdt);
	}
	
	/**
		Check if the platform is a file-only platform or if it allows data
		to be imported into the database.
		@return TRUE if only file data is allowed, FALSE if import to database
			is allowed
	*/
	public boolean isFileOnly()
	{
		return getData().isFileOnly();
	}
	
	/**
		Get the raw data type this platform is locked to. File-only platforms
		are always locked to a raw data type. Platforms that support import of
		data to the database may or may not be locked. Note! If the variant is
		a file-only variant that hasn't been committed to the database a 
		new instance of {@link RawDataType} is returned for each call to this
		method.
		@return The raw data type or null if the platform isn't locked
	*/
	public RawDataType getRawDataType()
	{
		return isInDatabase() ?
				RawDataTypes.getRawDataType(getData().getRawDataType()) :
				new RawDataType(getData());
	}
	
	/**
		Get a query that returns variants of this platform.
		@return An ItemQuery object
	*/
	public ItemQuery<PlatformVariant> getVariants()
	{
		ItemQuery<PlatformVariant> query = PlatformVariant.getQuery();
		query.restrictPermanent(
			Restrictions.eq(
				Hql.property("platform"), 
				Hql.entity(this)
			)
		);
		return query;
	}
	
	/**
		Create a new {@link PlatformVariant} of this platform. 
		@param systemId The systemId of the new variant (must be unique)
		@return The new <code>PlatformVariant</code> object
		@see PlatformVariant#getNew(DbControl, Platform, String)
	*/
	public PlatformVariant newVariant(String systemId)
	{
		return PlatformVariant.getNew(getDbControl(), this, systemId);
	}
	
	/**
		Get the file types registered with this platform.
		@see PlatformFileType#getQuery(Platform, PlatformVariant, boolean)
	*/
	public ItemQuery<PlatformFileType> getFileTypes(PlatformVariant variant, boolean restrict)
	{
		return PlatformFileType.getQuery(this, variant, restrict);
	}
	
	/**
		Add a file type to this platform/variant. If the file type has 
		already been registered this method can be used to change the
		<code>required</code> flag.
		
		@param type The file type
		@param required If a file of this type is required or not
		@param variant An optional variant
		@throws PermissionDeniedException If the logged in user doesn't
			have write permission
		@throws InvalidUseOfNullException If type is null
	*/
	public void addFileType(DataFileType type, boolean required, PlatformVariant variant)
	{
		checkPermission(Permission.WRITE);
		if (type == null) throw new InvalidUseOfNullException("type");
		FileTypeIndex index = new FileTypeIndex(variant == null ? null : variant.getData(), type.getData());
		if (!getData().getFileTypes().containsKey(index))
		{
			PlatformFileType platformFileType = 
				PlatformFileType.getNew(getDbControl(), this, variant, type, required);
			getData().getFileTypes().put(index, platformFileType.getData());
			getDbControl().saveItemIf(this, platformFileType, false);
		}
		else
		{
			PlatformFileTypeData pfData = getData().getFileTypes().get(index);
			pfData.setRequired(required);
		}
	}
	
	/**
		Remove a file type from this platform/variant. If the file
		type isn't registered this method does nothing. If a variant is given
		the file type is only removed from that variant. If no variant
		is given the file type is only removed from the platform, not
		from other variants that have registered with the file type.
		
		@param type The file type to remove
		@param variant An optional variant
		@throws PermissionDeniedException If the logged in user doesn't
			have write permission
		@throws InvalidUseOfNullException If type is null
	*/
	public void removeFileType(DataFileType type, PlatformVariant variant)
	{
		checkPermission(Permission.WRITE);
		if (type == null) throw new InvalidUseOfNullException("type");
		FileTypeIndex index = new FileTypeIndex(variant == null ? null : variant.getData(), type.getData());
		if (getData().getFileTypes().containsKey(index))
		{
			PlatformFileTypeData pfData = getData().getFileTypes().remove(index);
			if (pfData != null)
			{
				DbControl dc = getDbControl();
				PlatformFileType pf = dc.getItem(PlatformFileType.class, pfData);
				dc.deleteItem(pf);
			}
		}
	}
	
	/**
		Get information about a file type that has been registered with this
		platform/variant.
		
		@param type The file type to get information for
		@param variant An optional variant
		@return Null if the file type isn't registered with the platform/variant 
			or a {@link PlatformFileType} object
	*/
	public PlatformFileType getFileType(DataFileType type, PlatformVariant variant)
	{
		if (type == null) throw new InvalidUseOfNullException("type");
		PlatformFileType pf = null;
		FileTypeIndex index = new FileTypeIndex(variant == null ? null : variant.getData(), type.getData());
		if (getData().getFileTypes().containsKey(index))
		{
			PlatformFileTypeData pfData = getData().getFileTypes().get(index);
			if (pfData != null)
			{
				DbControl dc = getDbControl();
				pf = dc.getItem(PlatformFileType.class, pfData);
			}
		}
		return pf;
	}
	
}
