/*
	$Id: Project.java 5368 2010-06-22 12:51:41Z nicklas $

	Copyright (C) 2005 Samuel Andersson, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Martin Svensson
	Copyright (C) 2007 Johan Enell, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.Transactional.Action;
import net.sf.basedb.core.data.AnyToAnyData;
import net.sf.basedb.core.data.BasicData;
import net.sf.basedb.core.data.ProjectData;
import net.sf.basedb.core.data.ShareableData;
import net.sf.basedb.core.query.Restriction;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.EntityQuery;
import net.sf.basedb.core.query.ResultList;
import net.sf.basedb.core.query.Selects;
import net.sf.basedb.info.ProjectInfo;
import net.sf.basedb.info.ToTransferable;

import java.util.ArrayList;
import java.util.Date;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Collections;

/**
	This class is used to represent projects. A project is used to
	collect items of particular interest in one place. Users and
	groups can also be granted access to projects, allowing everyone easy
	access to shared items within the project.
	<p>
	While working with BASE it is possible to set one project as the
	active project. This enables some useful functionality in the core:
	<ul>
	<li>Every {@link Shareable} item the user creates is automatically
		shared to the project. 
	<li>All queries defaults to only return items that are shared
		to the project.
	</ul>
	<p>
	It is also possible to add an item to other
	projects, or remove it completely from all projects. Every member of
	a project has a special <code>project permission</code> which gives the highest
	permission for accessing items in the project. Normally this permission
	is {@link Permission#USE}. One person might be assigned
	administrator of the project and given higher permissions.
	<p>
	Items may also be given a permission within the project. The normal
	is to give full access, but it may also be set to for example
	{@link Permission#READ}. Then, not even the project administrator
	will get more than read permissions to the item. This feature can
	for example be used when a project needs to access data from
	a user that is not a member of the project.

	@author Nicklas
	@version 2.0
	@see Role
	@see Project
	@see SessionControl#setActiveProject(Project)
	@base.modified $Date: 2010-06-22 14:51:41 +0200 (Tue, 22 Jun 2010) $
*/
public class Project
	extends OwnedItem<ProjectData>
	implements Nameable, Removable, ToTransferable<ProjectInfo>, Registered
{
	/**
		The type of item represented by this class.
		@see Item#PROJECT
		@see #getType()
	*/
	public static final Item TYPE = Item.PROJECT;

	/**
		This filter will limit a query to only return projects
		where the logged in user is a member or owner unless the logged in user
		has generic read permission.
	*/
	private static final QueryRuntimeFilter RUNTIME_FILTER = new QueryRuntimeFilterImpl();
	
	/**
		Create a new <code>Project</code> item.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@return The new <code>Project</code> item
		@throws BaseException If there is an error
	*/
	public static Project getNew(DbControl dc)
		throws BaseException
	{
		Project p = dc.newItem(Project.class);
		p.setName("New project");
		p.getData().setEntryDate(new Date());
		p.setAutoPermission(EnumSet.of(Permission.READ, Permission.USE, Permission.WRITE, Permission.DELETE));
		return p;
	}

	/**
		Get a <code>Project</code> item when you know the ID.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The ID of the item to load
		@return The <code>Project</code> item
		@throws ItemNotFoundException If an item with the specified 
			ID is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have {@link Permission#READ} permission to the item
		@throws BaseException If there is another error
	*/
	public static Project getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		Project p = dc.loadItem(Project.class, id);
		if (p == null) throw new ItemNotFoundException("Project[id="+id+"]");
		return p;
	}

	/**
		Get a query configured to retrieve projects. If the logged in user
		doesn't have generic permission to all projects, only projects
		the user has access to are returned.

		@return An {@link ItemQuery} object
	*/
	public static ItemQuery<Project> getQuery()
	{
		return new ItemQuery<Project>(Project.class, RUNTIME_FILTER);
	}

	/**
		Get the project default entry that is used to attach item of a
		given <code>itemType</code> to <code>attachesToType</code>. Eg.
		The {@link Default#SAMPLING_PROTOCOL} is attaches a {@link Item#PROTOCOL}
		to a {@link Item#SAMPLE}.
		@param itemType The item type that is attached... (eg PROTOCOL)
		@param attachesToType ... to this item type (eg. SAMPLE)
		@return A default entry or null
		@since 2.10
	*/
	public static Default getDefaultType(Item itemType, Item attachesToType)
	{
		for (Default def : Default.values())
		{
			if (def.getItemType() == itemType && def.getAttachesToItemType() == attachesToType)
			{
				return def;
			}
		}
		return null;
	}
	
	private Map<Default, BasicData> defaults;
	
	Project(ProjectData projectData)
	{
		super(projectData);
	}

	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/*
		From the Nameable interface
		-------------------------------------------
	*/
	public String getName()
	{
		return getData().getName();
	}
	public void setName(String name)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setName(getData(), name);
	}
	public String getDescription()
	{
		return getData().getDescription();
	}
	public void setDescription(String description)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setDescription(getData(), description);
	}
	// -------------------------------------------
	/*
		From the Removable interface
		-------------------------------------------
	*/
	public boolean isRemoved()
	{
		return getData().isRemoved();
	}
	public void setRemoved(boolean removed)
		throws PermissionDeniedException
	{
		checkPermission(removed ? Permission.DELETE : Permission.WRITE);
		getData().setRemoved(removed);
	}
	// -------------------------------------------
	/*
		From the Registered interface
		-------------------------------------------
	*/
	public Date getEntryDate()
	{
		return DateUtil.copy(getData().getEntryDate());
	}
	// -------------------------------------------

	/*
		From the BasicItem class
		-------------------------------------------
	*/
	@Override
	void onBeforeCommit(Action action) 
		throws NotLoggedInException, BaseException 
	{
		super.onBeforeCommit(action);
		if (action == Action.DELETE)
		{
			SessionControl sc = getSessionControl();
			if (sc.getActiveProjectId() == getId())
			{
				sc.setActiveProject(null);
			}
		}
	}
	/**
		If the logged in user is a member of this project, read or use 
		permission is granted.
	*/
	void initPermissions(int granted, int denied)
		throws BaseException
	{
		int projectPermission = getSessionControl().getProjectPermission(this);
		granted |= projectPermission & Permission.grant(Permission.READ, Permission.USE);
		super.initPermissions(granted, denied);
	}
	/**
		Make sure plug-ins always has at least read access to the active project.
		@since 2.9
	*/
	PluginPermission getPluginPermissions()
	{
		PluginPermission pp = super.getPluginPermissions();
		if (getId() == getSessionControl().getActiveProjectId())
		{
			pp = new PluginPermission(pp.granted | Permission.grant(Permission.READ), pp.denied);
		}
		return pp;
	}
	// -------------------------------------------

	/**
		The permissions to use when automatically sharing new items to
		the project when it is the active project. The default is READ, 
		USE, WRITE and DELETE. Note that this setting is ignored when
		a permission template is used ({@link #getPermissionTemplate()}).
		@since 2.10
	*/
	public Set<Permission> getAutoPermission()
	{
		return Permission.fromInt(getData().getAutoPermission());
	}
	
	/**
		Set the default permissions to use when automatically sharing
		new item the project when it is the active project. Note that
		this setting is ignored when a permission template is used 
		({@link #getPermissionTemplate()}).
		@param permissions The permissions, must not be null or empty
		@throws PermissionDeniedException If the logged in user doesn't
			have write permission
		@throws InvalidDataException If the permissions are null or empty
		@since 2.10
	*/
	public void setAutoPermission(Set<Permission> permissions)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		if (permissions == null || permissions.isEmpty())
		{
			throw new InvalidUseOfNullException("permissions");
		}
		getData().setAutoPermission(Permission.grant(permissions));
	}

	/**
		Get the permission template that is used to share new items when this project
		is active. Note that if a permission template is used the new items are only
		shared to the active project if the project is part of the template.
		@return A permission template or null if no template has been specified
		@since 2.16
	*/
	public PermissionTemplate getPermissionTemplate()
	{
		return getDbControl().getItem(PermissionTemplate.class, getData().getPermissionTemplate());
	}
	
	/**
		Set the permission template that should be used to share new items when
		this project is active.
		@param template A permission template, or null to not use any template
		@since 2.16
	*/
	public void setPermissionTemplate(PermissionTemplate template)
	{
		checkPermission(Permission.WRITE);
		getData().setPermissionTemplate(template == null ? null : template.getData());
	}
	
	/**
		Grant a user permissions to this project. Use an empty <code>Set</code> or null
		to remove all permissions for the user.

		@param user The <code>User</code>
		@param permissions The permissions to grant, or null to revoke all permissions
		@throws PermissionDeniedException If the logged in user doesn't have
			{@link Permission#WRITE} permission for the project
		@throws InvalidDataException If the user is null
		@see Permission
	*/
	public void setPermissions(User user, Set<Permission> permissions)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		if (user == null) throw new InvalidUseOfNullException("user");
		if (permissions == null || permissions.isEmpty())
		{
			getData().getUsers().remove(user.getData());
		}
		else
		{
			getData().getUsers().put(user.getData(), Permission.grant(permissions));
		}
	}
	
	/**
		Get the permissions for a user in this project.
		@param user The <code>User</code> for which we want to get the permission
		@return A <code>Set</code> containing the granted permissions, or an
			empty set if no permissions have been granted
		@throws InvalidDataException If the user is null
		@see Permission
	*/
	public Set<Permission> getPermissions(User user)
		throws InvalidDataException
	{
		if (user == null) throw new InvalidUseOfNullException("user");
		return Permission.fromInt(getData().getUsers().get(user.getData()));
	}

	/**
		Get a query that returns the users that
		are members of this project. This query excludes users that the logged 
		in user doesn't have permission to read.
	 	@return A restricted item query. 
	 	@throws BaseException If the query couldn't be created or configured. 
		@see User#getQuery()
	*/
	public ItemQuery<User> getUsers()
		throws BaseException
	{
		ItemQuery<User> query = User.getQuery();
		query.joinPermanent(Hql.innerJoin("projects", Item.PROJECT.getAlias()));
		query.restrictPermanent(
			Restrictions.eq(
				Hql.alias(Item.PROJECT.getAlias()), 
				Hql.entity(this)
			)
		);
		return query;
	}
	
	/**
		Grant a group permissions to this project. Use an empty <code>Set</code> or null
		to remove all permissions for the group. To share to the {@link Group#EVERYONE}
		group {@link Permission#SHARE_TO_EVERYONE} is required.

		@param group The <code>Group</code>
		@param permissions The permissions to grant, or null to revoke all permissions
		@throws PermissionDeniedException If the logged in user doesn't have
			{@link Permission#WRITE} permission for the project
		@throws InvalidDataException If the group is null
		@throws BaseException If there is another error
		@see Permission
	*/
	public void setPermissions(Group group, Set<Permission> permissions)
		throws PermissionDeniedException, InvalidDataException, BaseException
	{
		checkPermission(Permission.WRITE);
		if (group == null) throw new InvalidUseOfNullException("group");
		if (permissions == null || permissions.isEmpty())
		{
			getData().getGroups().remove(group.getData());
		}
		else
		{
			if ((group.getId() == SystemItems.getId(Group.EVERYONE)) && 
				!getSessionControl().hasSystemPermission(Permission.SHARE_TO_EVERYONE))
			{
				throw new PermissionDeniedException("Not allowed to share to the EVERYONE group.");
			}
			getData().getGroups().put(group.getData(), Permission.grant(permissions));
		}
	}
	
	/**
		Get the permissions for a group in this project.
		@param group The <code>Group</code> for which we want to get the permission
		@return An <code>EnumSet</code> containing the granted permissions, or an
			empty set if no permissions have been granted
		@throws InvalidDataException If the group is null
		@see Permission
	*/
	public Set<Permission> getPermissions(Group group)
		throws InvalidDataException
	{
		if (group == null) throw new InvalidUseOfNullException("usgroup");
		return Permission.fromInt(getData().getGroups().get(group.getData()));
	}
	
	/**
		Get a query that returns the groups that
		are members of this project. This query excludes groups that the logged 
		in user doesn't have permission to read.

		@see Group#getQuery()
	*/
	public ItemQuery<Group> getGroups()
	{
		ItemQuery<Group> query = Group.getQuery();
		query.joinPermanent(Hql.innerJoin("projects", Item.PROJECT.getAlias()));
		query.restrictPermanent(
			Restrictions.eq(
				Hql.alias(Item.PROJECT.getAlias()), 
				Hql.entity(this)
			)
		);
		return query;
	}

	/**
		 @deprecated Use {@link #getItems(Item, int, int, Permission, Set)} instead
	*/
	public ResultList<Shareable> getItems(Item itemType, int firstItem, int maxItems, Permission permission)
	{
		return getItems(itemType == null ? null : Collections.singleton(itemType), 
				firstItem, maxItems, permission, null, null);
	}
	
	/**
		@see #getItems(Set, int, int, Permission, Set)
		@since 2.3.1
	*/
	public ResultList<Shareable> getItems(Item itemType, int firstItem, int maxItems, 
			Permission permission, Set<Include> include)
	{
		return getItems(itemType == null ? null : Collections.singleton(itemType), 
			firstItem, maxItems, permission, include, null);
	}
	
	/**
		@since 2.7.1
		@deprecated In 2.15, use {@link #getItems(Set, int, int, Permission, Set, Restriction)}
			instead
	*/
	public ResultList<Shareable> getItems(Set<Item> itemTypes, int firstItem, int maxItems, 
			Permission permission, Set<Include> include)
	{
		return getItems(itemTypes, firstItem, maxItems, permission, include, null);
	}

	
	/**
		Load the items in a project. If this project is the active project it will load
		all items in the project. If this project isn't the active project it will
		only load items that the logged in user owns.

		@param itemTypes Limit the list to only return items of the specified types,
			use null or an empty set to return items of all types
		@param firstItem The index of the first item to return (0-based)
		@param maxItems The maximum number of items to return, or 0 to return all items
		@param permission The permission the logged in user must have on the item
		@param include Options for which items that should be included/excluded from the
			result, or null to include all non-removed items
		@param restriction An optional restriction that is applied to the
			queries to further reduce the list
		@return A list containing shareable items
		@see SessionControl#setActiveProject(Project)
		@since 2.15
	*/
	public ResultList<Shareable> getItems(Set<Item> itemTypes, int firstItem, int maxItems, 
		Permission permission, Set<Include> include, Restriction restriction)
	{
		if (include == null) include = EnumSet.of(Include.NOT_REMOVED, Include.MINE, Include.OTHERS);
		
		DbControl dc = getDbControl();
		SessionControl sc = dc.getSessionControl();
		org.hibernate.Session session = dc.getHibernateSession();

		// First task is to get all project keys sharing items to this project
		Set<Integer> projectKeys = null;
		if (this.getId() != sc.getActiveProjectId())
		{
			// Load the project keys for the current project
			org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "GET_PROJECTKEYIDS_FOR_PROJECT");
			/*
				SELECT pk.keyId
				FROM ProjectKeys pk
				WHERE pk.projectId = :projectId
			*/
			query.setInteger("projectId", this.getId());
			projectKeys = new HashSet<Integer>(HibernateUtil.loadList(Integer.class, query, sc));
			
			include.remove(Include.OTHERS);
		}
		else
		{
			// The project keys for the active project are already loaded
			projectKeys = sc.getProjectKeys();
			
			// If a permission is specified we must filter on that too...
			if (permission != null && permission != Permission.READ)
			{
				Set<Integer> permissionKeys = sc.getProjectKeys(permission);
				if (permissionKeys == null || permissionKeys.size() == 0) 
				{
					permissionKeys = QueryRuntimeFilterFactory.ZERO_SET;
				}
				org.hibernate.Filter permissionFilter = HibernateUtil.enableFilter(session, "ownedByOrInProject");
				permissionFilter.setParameter("owner", sc.getLoggedInUserId());
				permissionFilter.setParameterList("projectKeys", permissionKeys);
			}
		}
		
		// Filter that excludes items not shared to this project
		org.hibernate.Filter keyFilter = HibernateUtil.enableFilter(session, "inProject");
		// The list must contain at least one element
		if (projectKeys == null || projectKeys.size() == 0) 
		{
			projectKeys = QueryRuntimeFilterFactory.ZERO_SET;
		}
		keyFilter.setParameterList("projectKeys", projectKeys);
		keyFilter.setParameter("owner", sc.getLoggedInUserId());
		
		// Add filters for 'removed' option
		if (!(include.contains(Include.REMOVED) && include.contains(Include.NOT_REMOVED)))
		{
			org.hibernate.Filter removedFilter = HibernateUtil.enableFilter(session, "isRemoved");
			removedFilter.setParameter("removed", include.contains(Include.REMOVED));
		}
		// Add filter for 'owned by me' and 'owned by others' options
		if (include.contains(Include.MINE) && include.contains(Include.OTHERS))
		{
			// Do nothing; all items are included
		}
		else if (include.contains(Include.MINE))
		{
			org.hibernate.Filter ownerFilter = HibernateUtil.enableFilter(session, "ownedBy");
			ownerFilter.setParameter("owner", sc.getLoggedInUserId());
		}
		else if (include.contains(Include.OTHERS))
		{
			org.hibernate.Filter otherFilter = HibernateUtil.enableFilter(session, "notOwnedBy");
			otherFilter.setParameter("owner", sc.getLoggedInUserId());			
		}
		else
		{
			org.hibernate.Filter denyAllFilter = HibernateUtil.enableFilter(session, "denyAll");
		}
		
		List<ShareableData> inProject = null;
		if (itemTypes != null && itemTypes.size() > 0)
		{
			inProject = new ArrayList<ShareableData>();
			for (Item item : itemTypes)
			{
				if (ShareableData.class.isAssignableFrom(item.getDataClass()))
				{
					// Query to load all shareable items of a specific subclass
					AbstractHqlQuery hql = 
						new AbstractHqlQuery(item.getDataClass().getName(), "item", false){};
					hql.select(Selects.expression(Hql.alias("item"), null, true));
					if (restriction != null) hql.restrict(restriction);
					/*
						SELECT item
						FROM <class> item
						[WHERE restriction]
					*/
					inProject.addAll(HibernateUtil.loadList(ShareableData.class, hql.getMainHqlQuery(dc), sc));
				}
			}
		}
		else
		{
			// Query to load all shareable items
			AbstractHqlQuery hql = 
				new AbstractHqlQuery("net.sf.basedb.core.data.ShareableData", "item", false){};
			hql.select(Selects.expression(Hql.alias("item"), null, true));
			if (restriction != null) hql.restrict(restriction);
			/*
				SELECT item
				FROM net.sf.basedb.core.data.ShareableData item
				[WHERE restriction]
			*/
			inProject = HibernateUtil.loadList(ShareableData.class, hql.getMainHqlQuery(dc), sc);
		}
		// Disable filters
		HibernateUtil.disableFilter(session, "ownedByOrInProject");
		HibernateUtil.disableFilter(session, "inProject");
		HibernateUtil.disableFilter(session, "ownedBy");
		HibernateUtil.disableFilter(session, "notOwnedBy");
		HibernateUtil.disableFilter(session, "isRemoved");
		HibernateUtil.disableFilter(session, "denyAll");
		
		List<Shareable> items = new ArrayList<Shareable>(inProject.size());
		int index = 0;
		int totalItems = 0;
		boolean useLimit = firstItem >= 0 && maxItems > 0;
		int lastItem = firstItem + maxItems;
		for (ShareableData sd : inProject)
		{
		BasicData data = (BasicData)sd;
		try
		{
			BasicItem item = dc.getItem(BasicItem.class, data);
			if (!useLimit || (index >= firstItem && index < lastItem))
			{
				items.add((Shareable)item);
			}
			index++;
			totalItems++;
		}
		catch (PermissionDeniedException ex)
		{}
		}
		return new UnmodifiableResultList<Shareable>(items, totalItems);
	}

	/**
	  	Sets item as default for project.
	  	
	  	@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
	  	@param defaultItem The {@link net.sf.basedb.core.BasicItem} to set as default.
	  		Use null when to remove a default value.
	  	@param defaultType The {@link Default} to be set.
	*/
	public void setDefaultItem(DbControl dc, BasicItem defaultItem, Default defaultType)
	{
		String lnkName = defaultType.getName();
		BasicData defaultData = null;
		if (defaultItem == null)
		{
			AnyToAny.unlinkFrom(dc, this, lnkName);
		}
		else
		{
			AnyToAny defaultLink = AnyToAny.getNewOrExisting(dc, this, lnkName, defaultItem, false);
			if (defaultLink.getId() == 0) dc.saveItem(defaultLink);
			defaultData = defaultItem.getData();
		}
		if (defaults == null) defaults = new HashMap<Default, BasicData>();
		defaults.put(defaultType, defaultData);
	}
	
	/**
	  	Gets the project's default item of a certain itemType defined in <code>defaultType</code>.
	  	If no item of the specified type is found this will return Null 
	  	
	  	@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
	  	@param defaultType The project's <code>Default</code> to get.
	  	@return A BasicItem or null if no item was found.
	*/
	public BasicItem getDefaultItem(DbControl dc, Default defaultType)
	{
		if (defaultType == null) throw new InvalidUseOfNullException("defaultType");
		BasicData defaultData = getDefaultData(dc, defaultType);
		return dc.getItem(BasicItem.class, defaultData);
	}
	
	/**
	  	Gets the project's default item as a data object.
	  	If no item of the specified type is found this will return Null
	  	@since 2.9
	*/
	BasicData getDefaultData(DbControl dc, Default defaultType)
	{
		BasicData defaultData = null;		
		if (defaults != null && defaults.containsKey(defaultType))
		{
			defaultData = defaults.get(defaultType);
		}
		else
		{
			org.hibernate.Session session = dc.getHibernateSession();
			AnyToAnyData any = AnyToAny.getDataByName(session, getId(), getType(), defaultType.getName());
			if (any != null)
			{
				defaultData = HibernateUtil.loadData(session, 
					defaultType.getItemType().getDataClass(), any.getToId());
			}
			if (defaults == null) defaults = new HashMap<Default, BasicData>();
			defaults.put(defaultType, defaultData);
		}
		return defaultData;
	}
	
	/**
	 	Gets the default raw data type for this project.
	 	If no default value has been set for raw data type this will return Null.
	 	
	  	@return A {@link net.sf.basedb.core.RawDataType} object
	*/
	public RawDataType getDefaultRawDataType()
	{
		RawDataType rdt = null;
		String id = getData().getProjectDefaults().get("raw_data_type");
		rdt = RawDataTypes.getRawDataType(id);		
		return rdt;
	}
	
	/**
	  	Sets a raw data type to be used as default for a project.
	  	The default value will be reset if <code>rawDataType</code> is Null
	  	
	  	@param rawDataType the {@link net.sf.basedb.core.RawDataType} to use as default.
	*/
	public void setDefaultRawDataType(RawDataType rawDataType)
	{
		checkPermission(Permission.WRITE);
		if (rawDataType == null) 
		{
			getData().getProjectDefaults().remove("raw_data_type");
		}
		else
		{		
			getData().getProjectDefaults().put("raw_data_type", rawDataType.getId());
		}
	}
	
	public ProjectInfo toTransferable(ProjectInfo info)
	{
		super.toTransferable(info);
		info.setDescription(this.getDescription());
		info.setName(this.getName());
		return info;
	}
	
	/**
		A runtime filter implementation that limits a query to only
		return projects where the logged in user is a member or owner
		unless the logged in user has generic read permission.
	*/
	private static class QueryRuntimeFilterImpl
		implements QueryRuntimeFilter
	{
		public void enableFilters(QueryRuntimeFilterManager manager, EntityQuery query, DbControl dc)
		{
			org.hibernate.Filter filter = null;
			SessionControl sc = dc.getSessionControl();
			int owner = sc.getLoggedInUserId();
			Set<Integer> projects = sc.getProjects();
			if (projects == null || projects.size() == 0) projects = Collections.singleton(0);

			if (sc.isDenied(query.getItemPermission(), Item.PROJECT))
			{
				manager.enableFilter("denyAll");
			}
			else if (query.isIncluded(Include.OTHERS) && sc.hasPermission(Permission.READ, query.getItemType()))
			{
				// Logged in user has generic read permission and wants other user's items...
				if (!query.isIncluded(Include.MINE))
				{
					// ... but not the one's with ownership
					filter = manager.enableFilter("notOwnedBy");
					if (filter != null)
					{
						filter.setParameter("owner", owner);
					}
				}
			}
			else if (query.isIncluded(Include.MINE, Include.SHARED))
			{
				// Logged in user wants own and shared projects
				filter = manager.enableFilter("ownedByOrMemberOf");
				if (filter != null)
				{
					filter.setParameter("owner", owner);
					filter.setParameterList("items", projects);
				}
			}
			else if (query.isIncluded(Include.SHARED))
			{
				// Logged in user wants only shared projects
				filter = manager.enableFilter("memberOf");
				if (filter != null)
				{
					filter.setParameter("owner", owner);
					filter.setParameterList("items", projects);
				}
			}
			else if (query.isIncluded(Include.MINE))
			{
				//Logged in user only wants own projects
				filter = manager.enableFilter("ownedBy");
				if (filter != null)
				{
					filter.setParameter("owner", owner);
				}
			}
			else
			{
				manager.enableFilter("denyAll");
			}
		}
	}
	
	public enum Default
	{		
		SOFTWARE("default_software", "Software", Item.SOFTWARE, SoftwareType.FEATURE_EXTRACTION, Item.RAWBIOASSAY),
		ARRAYDESIGN("default_arraydesign", "Array design", Item.ARRAYDESIGN, null, null),
		HYBRIDIZATION_HARDWARE("default_hybridization_hardware", "Hybridization station", Item.HARDWARE, HardwareType.HYBRIDIZATION_STATION, Item.HYBRIDIZATION),
		SCANNER_HARDWARE("default_scanner_hardware", "Scanner", Item.HARDWARE, HardwareType.SCANNER, Item.SCAN),	
		PRINTROBOT_HARDWARE("default_print_robot_hardware", "Print robot", Item.HARDWARE, HardwareType.PRINT_ROBOT, Item.ARRAYBATCH),
		SCANNING_PROTOCOL("default_scanning_protocol", "Scanning protocol", Item.PROTOCOL, ProtocolType.SCANNING, Item.SCAN),
		SAMPLING_PROTOCOL("default_sampling_protocol", "Sampling protocol", Item.PROTOCOL, ProtocolType.SAMPLING, Item.SAMPLE),
		EXTRACTING_PROTOCOL("default_extracting_protocol", "Extracting protocol", Item.PROTOCOL, ProtocolType.EXTRACTION, Item.EXTRACT),		
		LABELING_PROTOCOL("default_labeling_protocol", "Labeling protocol", Item.PROTOCOL, ProtocolType.LABELING, Item.LABELEDEXTRACT),
		HYBRIDIZATION_PROTOCOL("default_hybridization_protocol", "Hybridization protocol", Item.PROTOCOL, ProtocolType.HYBRIDIZATION, Item.HYBRIDIZATION),
		FEATURE_EXTRACTION_PROTOCOL("default_feature_extraction_protocol", "Feat. extraction protocol", Item.PROTOCOL, ProtocolType.FEATURE_EXTRACTION, Item.RAWBIOASSAY),
		POOLING_PROTOCOL("default_pooling_protocol", "Pooling protocol", Item.PROTOCOL, ProtocolType.POOLING, null),
		PRINTING_PROTOCOL("default_printing_protocol", "Printing protocol", Item.PROTOCOL, ProtocolType.PRINTING, Item.ARRAYBATCH),
		RAW_DATA_TYPE("default_raw_data_type", "Raw data type", null, null, null),
		PLATFORM("default_platform", "Platform", Item.PLATFORM, null, null),
		PLATFORM_VARIANT("default_variant", "Variant", Item.PLATFORMVARIANT, null, null);
		
		private String name;
		private String shortName;
		private Item itemType;
		private String type;
		private Item attachesToItemType;
		
		private Default(String name, String shortName, Item itemType, String type, Item attachesToItemType)
		{
			this.name = name;
			this.shortName = shortName;
			this.itemType = itemType;
			this.type = type;
			this.attachesToItemType = attachesToItemType;
		}
		
		public String getName()
		{
			return name;
		}
		
		public String getShortName()
		{
			return shortName;
		}
		
		public Item getItemType()
		{
			return itemType;
		}
		
		public String getType()
		{
			return type;
		}
		
		/**
			The item type an entry of this default type is usually attached to.
			For example a {@link Default#SAMPLING_PROTOCOL} attaches to a 
			{@link Item#SAMPLE}.
			@return An item object or null if it is not relevant
			@since 2.10
		*/
		public Item getAttachesToItemType()
		{
			return attachesToItemType;
		}
	}
}



