/*
	$Id: ShareableUtil.java 5384 2010-08-13 08:50:27Z nicklas $

	Copyright (C) 2005 Samuel Andersson, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.LinkedList;
import java.util.List;

import net.sf.basedb.core.data.GroupData;
import net.sf.basedb.core.data.ItemKeyData;
import net.sf.basedb.core.data.ProjectKeyData;
import net.sf.basedb.core.data.ShareableData;
import net.sf.basedb.core.signal.ThreadSignalHandler;

/**
	Utility methods that will make it easier to implement
	the {@link Shareable} interface, including data validation.
	
	@author Nicklas
	@version 2.0
	@see Shareable
	@base.modified $Date: 2010-08-13 10:50:27 +0200 (Fri, 13 Aug 2010) $
*/
public class ShareableUtil
{

	/**
		Set the item key of a {@link ShareableData} object.
	 	@param shareableData The data object to set the item key on. Null is not allowed. 
	 	@param itemKey Item key to be set.
	 	@deprecated Use {@link #setItemKey(DbControl, ShareableData, ItemKey)} instead
	*/
	public static void setItemKey(ShareableData shareableData, ItemKey itemKey)
	{
		shareableData.setItemKey(itemKey == null ? null : itemKey.getData());
	}
	
	/**
		Set the project key of a {@link ShareableData} object.
	 	@param shareableData The data object to set project key to. Null is not allowed. 
	 	@param projectKey Project key to be set.
	*/
	public static void setProjectKey(ShareableData shareableData, ProjectKey projectKey)
	{
		shareableData.setProjectKey(projectKey == null ? null : projectKey.getData());
	}
	
	/**
		Set the item key of a {@link ShareableData} object.
		If the logged in user doesn't have {@link Permission#SHARE_TO_EVERYONE} permission
		then new item key must, if it includes sharing to the everyone group, have exactly
		the same permissions as there is in the existing item key.
		
		@param dc A DbControl
		@param shareableData The item to share
		@param itemKey The new item key
		@throws PermissionDeniedException If the logged in user doesn't have permission 
			and the new item key contains a permission that is different from the
			existing permission
		@since 2.6
	*/
	public static void setItemKey(DbControl dc, ShareableData shareableData, ItemKey itemKey)
	{
		if (itemKey != null && !dc.getSessionControl().hasSystemPermission(Permission.SHARE_TO_EVERYONE))
		{
			GroupData everyone = HibernateUtil.loadData(dc.getHibernateSession(), GroupData.class, 
				SystemItems.getId(Group.EVERYONE));
			
			// Check permissions for Everyone in the new key
			ItemKeyData newKey = itemKey.getData();
			int newPermission = IntegerUtil.getInt(newKey.getGroups().get(everyone));
			if (newPermission != 0)
			{
				// Check permission for Everyone in the old key
				ItemKeyData oldKey = shareableData.getItemKey();
				int oldPermission = IntegerUtil.getInt(oldKey == null ? 0 : oldKey.getGroups().get(everyone));
				if (newPermission != oldPermission)
				{
					throw new PermissionDeniedException("Not allowed to change permissions for the to the EVERYONE group.");
				}
			}
		}
		shareableData.setItemKey(itemKey == null ? null : itemKey.getData());
	}
	
	/**
		Share all files and sub-directories in the given directory with the
		same permissions. This method will of course only change the
		permissions on items that the current user has {@link Permission#SET_PERMISSION}
		on.

		@param dc A DbControl for database access
		@param directory The directory
		@return The number of files and directories processed
		@since 2.8
	*/
	public static int shareRecursively(DbControl dc, Directory directory)
	{
		ItemKeyData itemKey = directory.getData().getItemKey();
		ProjectKeyData projectKey = directory.getData().getProjectKey();

		List<Directory> toProcess = new LinkedList<Directory>();
		toProcess.add(directory);
		int numItems = 0;
		
		while (toProcess.size() > 0)
		{
			ThreadSignalHandler.checkInterrupted();
			Directory currentDir = toProcess.remove(0);
			
			// Share files
			ItemQuery<File> fileQuery = currentDir.getFiles();
			fileQuery.include(Include.ALL);
			fileQuery.setItemPermission(Permission.SET_PERMISSION);
			for (File f : fileQuery.list(dc))
			{
				f.getData().setItemKey(itemKey);
				f.getData().setProjectKey(projectKey);
				numItems++;
			}
			
			// Share directories
			ItemQuery<Directory> dirQuery = currentDir.getSubDirectories();
			dirQuery.include(Include.ALL);
			dirQuery.setItemPermission(Permission.SET_PERMISSION);
			for (Directory d : dirQuery.list(dc))
			{
				d.getData().setItemKey(itemKey);
				d.getData().setProjectKey(projectKey);
				numItems++;
				
				// Make sure that we go into this subdirectory
				toProcess.add(d);
			}
		}
		return numItems;
	}
	
	/**
		Copy permissions from one item to another. If the 'from' item
		is shared to the 'EVERYONE' group, but the logged in user lacks
		the {@link Permission#SHARE_TO_EVERYONE} permission, the share to
		everyone is removed. If the 'to' item is a new item and a project 
		is active the new item is also shared with the settings from 
		the project. 

		@param from The item to copy permissions from
		@param to The item to copy permissions to
		@since 2.8
	*/
	public static void copyPermissions(DbControl dc, Shareable from, Shareable to)
	{
		SessionControl sc = dc.getSessionControl();
		
		BasicItem fromItem = (BasicItem)from;
		if (fromItem.isDetached()) dc.reattachItem(fromItem, false);
		
		ItemKey itemKey = from.getItemKey();
		ProjectKey projectKey = from.getProjectKey();
		
		GroupPermissions newGroupPermissions = null;
		UserPermissions newUserPermissions = null;
		ProjectPermissions newProjectPermissions = null;
		
		// If 'to' is a new item we may need to merge permission from the active project
		if (to.getId() == 0)
		{
			// Merge with item key permissions from active project
			if (sc.getItemKeyId() != 0 && itemKey != null)
			{
				ItemKey activeKey = ItemKey.getById(dc, sc.getItemKeyId());
				newUserPermissions = new UserPermissions(itemKey);
				newUserPermissions.merge(activeKey);
				newGroupPermissions = new GroupPermissions(itemKey);
				newGroupPermissions.merge(activeKey);
			}
			if (sc.getProjectKeyId() != 0 && projectKey != null)
			{
				newProjectPermissions = new ProjectPermissions(projectKey);
				newProjectPermissions.merge(ProjectKey.getById(dc, sc.getProjectKeyId()));
			}
		}
		
		// Check if we need to remove permissions for the EVERYONE group
		if (itemKey != null && !sc.hasSystemPermission(Permission.SHARE_TO_EVERYONE))
		{
			GroupData everyone = HibernateUtil.loadData(dc.getHibernateSession(), GroupData.class, 
				SystemItems.getId(Group.EVERYONE));

			if (itemKey.getData().getGroups().containsKey(everyone) || 
				(newGroupPermissions != null && newGroupPermissions.getGroupPermissions().containsKey(everyone)))
			{
				// Remove the EVERYONE grup
				if (newGroupPermissions == null) newGroupPermissions = new GroupPermissions(itemKey);
				if (newUserPermissions == null) newUserPermissions = new UserPermissions(itemKey);
				newGroupPermissions.getGroupPermissions().remove(everyone);
			}
		}

		if (newUserPermissions != null || newGroupPermissions != null)
		{
			int numUsers = newUserPermissions == null ? 0 : newUserPermissions.size();
			int numGroups = newGroupPermissions == null ? 0 : newGroupPermissions.size();
			if (numUsers == 0 && numGroups == 0)
			{
				itemKey = null;
			}
			else
			{
				itemKey = ItemKey.getNewOrExisting(dc, newUserPermissions, newGroupPermissions);
			}
		}
		if (newProjectPermissions != null)
		{
			if (newProjectPermissions.size() == 0)
			{
				projectKey = null;
			}
			else
			{
				projectKey = ProjectKey.getNewOrExisting(dc, newProjectPermissions);
			}
		}
		
		if (itemKey != null) to.setItemKey(itemKey);
		if (projectKey != null) to.setProjectKey(projectKey);
	}
	
	/**
		Merge one or more item keys into a new item key that combines all
		the permissions.
		
		@param keys An array with the item keys (null elements are ignored)
		@return A new or existing item key
		@since 2.16
	*/
	public static ItemKey merge(DbControl dc, ItemKey... keys)
	{
		ItemKey result = null;
		UserPermissions up = null;
		GroupPermissions gp = null;
		// We only need to merge if there are two or more non-null keys
		boolean needMerge = false;
		for (ItemKey key : keys)
		{
			if (key != null)
			{
				if (needMerge)
				{
					// We need to merge when at least two keys are not null
					if (up == null) 
					{
						up = new UserPermissions(result);
						gp = new GroupPermissions(result);
						result = null;
					}
					up.merge(key);
					gp.merge(key);
				}
				else
				{
					// First key that is not null...hope that we can use this as the result
					// but we need to merge otherwise
					result = key;
					needMerge = true;
				}
			}
		}
		if (up != null && (up.size() > 0 || gp.size() > 0)) 
		{
			result = ItemKey.getNewOrExisting(dc, up, gp);
		}
		return result;
	}
	
	/**
		Merge one or more project keys into a new project key that combines all
		the permissions.
		
		@param keys An array with the project keys (null elements are ignored)
		@return A new or existing project key
		@since 2.16
	*/
	public static ProjectKey merge(DbControl dc, ProjectKey... keys)
	{
		ProjectKey result = null;
		ProjectPermissions pp = null;
		// We only need to merge if there are two or more non-null keys
		boolean needMerge = false;
		for (ProjectKey key : keys)
		{
			if (key != null)
			{
				if (needMerge)
				{
					// We need to merge when at least two keys are not null
					if (pp == null) 
					{
						pp = new ProjectPermissions(result);
						result = null;
					}
					pp.merge(key);
				}
				else
				{
					// First key that is not null...hope that we can use this as the result
					// but we need to merge otherwise
					result = key;
					needMerge = true;
				}
			}
		}
		if (pp != null && pp.size() > 0) result = ProjectKey.getNewOrExisting(dc, pp);
		return result;
	}
}
