/*
  $Id: DataFileTypeData.java 4517 2008-09-11 20:11:02Z jari $

	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

  BASE is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 3
  of the License, or (at your option) any later version.

  BASE is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.data;

import java.util.Set;

/**
	This class hols information about individual files attached
	to items that store data in files instead of in the
	database.
	
	@author Nicklas
	@version 2.5
	@see net.sf.basedb.core.DataFileType
	@base.modified $Date: 2008-09-11 22:11:02 +0200 (Thu, 11 Sep 2008) $
	@see <a href="../../../../../../html/developerdoc/api/data_api.platforms.html">Data API - Experimental platforms</a>
	@hibernate.class table="`DataFileTypes`" lazy="false"
*/
public class DataFileTypeData
	extends BasicData
	implements NameableData, RemovableData
{

	/*
		From the NameableData interface
		-------------------------------------------
	*/
	private String name;
	public String getName()
	{
		return name;
	}
	public void setName(String name)
	{
		this.name = name;
	}
	
	private String description;
	public String getDescription()
	{
		return description;
	}
	public void setDescription(String description)
	{
		this.description = description;
	}
	/*
		-------------------------------------------		
	*/
	
	/*
		From the RemovableData interface
		-------------------------------------------
	*/
	private boolean removed;
	public boolean isRemoved()
	{
		return removed;
	}
	public void setRemoved(boolean removed)
	{
		this.removed = removed;
	}
	// -------------------------------------------

	/**
		The maximum length of the external ID that can be stored in the database.
		@see #setExternalId(String)
	*/
	public static final int MAX_EXTERNAL_ID_LENGTH = 255;
	private String externalId;
	/**
		Get the external id for the item. 
		@return The external id of the item
		@hibernate.property column="`external_id`" type="string" length="255" 
			not-null="true" update="false" unique="true"
	*/
	public String getExternalId()
	{
		return externalId;
	}
	public void setExternalId(String externalId)
	{
		this.externalId = externalId;
	}
	
	/**
		The maximum length of the file extension that can be stored in the
		database.
	*/
	public static final int MAX_EXTENSION_LENGTH = 255;
	private String extension;
	/**
		Get the extension that files of this type usually have. Null if 
		there is no specific extension.
		@hibernate.property column="extension" type="string" length="255"
	*/
	public String getExtension()
	{
		return extension;
	}
	public void setExtension(String extension)
	{
		this.extension = extension;
	}

	/**
		The maximum length of the class name that can be stored in the
		database.
	*/
	public static final int MAX_CLASSNAME_LENGTH = 255;
	private String validatorClass;
	/**
		Get the class name of the Java class that is the validator for
		this file type. The class must implement the {@link net.sf.basedb.core.filehandler.DataFileValidator}
		interface. Null if there is no validator for this file type.
		@hibernate.property column="`validator_class`" type="string" length="255" not-null="false"
	*/
	public String getValidatorClass()
	{
		return validatorClass;
	}
	public void setValidatorClass(String validatorClass)
	{
		this.validatorClass = validatorClass;
	}
	
	public static final int MAX_JARPATH_LENGTH = 65535;
	private String validatorJarPath;
	/**
		Get the path to the JAR file where the validator class is located. If the
		value is null the value must be in the classpath.
		@return The path to the JAR file
		@hibernate.property column="`validator_jarpath`" type="text" not-null="false"
		@since 2.6
	*/
	public String getValidatorJarPath()
	{
		return validatorJarPath;
	}
	public void setValidatorJarPath(String validatorJarPath)
	{
		this.validatorJarPath = validatorJarPath;
	}
	
	
	private String metadataReaderClass;
	/**
		Get the class name of the Java class that extracts metadata from
		this file type. The class must implement the {@link net.sf.basedb.core.filehandler.DataFileMetadataReader}
		interface. Null if there is no extractor for this file type.
		@hibernate.property column="`extractor_class`" type="string" length="255" not-null="false"
	*/
	public String getMetadataReaderClass()
	{
		return metadataReaderClass;
	}
	public void setMetadataReaderClass(String metadataReaderClass)
	{
		this.metadataReaderClass = metadataReaderClass;
	}
	
	private String metadataReaderJarPath;
	/**
		Get the path to the JAR file where the validator class is located. If the
		value is null the value must be in the classpath.
		@return The path to the JAR file
		@hibernate.property column="`extractor_jarpath`" type="text" not-null="false"
		@since 2.6
	*/
	public String getMetadataReaderJarPath()
	{
		return metadataReaderJarPath;
	}
	public void setMetadataReaderJarPath(String metadataReaderJarPath)
	{
		this.metadataReaderJarPath = metadataReaderJarPath;
	}
	
	private int itemType;
	/**
		The item type this file type can be associated with.
		@hibernate.property column="`item_type`" type="int" not-null="true" update="false"
	*/
	public int getItemType()
	{
		return itemType;
	}
	public void setItemType(int itemType)
	{
		this.itemType = itemType;
	}
	
	private FileTypeData genericType;
	/**
		Get the generic file type of files with this member type.
		@hibernate.many-to-one outer-join="false"
		@hibernate.column name="`filetype_id`" not-null="false"
	*/
	public FileTypeData getGenericType()
	{
		return genericType;
	}
	public void setGenericType(FileTypeData genericType)
	{
		this.genericType = genericType;
	}

	private Set<PlatformFileTypeData> platforms;
	/**
		This is the inverse end.
		@see PlatformFileTypeData#getDataFileType()
		@hibernate.set lazy="true" inverse="true" 
		@hibernate.collection-key column="`datafiletype_id`"
		@hibernate.collection-one-to-many class="net.sf.basedb.core.data.PlatformFileTypeData"
	*/
	Set<PlatformFileTypeData> getPlatforms()
	{
		return platforms;
	}
	void setPlatforms(Set<PlatformFileTypeData> platforms)
	{
		this.platforms = platforms;
	}
	
	private Set<FileSetMemberData> members;
	/**
		This is the inverse end.
		@see FileSetMemberData#getDataFileType()
		@hibernate.set lazy="true" inverse="true" 
		@hibernate.collection-key column="`datafiletype_id`"
		@hibernate.collection-one-to-many class="net.sf.basedb.core.data.FileSetMemberData"
	*/
	Set<FileSetMemberData> getFileSetMembers()
	{
		return members;
	}
	void setFileSetMembers(Set<FileSetMemberData> members)
	{
		this.members = members;
	}

}
