/*
	$Id: AffymetrixFileHandler.java 5327 2010-04-29 10:14:59Z nicklas $

	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.filehandler;

import affymetrix.fusion.cdf.FusionCDFData;
import affymetrix.fusion.cel.FusionCELData;
import net.sf.basedb.core.File;
import net.sf.basedb.core.FileStoreEnabled;
import net.sf.basedb.core.InvalidDataException;
import net.sf.basedb.core.InvalidUseOfNullException;
import net.sf.basedb.core.ItemNotFoundException;
import net.sf.basedb.core.Platform;

/**
	This class is used as a superclass for all Affymetrix-based
	validators and metadata readers. It also has some useful utility
	methods.
	
	@author Nicklas
	@version 2.5
	@base.modified $Date: 2010-04-29 12:14:59 +0200 (Thu, 29 Apr 2010) $
*/
public abstract class AffymetrixFileHandler
	extends AbstractDataFileHandler
{

	protected AffymetrixFileHandler()
	{}
	
	/**
		Check if the item is using the affymetrix platform.
	 	@param item Item to check. Null is not allowed.
	 	@return TRUE if platform is used, FALSE otherwise.
	*/
	public boolean isAffymetrix(FileStoreEnabled item)
	{
		if (item == null) throw new InvalidUseOfNullException("item");
		Platform platform = item.getPlatform();
		if (platform == null || !Platform.AFFYMETRIX.equals(platform.getExternalId()))
		{
			return false;
		}
		return true;
	}

	/**
		Load a CDF file using the Affymetric Fusion SDK. This method
		checks that the file exists, and read all headers.
		
		@param cdfFile The file to load
		@return A FusionCDFData object representing the CDF file
		@throws ItemNotFoundException If the actual file is not on the server
		@throws InvalidDataException If the file is not a CDF file
	*/
	public FusionCDFData loadCdfFile(File cdfFile)
		throws ItemNotFoundException, InvalidDataException
	{
		if (cdfFile == null) throw new InvalidUseOfNullException("cdfFile");
		if (!cdfFile.getLocation().isDownloadable())
		{
			throw new ItemNotFoundException("Can't download file data for file '" + 
				cdfFile.getName() + "'; location=" + cdfFile.getLocation());
		}
		FusionCDFData cdf = new FusionCDFData();
		cdf.setFileName(cdfFile.getName());
		cdf.setInputStream(cdfFile.getDownloadStream(0));
		if (!cdf.readHeader())
		{
			throw new InvalidDataException("Could not read CDF file '" +
					cdfFile.getPath()+"': " + cdf.getError());
		}
		return cdf;
	}
	
	/**
		Load a CEL file using the Affymetric Fusion SDK. This method
		checks that the file exists, and read all headers.
		
		@param celFile The file to load
		@return A FusionCELData object representing the CEL file
		@throws ItemNotFoundException If the actual file is not on the server
		@throws InvalidDataException If the file is not a CEL file
	*/
	public FusionCELData loadCelFile(File celFile)
	{
		if (celFile == null) throw new InvalidUseOfNullException("celFile");
		if (!celFile.getLocation().isDownloadable())
		{
			throw new ItemNotFoundException("Can't download file data for file '" + 
					celFile.getName() + "'; location=" + celFile.getLocation());
		}
		FusionCELData cel = new FusionCELData();
		cel.setFileName(celFile.getName());
		cel.setInputStream(celFile.getDownloadStream(0));
		if (!cel.readHeader())
		{
			throw new InvalidDataException("Could not read CEL file '" +
					celFile.getPath()+"': " + cel.getError());
		}
		return cel;
	}
}
