/*
	$Id $

	Copyright (C) 2010 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util;

import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Locale;

import org.apache.http.Header;
import org.apache.http.HttpResponse;
import org.apache.http.client.HttpClient;

/**
	Useful methods related to HTTP stuff in general and
	Apache HTTP components in particular.
	
	@author Nicklas
	@since 2.16
	@base.modified $Date$
*/
public class HttpUtil 
{
	
	private static final SimpleDateFormat HTTP_DATE_FORMAT = new SimpleDateFormat("EEE, d MMM yyyy HH:mm:ss z", Locale.ENGLISH);

	/**
		Get the content length from the headers in the response.
		@param response The response from a HTTP request
		@return The value of the "Content-Length" header or -1 if not present 
			if it is unparsable
	*/
	public static long getContentLength(HttpResponse response)
	{
		return getLong(response, "Content-Length", -1L);
	}

	/**
		Get the date the content was last modified from the headers 
		in the response.
		@param response The response from a HTTP request
		@return The value of the "Last-Modified" header or null if not present 
			if it is unparsable
	*/
	public static Date getLastModified(HttpResponse response)
	{
		return getDate(response, "Last-Modified", null);
	}

	/**
		Get the content type of the content from the headers 
		in the response.
		@param response The response from a HTTP request
		@return The value of the "Content-Type" header or null if not present 
			if it is unparsable
	*/
	public static String getContentType(HttpResponse response)
	{
		return getString(response, "Content-Type", null);
	}

	
	public static String getString(HttpResponse response, String header, String defaultValue)
	{
		String result = defaultValue;
		Header h = response.getFirstHeader(header);
		if (h != null) result = h.getValue();
		return result;
	}
	
	public static Long getLong(HttpResponse response, String header, Long defaultValue)
	{
		Long result = defaultValue;
		Header h = response.getFirstHeader(header);
		if (h != null) result = Values.getLong(h.getValue(), defaultValue);
		return result;
	}
	
	public static Date getDate(HttpResponse response, String header, Date defaultValue)
	{
		Date result = defaultValue;
		Header h = response.getFirstHeader(header);
		if (h != null) 
		{
			try 
			{ 
				result = HTTP_DATE_FORMAT.parse(h.getValue());
			}
			catch (Exception ex) 
			{}
		}
		return result;
	}
	
	/**
		Safely shuts down a http client and it's connection manager
		without trowing an exception. Useful to use in try-catch-finally
		clauses.
	*/
	public static void shutdown(HttpClient client)
	{
		try
		{
			client.getConnectionManager().shutdown();
		}
		catch (RuntimeException ex)
		{}
	}
	
}
