/**
	$Id: ExtractToLabeledExtractTransformer.java 4723 2009-01-13 08:20:26Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.biomaterial;

import java.util.Collection;
import java.util.HashSet;
import java.util.Set;

import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Extract;
import net.sf.basedb.core.Include;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.LabeledExtract;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.util.collections.BasicItemToIdTransformer;
import net.sf.basedb.util.collections.CollectionTransformer;
import net.sf.basedb.util.collections.MultiStepCollectionTransformer;

/**
	Collection transformer that given a collection of extracts finds
	all labeled extracts that has the extracts as parents. Note that this
	transformer is only capable of finding labeled extracts that are immediate
	children to the extracts. If you want to include labeled extracts that are
	children to pooled extracts, you should combine this transformer
	with a {@link PooledChildrenTransformer} with the help of a
	{@link MultiStepCollectionTransformer}.

	@author Nicklas
	@version 2.10
	@base.modified $Date: 2009-01-13 09:20:26 +0100 (Tue, 13 Jan 2009) $
*/
public class ExtractToLabeledExtractTransformer
	implements CollectionTransformer<Extract, LabeledExtract>
{

	private final DbControl dc;
	
	/**
		Create a new Extract-to-labeled extract transformer.
		@param dc The DbControl to use for database access
	*/
	public ExtractToLabeledExtractTransformer(DbControl dc)
	{
		this.dc = dc;
	}
	
	/*
		From the CollectionTransformer interface
		----------------------------------------
	*/
	/**
		Add all labeled extracts that are children to any of the extracts in the
		source collection.
		@return The number of labeled extracts added
	*/
	@Override
	public int transform(Collection<? extends Extract> extracts,
			Collection<? super LabeledExtract> labeledExtracts)
	{
		// Copy source ID:s so we can use them as a parameter in the query
		Set<Integer> sourceIds = new HashSet<Integer>();
		BasicItemToIdTransformer idTransformer = new BasicItemToIdTransformer();
		idTransformer.transform(extracts, sourceIds);

		ItemQuery<LabeledExtract> query = createQuery();
		query.restrict(
			Restrictions.in(
				Hql.property("parent"), 
				Expressions.parameter("extracts")
			)
		);
		query.setParameter("extracts", sourceIds, Type.INT);
		int size = labeledExtracts.size();
		labeledExtracts.addAll(query.list(dc));
		return labeledExtracts.size() - size;
	}
	// -----------------------------------------------
	
	/**
		Create a query for loading labeled extracts. The default implemenation will
		simply load all labeled extracts. The restriction needed to locate the extracts
		that are parents is automatically applied by the <code>transform</code>
		method. Subclasses may override this method to create an alternate query
		(for example, with more filters).
		@return A query object
	*/
	protected ItemQuery<LabeledExtract> createQuery()
	{
		ItemQuery<LabeledExtract> query = LabeledExtract.getQuery();
		query.include(Include.ALL);
		return query;
	}

}
