/**
	$Id: SynchronizedSpotDataParser.java 5319 2010-04-20 08:02:16Z nicklas $

	Copyright (C) 2010 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.importer.spotdata;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import net.sf.basedb.core.signal.ThreadSignalHandler;
import net.sf.basedb.util.bfs.AnnotationParser;
import net.sf.basedb.util.bfs.BfsParser;
import net.sf.basedb.util.bfs.DataParser;
import net.sf.basedb.util.bfs.EventHandler;
import net.sf.basedb.util.bfs.EventType;

/**
	A parser that synchronizes the parsing of multiple underlying
	data parsers and a master reporter annotations parser. Before
	it can be used a master parser must be set by {@link #setMasterParser(AnnotationParser)}.
	Then, one or more data parsers is needed {@link #addDataParser(DataParser)}. 
	All parsers must have been parsed to beginning-of-file 
	(eg. {@link AnnotationParser#parseToBof()} and {@link DataParser#parseToBof()}) before
	they are added to this class. The {@link #parse(EventHandler)} method will then
	call {@link AnnotationParser#nextData()} and {@link DataParser#nextData()}
	for each parser as long as data is returned. All parsers must resturn at least
	as many data lines as the master parser or an exception is thrown. 

	@author Nicklas
	@version 2.15
	@base.modified $Date: 2010-04-20 10:02:16 +0200 (Tue, 20 Apr 2010) $
*/
public class SynchronizedSpotDataParser
	implements BfsParser
{

	/**
		Event type that is issued for each data line. The event data is an
		array with {@link SynchronizedData} objects. The first entry in the
		array contains information from the master parser, the following
		entries contain information for the data parsers in the order
		they were added.
	*/
	public static final EventType<SynchronizedData[]> DATA_EVENT = 
		new EventType<SynchronizedData[]>("synchonized.data");

	
	private AnnotationParser masterParser;
	private List<DataParser> dataParsers;
	
	/**
		Create a new synchronized data parser.
	*/
	public SynchronizedSpotDataParser()
	{
		this.dataParsers = new ArrayList<DataParser>();
	}
	
	/**
		Sets the master parser. This parser should normally parse a
		reporter annotations file.
	*/
	public void setMasterParser(AnnotationParser parser)
	{
		this.masterParser = parser;
	}
	
	/**
		Adds a data parser. The parser should parse a data file
		that has the same number of data lines as the master parser.
		Each line contains data identified by the corresponding 
		reporter in the master parser file.
	*/
	public void addDataParser(DataParser parser)
	{
		this.dataParsers.add(parser);
	}
	
	/**
		@return Always null, since we are parsing multiple files
	*/
	@Override
	public String getFilename()
	{
		return null;
	}

	/**
		@return The file size of the master parser
	*/
	@Override
	public long getFileSize()
	{
		return masterParser.getFileSize();
	}
	
	/**
		@return The current line number in the file parsed by the 
			master parser
	*/
	@Override
	public int getCurrentLine()
	{
		return masterParser.getCurrentLine();
	}
	
	/**
		@return The number of bytes parsed by the master parser
	*/
	@Override
	public long getParsedBytes()
	{
		return masterParser.getParsedBytes();
	}
	
	@Override
	public void close() 
	{
		if (masterParser != null) masterParser.close();
		for (DataParser parser : dataParsers)
		{
			parser.close();
		}
	}

	/**
		Begin parsing and issue events to the given event handler. For each
		data line a {@link #DATA_EVENT} is sent to the event handler. When 
		the end-of-file has been a {@link DataParser#END_OF_FILE_EVENT} is
		sent.
		
		@param handler An event handler
		@throws IOException If there is an error reading the stream data
		@throws NullPointerException If the event handler or master parser is null
	*/
	public void parse(EventHandler handler)
		throws IOException
	{
		if (handler == null) throw new NullPointerException("handler");
		if (masterParser == null) throw new NullPointerException("masterParser");
		
		SynchronizedData[] data = new SynchronizedData[1+dataParsers.size()];
		data[0] = new SynchronizedData(masterParser);
		int i = 1;
		for (DataParser parser : dataParsers)
		{
			data[i] = new SynchronizedData(parser);
			++i;
		}
		
		String[] tmp = masterParser.nextData();
		while (tmp != null)
		{
			ThreadSignalHandler.checkInterrupted();
			data[0].setData(tmp);
			i = 1;
			for (DataParser parser : dataParsers)
			{
				tmp = parser.nextData();
				if (tmp == null)
				{
					throw new IOException("Unexpected end of file at line " + 
						parser.getCurrentLine() + " in file " + parser.getFilename());
				}
				data[i].setData(tmp);
				++i;
			}
			handler.handleEvent(DATA_EVENT, data, this);
			tmp = masterParser.nextData();
		}
		handler.handleEvent(DataParser.END_OF_FILE_EVENT, null, this);
	}
	
	/**
		This class holds information that is sent to the event
		handler when {@link SynchronizedSpotDataParser#DATA_EVENT}:s are
		sent. Each instance contains information about the underlying 
		parser (master parser or data parser) and the data on the
		line that is currently being processed.
		NOTE! There is usually only one instance of this class for 
		each parser. If the event handler needs to collect information
		accross multiple lines the data should be copied to the event
		handler.
	*/
	public static class SynchronizedData
	{
		private final BfsParser parser;
		private String[] data;
	
		SynchronizedData(BfsParser parser)
		{
			this.parser = parser;
		}
		
		void setData(String[] data)
		{
			this.data = data;
		}
		
		/**
			Get the current data line, split into columns.
		*/
		public String[] getData()
		{
			return data;
		}
		
		/**
			The underlying parser.
		*/
		public BfsParser getParser()
		{
			return parser;
		}
	}
	
}
