/**
	$Id: SSLUtil.java 5362 2010-06-16 12:28:36Z nicklas $

	Copyright (C) 2010 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.ssl;

import java.io.ByteArrayInputStream;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.NoSuchProviderException;
import java.security.UnrecoverableKeyException;
import java.security.cert.Certificate;
import java.security.cert.CertificateException;
import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;

import javax.net.ssl.HttpsURLConnection;
import javax.net.ssl.KeyManager;
import javax.net.ssl.KeyManagerFactory;
import javax.net.ssl.SSLContext;
//import javax.net.ssl.SSLSocketFactory;
import javax.net.ssl.TrustManager;
import javax.net.ssl.TrustManagerFactory;

import org.apache.http.conn.ssl.SSLSocketFactory;

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.Config;

/**
	Utility class for working with SSL connections.
	
	@author Nicklas
	@since 2.16
	@base.modified $Date: 2010-06-16 14:28:36 +0200 (Wed, 16 Jun 2010) $
*/
public class SSLUtil
{

	private static KeyManager[] defaultKeyManagers;
	private static TrustManager[] defaultTrustManagers;
	private static SSLSocketFactory defaultSocketFactory;

	/**
		Get a SSL Socket Factory object that is used to create SSL sockets.
		Typical usage is:
		<pre class="code">
URL u = ... // A https url
URLConnection conn = u.openConnection();
if (conn instanceof HttpsURLConnection)
{
  ((HttpsURLConnection)conn).setSSLSocketFactory(SSLUtil.getSSLSocketFactory());
}
</pre>
		
		Unless a key-store and/or trust-store has been configured this method returns a
		default socket factory provided by {@link HttpsURLConnection#getDefaultSSLSocketFactory()}.
		<p>
		A key-store is used to store personal certificates that are used for 
		authentication with the server. A key-store can be setup with the 
		following configuration options in base.config.
		
		<ul>
		<li>ssl.keystore.file: The full path to a file containing certificate keys.
		<li>ssl.keystore.password: The password that is required to unlock the certificates. 
			All certificates must use the same password.
		<li>ssl.keystore.type (optional): The type of the keystore. The default value is 'JKS'.
		<li>ssl.keystore.provider (optional): The cryptographic provider implementation to use.
			If not specified the list of registered providers is searched.
		<li>ssl.keystore.algorithm (optional): The encryption algorithm used in the keystore.
			If not specified, 'SunX509' is used.
		</ul>
		
		<p>
		A trust-store is used to store public certificates of of servers that are trusted.
		The default trust-store uses the certificates that are shipped with the java runtime
		in $JAVA_HOME/jre/lib/security/cacerts. A trust-store can be setup with the following 
		configuration options in base.config. 
		
		<ul>
		<li>ssl.truststore.file: The full path to a file containing the certificates.
		<li>ssl.truststore.password: The password that is required to unlock the certificates. 
			All certificates must use the same password.
		<li>ssl.truststore.type (optional): The type of the keystore. The default value is 'JKS'.
		<li>ssl.truststore.provider (optional): The cryptographic provider implementation to use.
			If not specified the list of registered providers is searched.
		<li>ssl.truststore.algorithm (optional): The encryption algorithm used in the keystore.
			If not specified, 'PKIX' is used.
		</ul>
		
		For more information about algorithms, keystore types, etc. see 
		<a href="http://java.sun.com/javase/6/docs/technotes/guides/security/jsse/JSSERefGuide.html">
		Java Secure Socket Extension Reference Guide</a>
		
		@return A SSL socket factory
	*/
	public static SSLSocketFactory getSSLSocketFactory()
	{
		if (defaultSocketFactory == null)
		{
			String sslProtocol = Config.getString("ssl.context.protocol", "TLS");
			String sslProvider = Config.getString("ssl.context.provider");
		
			try
			{
				KeyManager[] keyManagers = getDefaultKeyManagers();
				TrustManager[] trustManagers = getDefaultTrustManagers();
	
				if (keyManagers != null || trustManagers != null)
				{
					SSLContext context = sslProvider == null ? 
							SSLContext.getInstance(sslProtocol) : 
							SSLContext.getInstance(sslProtocol, sslProvider);
					context.init(keyManagers, trustManagers, null);
					defaultSocketFactory = new SSLSocketFactory(context);
				}
				else
				{
					defaultSocketFactory = new SSLSocketFactory(SSLContext.getDefault());
				}
			}
			catch (Exception ex)
			{
				throw new BaseException(ex);
			}
		}
		return defaultSocketFactory;
	}

	/**
		Create a SSL socket factory that optionally uses the given trusted server certificate
		and client certificate. If both parameters are null then the socket factory from 
		{@link #getSSLSocketFactory()} is returned.
		
		@param serverCertificate The trusted server certificate data or null to use
			the default trusted certificates (see {@link #getSSLSocketFactory()}. The
			data must represent a X.509 certificate in binary or base64-encoded DER
			format
		@param clientCertificate The client certificate to use for authorization with
			the server or null to use the default certificates. The data must represent
			a PKCS #12 cretificate in binary form
		@param clientCertificatePassword A password used to unlock the client
			certificate
		@return A socket factory
	*/
	public static SSLSocketFactory getSSLSocketFactory(byte[] serverCertificate, 
			byte[] clientCertificate, String clientCertificatePassword)
	{
		if (serverCertificate == null && clientCertificate == null)
		{
			return getSSLSocketFactory();
		}

		String sslProtocol = Config.getString("ssl.context.protocol", "TLS");
		String sslProvider = Config.getString("ssl.context.provider");
		SSLSocketFactory factory = null;
		try
		{
			KeyManager[] keyManagers = null;
			if (clientCertificate != null)
			{
				String provider = Config.getString("ssl.keystore.provider");
				String algorithm = Config.getString("ssl.keystore.algorithm", "SunX509");
				KeyStore keyStore = createKeyStore(new ByteArrayInputStream(clientCertificate), clientCertificatePassword, 
						"PKCS12", provider);
				keyManagers = new KeyManager[] { 
						new StaticKeyManager(keyStore, null, clientCertificatePassword) 
					};
			}
			else
			{
				keyManagers = getDefaultKeyManagers();
			}
			TrustManager[] trustManagers = null;
			if (serverCertificate != null)
			{
				String trustStoreType = Config.getString("ssl.truststore.type", "JKS");
				String trustStoreProvider = Config.getString("ssl.truststore.provider");
				String trustStoreAlgorithm = Config.getString("ssl.truststore.algorithm", "PKIX");
				Certificate cert = getCertificate(new ByteArrayInputStream(serverCertificate), "X.509", trustStoreProvider);
				KeyStore keyStore = createKeyStore(null, null, trustStoreType, trustStoreProvider);
				keyStore.setCertificateEntry("tmp", cert);
				trustManagers = createTrustManagers(keyStore, trustStoreProvider, trustStoreAlgorithm);
			}
			else
			{
				trustManagers = getDefaultTrustManagers();
			}
			
			if (keyManagers != null || trustManagers != null)
			{
				SSLContext context = sslProvider == null ? 
						SSLContext.getInstance(sslProtocol) : 
						SSLContext.getInstance(sslProtocol, sslProvider);
				context.init(keyManagers, trustManagers, null);
				factory = new SSLSocketFactory(context);
				factory.setHostnameVerifier(new AllowAnyHostNameVerifyer());
			}
			else
			{
				factory = new SSLSocketFactory(SSLContext.getDefault());
			}
		}
		catch (Exception ex)
		{
			throw new BaseException(ex);
		}

		return factory;

	}

	/**
		Get the default trust manages as configured in the base.config file by
		ssl.truststore.* settings.
		@return An array with trust managers or null if no trust-store file has been
			configured
	*/
	private static TrustManager[] getDefaultTrustManagers() 
		throws KeyStoreException, NoSuchAlgorithmException, CertificateException,
		NoSuchProviderException, IOException
	{
		if (defaultTrustManagers == null)
		{
			String file = Config.getString("ssl.truststore.file");
			String password = Config.getString("ssl.truststore.password");
			String type = Config.getString("ssl.truststore.type", "JKS");
			String provider = Config.getString("ssl.truststore.provider");
			String algorithm = Config.getString("ssl.truststore.algorithm", "PKIX");
			
			if (file == null)
			{
				defaultTrustManagers = new TrustManager[0];
			}
			else
			{
				KeyStore keyStore = createKeyStore(new FileInputStream(file), password, type, provider);
				createTrustManagers(keyStore, provider, algorithm);
			}
		
		}
		return defaultTrustManagers.length == 0 ? null : defaultTrustManagers;
	}

	/**
		Get the default keys manages as configured in the base.config file by
		ssl.keystore.* settings.
		@return An array with key managers or null if no key-store file has been
			configured
	*/
	private static KeyManager[] getDefaultKeyManagers() 
		throws UnrecoverableKeyException, KeyStoreException, NoSuchAlgorithmException, 
		CertificateException, NoSuchProviderException, IOException
	{
		if (defaultKeyManagers == null)
		{
			String file = Config.getString("ssl.keystore.file");
			String password = Config.getString("ssl.keystore.password");
			String type = Config.getString("ssl.keystore.type", "JKS");
			String provider = Config.getString("ssl.keystore.provider");
			String algorithm = Config.getString("ssl.keystore.algorithm", "SunX509");
			
			if (file == null)
			{
				defaultKeyManagers = new KeyManager[0];
			}
			else
			{
				KeyStore keyStore = createKeyStore(new FileInputStream(file), password, type, provider);
				defaultKeyManagers = createKeyManagers(keyStore, password, provider, algorithm);	
			}
		}
		return defaultKeyManagers.length == 0 ? null : defaultKeyManagers;
	}

	/**
		Create a key-store from the given input stream
		@param store The stream with key-store data
		@param password The password to unlock the keystore
		@return A KeyStore object
	*/
	public static KeyStore createKeyStore(InputStream store, String password, String type, String provider) 
		throws KeyStoreException, NoSuchProviderException, NoSuchAlgorithmException, CertificateException, IOException
	{
		char[] pw = password == null ? null : password.toCharArray();
		KeyStore keyStore = provider == null ? 
				KeyStore.getInstance(type) : 
				KeyStore.getInstance(type, provider);
		keyStore.load(store, pw);
		return keyStore;
	}

	/**
		Get the certificate that is found in the given input stream.
		
		@param cert The stream with the certificate information
		@param type The type of the certificate, usually "X.509"
		@return A Certificate object (can usually be subclassed to {@link X509Certificate}
	*/
	public static Certificate getCertificate(InputStream cert, String type, String provider) 
		throws CertificateException, NoSuchProviderException
	{
		CertificateFactory factory = provider == null ? 
				CertificateFactory.getInstance(type) :
				CertificateFactory.getInstance(type, provider);
		return factory.generateCertificate(cert);
	}
	
	private static KeyManager[] createKeyManagers(KeyStore keyStore, String password, String provider, String algorithm) 
		throws KeyStoreException, NoSuchAlgorithmException, CertificateException, 
			IOException, UnrecoverableKeyException, NoSuchProviderException
	{
		char[] pw = password == null ? null : password.toCharArray();
		
		// Create key manager factory
		KeyManagerFactory keyManagerFactory = provider == null ? 
				KeyManagerFactory.getInstance(algorithm) : 
				KeyManagerFactory.getInstance(algorithm, provider);
		keyManagerFactory.init(keyStore, pw);
		return keyManagerFactory.getKeyManagers();
	}
	
	private static TrustManager[] createTrustManagers(KeyStore keyStore, String provider, String algorithm) 
		throws KeyStoreException, NoSuchAlgorithmException, CertificateException, 
			IOException, NoSuchProviderException 
	{
		// Create trust manager factory
		TrustManagerFactory trustManagerFactory = provider == null ?
				TrustManagerFactory.getInstance(algorithm) :
				TrustManagerFactory.getInstance(algorithm, provider);
		trustManagerFactory.init(keyStore);
		return trustManagerFactory.getTrustManagers();
	}

	
}
