/*
 $Id: Base1PluginExecuter.java 5405 2010-09-10 11:09:05Z nicklas $

 Copyright (C) 2006 Johan Enell, Jari Häkkinen, Nicklas Nordborg
 Copyright (C) 2007 Johan Enell, Nicklas Nordborg
 Copyright (C) 2008 Martin Svensson
 
 This file is part of BASE - BioArray Software Environment.
 Available at http://base.thep.lu.se/

 BASE is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 3
 of the License, or (at your option) any later version.

 BASE is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/

package net.sf.basedb.plugins;

import net.sf.basedb.core.AnnotationType;
import net.sf.basedb.core.AnyToAny;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.BioAssaySet;
import net.sf.basedb.core.BooleanParameterType;
import net.sf.basedb.core.Config;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Directory;
import net.sf.basedb.core.Experiment;
import net.sf.basedb.core.ExtraValue;
import net.sf.basedb.core.File;
import net.sf.basedb.core.FileParameterType;
import net.sf.basedb.core.FloatParameterType;
import net.sf.basedb.core.Include;
import net.sf.basedb.core.IntegerParameterType;
import net.sf.basedb.core.IntensityTransform;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemNotFoundException;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.Job;
import net.sf.basedb.core.ParameterType;
import net.sf.basedb.core.Path;
import net.sf.basedb.core.PathParameterType;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.PluginConfiguration;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.ProgressReporter;
import net.sf.basedb.core.RequestInformation;
import net.sf.basedb.core.StringParameterType;
import net.sf.basedb.core.Transformation;
import net.sf.basedb.core.User;
import net.sf.basedb.core.Version;
import net.sf.basedb.core.plugin.About;
import net.sf.basedb.core.plugin.AboutImpl;
import net.sf.basedb.core.plugin.AbstractAnalysisPlugin;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.core.plugin.InteractivePlugin;
import net.sf.basedb.core.plugin.Permissions;
import net.sf.basedb.core.plugin.Request;
import net.sf.basedb.core.plugin.Response;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Orders;
import net.sf.basedb.core.signal.SignalException;
import net.sf.basedb.core.signal.SignalHandler;
import net.sf.basedb.core.signal.SignalTarget;
import net.sf.basedb.core.signal.ThreadSignalHandler;
import net.sf.basedb.util.ChainedProgressReporter;
import net.sf.basedb.util.Enumeration;
import net.sf.basedb.util.FileCopyRunnable;
import net.sf.basedb.util.FileUtil;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.XMLUtil;
import net.sf.basedb.util.basefile.BaseFileWriter;
import net.sf.basedb.util.export.spotdata.Base1FieldConverter;
import net.sf.basedb.util.export.spotdata.ExportableFieldConverter;
import net.sf.basedb.util.export.spotdata.BaseFileExporter;
import net.sf.basedb.util.export.spotdata.DynamicField;
import net.sf.basedb.util.export.spotdata.ExportableFieldFactory;
import net.sf.basedb.util.export.spotdata.MatrixBaseFileExporter;
import net.sf.basedb.util.export.spotdata.SerialBaseFileExporter;
import net.sf.basedb.util.formatter.DateFormatter;
import net.sf.basedb.util.importer.BaseFileWrapper;
import net.sf.basedb.util.importer.spotdata.BaseFileImporter;
import net.sf.basedb.util.parser.FlatFileParser;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.filter.ElementFilter;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;

import java.io.ByteArrayOutputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.net.URL;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Date;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.regex.Pattern;

/**
	This plug-in executes a plug-in from the Base1 system. It takes a base1 configuration
	file to create its configuration.
	
	@author Johan Enell
	@version 2.2
	@base.modified $Date: 2010-09-10 13:09:05 +0200 (Fri, 10 Sep 2010) $
*/
public class Base1PluginExecuter 
	extends AbstractAnalysisPlugin 
	implements InteractivePlugin, SignalTarget
{

	/**
		Plugin description
	*/
	private static final About about = new AboutImpl(
		"Base1PluginExecuter",
		"This plug-in simulates the plug-in runner from Base 1.2 to enable old plug-ins to be executed.",
		Version.getMajor() + "." + Version.getMinor() + "." + Version.getMaintenance(),
		"2006, Base 2 development team",
		null,
		null,
		"http://base.thep.lu.se");
 
	/**
		The plug-in permissions.
	*/
	private static final Set<Permissions> permissions = new HashSet<Permissions>();
	
	private static final URL dtdFile = Base1PluginExecuter.class.getResource("/net/sf/basedb/core/dtd/base1-plugin-configuration-file.dtd");
	
	/* Configuration parameters*/

	/**
		Parameter definition that asks for the configuration file that should be imported.
	*/
	private static final PluginParameter<File> fileParameter = new PluginParameter<File>(
		"file",
		"File",
		"The file to import the data from. If you leave this parameter blank you can manually configure the plug-in.",
		new FileParameterType(null, false, 1));

	/**
		The path to where the plug-in executer can find the executables. In Base1 this was 
		default to /usr/local/base/plugins.
	*/
	private static PluginParameter<String> execPathParameter = new PluginParameter<String>(
		"execPath",
		"Plugin executables path",
		"The path to the executables",
		new StringParameterType(255, null, true));

	/**
		This parameter describes if the data should be exported in serial format or
		in matrix.
	*/
	private PluginParameter<Boolean> serialFormatParameter;
	
	/**
		The version number of the plug-in.
	*/
	private PluginParameter<String> versionNumberParameter;
	
	/**
		The name of the executable.
	*/
	private PluginParameter<String> execNameParameter;
	
	/**
		If the data should be exported averaged per gene.
	*/
	private PluginParameter<Boolean> geneAveragesParameter;
	
	/**
		The minimum number of channels. 
	*/
	private PluginParameter<Integer> minChannelsParameter;
	
	/**
		The maximum number of channels, 0 means unlimited.
	*/
	private PluginParameter<Integer> maxChannelsParameter;
	
	/**
		If the plug-in should leave the stdin.txt after execution.
	*/
	private PluginParameter<Boolean> leaveStdinParameter;
	
	/**
		If the plug-in should leave the stdout.txt after execution.
	*/
	private PluginParameter<Boolean> leaveStdoutParameter;
	
	/**
		The reporter fields that should be exported as a tab separated string
		of column headers.
	*/
	private PluginParameter<String> usedColumnsParameter;
	
	/**
		The spot fields that should be exported as a tab separated string of 
		column headers.
	*/
	private PluginParameter<String> usedFieldsParameter;
	
	/**
		This parameter describes the job parameters that the base1 plugin needs. 
	*/
	private PluginParameter<String> jobParametersParameter = new PluginParameter<String>(
		"jobParameters",
		"Job parameters",
		"An XML text that describes the job parameters.",
		new StringParameterType(null, null, true));
	
	/* Job parameters*/
	/**
		The directory where the plug-in will store it's output.
	*/
	private PluginParameter<String> pluginDirectoryParameter;

	/**
		This command is used if the user decides to configure the plugin by hand.
	*/
	private static final String COMMAND_MANUAL_CONFIG = "COMMAND_MANUAL_CONFIG";

	/**
		Holds info about a job configuration request.
	*/
	private RequestInformation configureJob;
	
	/**
		Holds information about a plugin configuration request.
	*/
	private RequestInformation configurePlugin;
	
	/**
		Holds information about a manual plugin configuration request.
	*/
	private RequestInformation configureManual;

	/**
		This list holds the job parameters derived from the configuration file.
	*/
	private List<PluginParameter<?>> jobParameters;

	/**
		This is the directory where the plugin will generate its output.
	*/
	private java.io.File execDirectory;
	
	private ThreadSignalHandler signalHandler;

	/* From InteractivePlugin interface*/

	/**
		Checks if the BioAssaySet has the number of channels specified by the 
		configuration and that it has the extra values.
		
		@param context The context from where the plugin is invoked.
		@param item The item that should be checked if it is in context.
		@return null if the item is in context, otherwise an error message.
		@see net.sf.basedb.core.plugin.AbstractAnalysisPlugin#isInContext(net.sf.basedb.core.plugin.GuiContext, java.lang.Object)
	*/
	@Override
	public String isInContext(GuiContext context, Object item)
	{
		String message = super.isInContext(context, item);
		if (message == null)
		{
			BioAssaySet bas = (BioAssaySet)item;
			if (bas.getNumSpots() <= 0)
			{
				message = "This plug-in requires spot data in the database";
			}
			else
			{
				DbControl dc = null;
				try
				{
					dc = sc.newDbControl();
					//getManualConfigureParameters();
	
					// Check number of channels
					int channels = bas.getRawDataType().getChannels();
					int maxChannels = Values.getInt(String.valueOf(configuration.getValue("maxChannels")), 0);
					int minChannels = Values.getInt(String.valueOf(configuration.getValue("minChannels")), 0);
					if (channels < minChannels || ( maxChannels != 0 && channels > maxChannels))
					{
						return configuration.getPluginConfiguration().getName() + " can't work on " + channels + "-channel data; min=" + minChannels + "; max=" + maxChannels;
					}
					
					// Check source data intensity transform
					IntensityTransform transform = bas.getIntensityTransform();
					String requiredTransform = (String)configuration.getValue("sourceTransform");
					if (requiredTransform != null)
					{
						if (!requiredTransform.equals(transform.name()))
						{
							String transformed =  transform.isTransformed() ? 
									transform.toString() + "-transformed" : "untransformed";
							return configuration.getPluginConfiguration().getName() + 
								" can't work on " + transformed + " data";
						}
					}
					List<String> extraValues = new ArrayList<String>();
					for (ExtraValue ev : bas.getExtraValues().list(dc))
					{
						extraValues.add(ev.getExtraValueType().getExternalId());
					}
					String[] usedFields = String.valueOf(configuration.getValue("usedFields")).split("\\\\t");
					for (String field : usedFields)
					{
						if (field.startsWith("_xc_") && !extraValues.contains(field.substring(4)))
						{
							return "Requires extra value '"+ field.substring(4) +"' attached to BioAssaySet";
						}
					}
				}
				finally
				{
					if (dc != null) dc.close();
				}
			}
		}
		return message;
	}

	public RequestInformation getRequestInformation(GuiContext context, String command)
		throws BaseException
	{
		RequestInformation requestInformation = null;

		if (command.equals(Request.COMMAND_CONFIGURE_PLUGIN))
		{
			requestInformation = getConfigurePluginParameters();
		}
		else if (command.equals(Base1PluginExecuter.COMMAND_MANUAL_CONFIG))
		{
			requestInformation = getManualConfigureParameters();
		}
		else if (command.equals(Request.COMMAND_CONFIGURE_JOB))
		{
			requestInformation = getConfigureJobParameters();
		}
		else
		{
			throw new BaseException("Unknown command: "+command);
		}
		return requestInformation;
	}
	
	
	public void configure(GuiContext context, Request request, Response response)
	{
		String command = request.getCommand();
		DbControl dc = null;
		try
		{
			if (command.equals(Request.COMMAND_CONFIGURE_PLUGIN))
			{
				RequestInformation ri = getConfigurePluginParameters();
				List<Throwable> errors = validateRequestParameters(ri.getParameters(), request);
				if (errors != null)
				{
					response.setError(errors.size() + " invalid parameter(s) were found in the request", errors);
					return;
				}
				storeValue(configuration, request, execPathParameter);
				storeValue(configuration, request, ri.getParameter("sourceTransform"));
				storeValue(configuration, request, ri.getParameter("resultTransform"));
				
				dc = sc.newDbControl();
				File f = (File) request.getParameterValue(fileParameter.getName());
				if (f != null)
				{
					if (f.isInDatabase()) f = File.getById(dc, f.getId());
					
					getManualConfigureParameters();
					Map<String, String> headers = getParametersFromBaseFile(f);
					configuration.setValue(serialFormatParameter.getName(), serialFormatParameter.getParameterType(), getBoolean(headers.get(serialFormatParameter.getName())));
					configuration.setValue(versionNumberParameter.getName(), versionNumberParameter.getParameterType(), headers.get(versionNumberParameter.getName()));
					configuration.setValue(execNameParameter.getName(), execNameParameter.getParameterType(), headers.get(execNameParameter.getName()));
					configuration.setValue(geneAveragesParameter.getName(), geneAveragesParameter.getParameterType(), getBoolean(headers.get(geneAveragesParameter.getName())));
					configuration.setValue(minChannelsParameter.getName(), minChannelsParameter.getParameterType(), new Integer(headers.get(minChannelsParameter.getName())));
					configuration.setValue(maxChannelsParameter.getName(), maxChannelsParameter.getParameterType(), new Integer(headers.get(maxChannelsParameter.getName())));
					configuration.setValue(leaveStdinParameter.getName(), leaveStdinParameter.getParameterType(), getBoolean(headers.get(leaveStdinParameter.getName())));
					configuration.setValue(leaveStdoutParameter.getName(), leaveStdoutParameter.getParameterType(), getBoolean(headers.get(leaveStdoutParameter.getName())));
					configuration.setValue(usedColumnsParameter.getName(), usedColumnsParameter.getParameterType(), headers.get(usedColumnsParameter.getName()));
					configuration.setValue(usedFieldsParameter.getName(), usedFieldsParameter.getParameterType(), headers.get(usedFieldsParameter.getName()));
					configuration.setValue(jobParametersParameter.getName(), jobParametersParameter.getParameterType(), headers.get(jobParametersParameter.getName()));
					PluginConfiguration config = configuration.getPluginConfiguration();
					if (config != null)
					{
						if (config.isInDatabase()) config = PluginConfiguration.getById(dc,config.getId());
						config.setDescription(headers.get("descr").replaceAll("\\\\r\\\\n|\\\\n", "\n"));
					}
					dc.commit();
					response.setDone("Configured plug-in with file.");
				}
				else
				{
					response.setContinue(Base1PluginExecuter.COMMAND_MANUAL_CONFIG);
				}
			}
			else if (command.equals(Base1PluginExecuter.COMMAND_MANUAL_CONFIG))
			{
				RequestInformation ri = getManualConfigureParameters();
				List<Throwable> errors = validateRequestParameters(ri.getParameters(), request);
				if (errors != null)
				{
					response.setError(errors.size() + " invalid parameter(s) were found in the request", errors);
					return;
				}
				
				storeValue(configuration, request, serialFormatParameter);
				storeValue(configuration, request, versionNumberParameter);
				storeValue(configuration, request, execNameParameter);
				storeValue(configuration, request, geneAveragesParameter);
				storeValue(configuration, request, minChannelsParameter);
				storeValue(configuration, request, maxChannelsParameter);
				storeValue(configuration, request, leaveStdinParameter);
				storeValue(configuration, request, leaveStdoutParameter);
				storeValue(configuration, request, ri.getParameter("overrideLeaveStd"));
				storeValue(configuration, request, usedColumnsParameter);
				storeValue(configuration, request, usedFieldsParameter);
				storeValue(configuration, request, jobParametersParameter);
				// test the xml in the jobParameters parameters
				try
				{
					getJobParametersFromXML(String.valueOf(configuration.getValue(jobParametersParameter.getName())), false);
				}
				catch (BaseException e)
				{
					throw new BaseException("Error in the jobparameters xml. "+e.getMessage());
				}
				response.setDone("Plugin configuration done.");
			}
			else if (command.equals(Request.COMMAND_CONFIGURE_JOB))
			{
				RequestInformation ri = getConfigureJobParameters();
				List<Throwable> errors = validateRequestParameters(ri.getParameters(), request);
				if (errors != null)
				{
					response.setError(errors.size() + " invalid parameter(s) were found in the request", errors);
					return;
				}
				
				storeValue(job, request, ri.getParameter(SOURCE_BIOASSAYSET));
				storeValue(job, request, ri.getParameter("pluginDirectory"));
				storeValue(job, request, ri.getParameter(CHILD_NAME));
				storeValue(job, request, ri.getParameter(CHILD_DESCRIPTION));
				storeValue(job, request, ri.getParameter("leaveStdin"));
				storeValue(job, request, ri.getParameter("leaveStdout"));
				for (PluginParameter<?> p : getJobParameters())
				{
					ParameterType<?> pType = p.getParameterType();
					Object value = request.getParameterValue(p.getName());
					if (pType.isEnumeration() && value != null)
					{
						// The 'key' value is sent in the request, but we
						// want to store the 'value' value instead
						Enumeration<?, String> e = pType.getEnumeration();
						for (int i = 0; i < e.size(); ++i)
						{
							if (value.equals(e.getKey(i))) 
							{
								value = e.getValue(i);
								break;
							}
						}
						ParameterType<String> tmp = new StringParameterType(65535, null, false);
						job.setValue(p.getName(), tmp, (String)value);
					}
					else
					{
						storeValue(job, request, p);
					}
				}
				response.setDone("Job configuration complete", Job.ExecutionTime.SHORT);
			}
			else
			{
				throw new BaseException("Command unknown: "+command);
			}
		}
		catch (Throwable ex)
		{
			response.setError(ex.getMessage(), Arrays.asList(ex));
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
	
	/**
		Copies the content from a File to a directory on the file system.
		
		@param f The file to be copied.
		@param d The directory where the file will be written.
		@throws IOException If there is any error with target.
	*/
	private void copy(File f, java.io.File d)
		throws IOException
	{
		if (f != null && d.isDirectory())
		{
			java.io.File target = new java.io.File(d, f.getName());
			InputStream in = null;
			OutputStream out = null;
			try
			{
				in = f.getDownloadStream(0);
				out = FileUtil.getOutputStream(target);
				FileUtil.copy(in, out);
			}
			finally
			{
				FileUtil.close(in);
				FileUtil.close(out);
			}
		}
	}
	
	/**
		Deletes all files and sub directories under dir.
		
		@param dir
		@return true if all deletions were successful
	*/
	private boolean deleteDir(java.io.File dir)
	{
		if (dir.isDirectory())
		{
			java.io.File[] children = dir.listFiles();
			for (java.io.File child : children)
			{
				boolean success = deleteDir(child);
				if (!success)
				{
					return false;
				}
			}
		}
		return dir.delete();
	}
	
	@Override
	public Collection<Permissions> getPermissions()
	{
		if (permissions.size() == 0)
		{
			permissions.add(new Permissions(Item.ANNOTATIONTYPE, null, EnumSet.of(Permission.READ)));
			permissions.add(new Permissions(Item.DIRECTORY, null, EnumSet.of(Permission.CREATE, Permission.DELETE)));
			permissions.add(new Permissions(Item.EXPERIMENT, null, EnumSet.of(Permission.WRITE)));
			permissions.add(new Permissions(Item.FILE, null, EnumSet.of(Permission.CREATE, Permission.DELETE)));
			permissions.add(new Permissions(Item.PLUGINCONFIGURATION, EnumSet.of(Permission.WRITE), null));
			permissions.add(new Permissions(Item.RAWBIOASSAY, null, EnumSet.of(Permission.READ)));
			permissions.add(new Permissions(Item.REPORTER, null, EnumSet.of(Permission.READ)));
			permissions.add(new Permissions(Item.EXTRAVALUETYPE, null, EnumSet.of(Permission.CREATE, Permission.USE)));
			permissions.add(new Permissions(Item.USER, null, EnumSet.of(Permission.READ)));
			permissions.add(new Permissions(Item.JOB, null, EnumSet.of(Permission.READ)));
			permissions.add(new Permissions(Item.FORMULA, null, EnumSet.of(Permission.READ)));
			permissions.add(new Permissions(Item.REPORTERLIST, null, EnumSet.of(Permission.CREATE, Permission.USE)));
		}
		return permissions;
	}
	
	// ***

	// *** From Plugin interface

	public About getAbout()
	{
		return about;
	}
	

	@Override
	public boolean supportsConfigurations()
	{
		return true;
	}
	

	@Override
	public boolean requiresConfiguration()
	{
		return true;
	}
	

	public void run(Request request, Response response, ProgressReporter progress)
	{
		if (signalHandler != null) signalHandler.setWorkerThread(null);
		ChainedProgressReporter chainedProgress = progress == null ? 
				null : new ChainedProgressReporter(progress);;
		String command = request.getCommand();
		if (command.equals(Request.COMMAND_EXECUTE))
		{
			String msg = "";
			DbControl dc = null;
			File stdin = null;
			File stdout = null;
			Transformation trans = null; 
			Directory workingDirectory = null;
			java.io.File tmpDir = null;

			boolean overrideLeaveStd = Boolean.TRUE.equals(configuration.getValue("overrideLeaveStd"));
			boolean leaveStdin = Boolean.TRUE.equals(configuration.getValue("leaveStdin"));
			boolean leaveStdout = Boolean.TRUE.equals(configuration.getValue("leaveStdout"));
			if (overrideLeaveStd)
			{
				if (job.getValue("leaveStdin") != null) 
				{
					leaveStdin = Boolean.TRUE.equals(job.getValue("leaveStdin"));
				}
				if (job.getValue("leaveStdout") != null) 
				{
					leaveStdout = Boolean.TRUE.equals(job.getValue("leaveStdout"));
				}
			}
			
			String childName = (String)job.getValue(CHILD_NAME);
			String childDescription = (String)job.getValue(CHILD_DESCRIPTION);
			IntensityTransform resultTransform = null;
			if (configuration.getValue("resultTransform") != null)
			{
				resultTransform = IntensityTransform.valueOf((String)configuration.getValue("resultTransform"));
			}

			try
			{
				//Export
				try
				{
					dc = sc.newDbControl();
					String workingPath = (String) job.getValue("pluginDirectory") + "/" + job.getId();
					workingDirectory = getPluginDirectory(dc, workingPath);
					progress.display(0, "Exporting data to be used by plugin.");
					if (chainedProgress != null) chainedProgress.setRange(0, 20);
					stdin = exportData(dc, workingDirectory, chainedProgress);
					ThreadSignalHandler.checkInterrupted();
					dc.commit();
				}
				catch(Exception e)
				{
					response.setError("Error during export: " + e.getMessage(), Arrays.asList(e));
					return;
				}
				finally
				{
					if (dc != null) dc.close();
				}
				
				//Execution
				// This thread group is for thread readin/writing stream to program
				ThreadGroup fileCopyGroup = new ThreadGroup("File copy");		
				Process externalProcess = null;
				try
				{
					ThreadSignalHandler.checkInterrupted();
					tmpDir = getExecDirectory();
					copy(stdin, tmpDir);
					String executable = (String)configuration.getValue("execName");
					progress.display(25, executable);
					externalProcess = Runtime.getRuntime().exec(getExecLine(), null, tmpDir);
	
					// Send stdin.txt to the program
					InputStream in = FileUtil.getInputStream(new java.io.File(tmpDir, "stdin.txt"));
					FileCopyRunnable inCopy = new FileCopyRunnable(in, externalProcess.getOutputStream());
					new Thread(fileCopyGroup, inCopy, "stdin").start();
					
					// Send output from the program to stdout.txt
					OutputStream out = FileUtil.getOutputStream(new java.io.File(tmpDir, "stdout.txt"));
					FileCopyRunnable outCopy = new FileCopyRunnable(externalProcess.getInputStream(), out);
					new Thread(fileCopyGroup, outCopy, "stdout").start();
					
					// Always read stderr into a byte[]
					ByteArrayOutputStream err = new ByteArrayOutputStream();
					FileCopyRunnable errCopy = new FileCopyRunnable(externalProcess.getErrorStream(), err);
					Thread errThread = new Thread(fileCopyGroup, errCopy, "stderr");
					errThread.start();

					int exitValue = 0;
					try
					{
						exitValue = externalProcess.waitFor();
					}
					catch (InterruptedException ex)
					{
						// Aborted by user
						externalProcess.destroy();
						throw new SignalException("Aborted by user", ex);
					}
	
					progress.display(50, "Reading from error stream.");
					for(int i = 0; errThread.isAlive(); ++i)
					{
						Thread.sleep(1000);
						if (i > 300) throw new BaseException("Waited 5 minutes for the errormessage. I give up.");
					}
										
					err.flush();
					msg = err.toString();
					out.flush();
					
					if (exitValue != 0)
					{
						response.setError(msg, null);
						return;
					}
				}
				catch (Exception e)
				{
					response.setError("Error during execution: " + e.getMessage(), Arrays.asList(e));
					return;
				}
				finally
				{
					fileCopyGroup.interrupt();
				}
				
				// Import analysis files created by the BASE1 plug-in
				try
				{
					//Copy files
					dc = sc.newDbControl();
					progress.display(55, "Copying output files to server.");
					importTempFiles(dc, tmpDir.listFiles(), workingDirectory);
					dc.commit();
				}
				finally
				{
					if (dc != null) dc.close();
				}
							
				//Import
				try
				{
					dc = sc.newDbControl();
					ThreadSignalHandler.checkInterrupted();
					progress.display(60, "Importing data from plugin.");
					BioAssaySet parentBas = getSourceBioAssaySet(dc);
					trans = parentBas.newTransformation(Job.getById(dc, job.getId()));
					trans.setName(PluginConfiguration.getById(dc, configuration.getId()).getName());
					dc.saveItem(trans);
					stdout = File.getFile(dc, workingDirectory, "stdout.txt", true);
					associateFiles(dc, trans, workingDirectory);
					if (chainedProgress != null) chainedProgress.setRange(65, 95);
					BioAssaySet child = importData(dc, stdout, trans, resultTransform, chainedProgress);
					if (child != null)
					{
						child.setName(childName);
						child.setDescription(childDescription);
					}
					dc.commit();
				}
				catch (Exception ex)
				{
					response.setError(ex.getMessage(), Arrays.asList(ex));
					return;
				}
				finally
				{
					if (dc != null) dc.close();
				}
			}
			finally
			{
				// Cleanup temporary working directory
				if (tmpDir != null) FileUtil.deleteTempDirectory(tmpDir);

				// Cleanup stdin and stdout
				if (workingDirectory != null && workingDirectory.isInDatabase())
				{
					try
					{
						dc = sc.newDbControl();
						if (!leaveStdin && File.exists(dc, workingDirectory, "stdin.txt"))
						{
							stdin = File.getFile(dc, workingDirectory, "stdin.txt", false);
							dc.deleteItem(stdin);
						}
						if (!leaveStdout && File.exists(dc, workingDirectory, "stdout.txt"))
						{
							stdout = File.getFile(dc, workingDirectory, "stdout.txt", false);
							dc.deleteItem(stdout);
						}
						dc.commit();
						dc = sc.newDbControl();
						// Remove the working directory if the plug-in didn't create
						// any files that should be saved
						workingDirectory = Directory.getById(dc, workingDirectory.getId());
						if (!workingDirectory.isUsed()) 
						{
							dc.deleteItem(workingDirectory);
						}
						dc.commit();
					}
					catch (Exception ex)
					{
						response.setError("Plug-in completed successfully, but working directory" +
								" and/or stdin.txt/stdout.txt could not be removed: " +
								ex.getMessage(), Arrays.asList(ex));
						return;
					}
					finally
					{
						if (dc != null) dc.close();
					}
				}
			}
			response.setDone(msg);
		}
		else
		{
			response.setError("Unknown command: " + command, null);
		}
	}
	/*
		From the SignalTarget interface
		-------------------------------------------
	*/
	public SignalHandler getSignalHandler()
	{
		signalHandler = new ThreadSignalHandler();
		return signalHandler;
	}
	// -------------------------------------------


	/**
		Export the bioassay set to a file called 'stdin.txt' in the plugin directory.
		@param dc DbControl used to access the database
		@param workingDirectory Where in base' file system the exported BASEFile should be stored.
		@return The file where the data is exported to.
	*/
	private File exportData(DbControl dc, Directory workingDirectory, ProgressReporter progress)
	{
		try
		{
			// Exporting data to 'stdin.txt' which is sent to the plug-in
			File file = File.getFile(dc, workingDirectory, "stdin.txt", true);
			file.setMimeTypeAuto("text/plain", null);
			dc.saveItem(file);
			BaseFileWriter out = new BaseFileWriter(
					new OutputStreamWriter(file.getUploadStream(false), Config.getCharset()));
			
			// Get configuration parameters
			BioAssaySet source = getSourceBioAssaySet(dc);
			boolean averageOnReporter = Boolean.TRUE.equals(configuration.getValue("geneAverages"));
			
			// Create the exporter
			BaseFileExporter exporter = null;
			if (Boolean.TRUE.equals(configuration.getValue("serialFormat")))
			{
				exporter = new SerialBaseFileExporter();
			}
			else
			{
				exporter = new MatrixBaseFileExporter();
			}
			exporter.setDbControl(dc);
			exporter.setSource(source);
			exporter.setProgressReporter(progress);
			exporter.setAverageOnReporter(averageOnReporter);
			exporter.setWriter(out);
			
			// Set job parameters
			for (PluginParameter<?> pp : getJobParametersFromXML((String)(configuration.getValue("jobParameters")), true))
			{
				String name = pp.getName();
				Object value = job.getValue(name);
				ParameterType<?> pType = pp.getParameterType();
				if (pType.isEnumeration() && value != null)
				{
					// The 'value' value is stored in the database, but
					// we need to export the 'key' value
					Enumeration<?, String> e = pType.getEnumeration();
					for (int i = 0; i < e.size(); ++i)
					{
						if (value.equals(e.getValue(i))) 
						{
							value = e.getKey(i);
							break;
						}
					}
				}
				exporter.setParameter(name, value == null ? "" : value.toString());
			}
			
			// Set reporter fields
			ExportableFieldConverter fieldConverter = new Base1FieldConverter(dc, source);
			String usedColumns = (String)configuration.getValue("usedColumns");
			if (usedColumns != null)
			{
				for (String reporterField : usedColumns.split("\\\\t"))
				{
					DynamicField fld = fieldConverter.getReporterField(reporterField, true);
					exporter.addReporterField(fld);
				}
			}
			
			// Set spot fields
			String usedFields = (String)configuration.getValue("usedFields");
			if (usedFields != null)
			{
				for (String spotField : usedFields.split("\\\\t"))
				{
					if ("_allIntensities".equals(spotField))
					{
						int channels = source.getRawDataType().getChannels();
						for (int i = 1; i <= channels; ++i)
						{
							exporter.addSpotField(ExportableFieldFactory.channel(i, "intensity" + i, null));
						}
					}
					else
					{
						DynamicField fld = fieldConverter.getSpotField(spotField, true);
						exporter.addSpotField(fld);
					}
				}
			}
			exporter.doExport();
			return file;
		}
		catch (IOException e)
		{
			throw new BaseException(e);
		}
	}
	
	/**
		Get the experimental factors connected to a experiment.
		
		@return A list of {@link AnnotationType}:s
	*/
	private List<AnnotationType> getAnnotationTypes()
	{
		DbControl dc = null;
		List<AnnotationType> list = new ArrayList<AnnotationType>();
		try
		{
			dc = sc.newDbControl();
			Experiment e = super.getCurrentExperiment(dc);
			if (e != null)
			{
				ItemQuery<AnnotationType> query = e.getExperimentalFactors();
				query.include(Include.MINE, Include.SHARED, Include.IN_PROJECT, Include.OTHERS);
				query.order(Orders.asc(Hql.property("name")));
				list.addAll(query.list(dc));
			}
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return list;
	}
	
	
	/**
		Get the data from the configuration file.
		
		@param f The file.
		@return A map of name->value pairs.
		@throws IOException If there i any problem reading from file.
	*/
	private Map<String, String> getParametersFromBaseFile(File f)
		throws IOException
	{
		FlatFileParser ffp = getInitializedFlatFileParser(f.getDownloadStream(0), f.getCharacterSet());
		ffp.setMinDataColumns(8);

		FlatFileParser.Line s = ffp.nextSection();
		ffp.parseHeaders();
		if (!s.name().equals("plugin"))
		{
			throw new BaseException("Can't find section 'plugin' in configuration file: " + f.getName());
		}
		
		HashMap<String, String> headers = new HashMap<String, String>();
		for (String header : ffp.getHeaderNames())
		{
			headers.put(header, ffp.getHeader(header));
		}
		
		Element parameters = new Element("jobparameters");
		while (ffp.hasMoreData())
		{
			FlatFileParser.Data dataline = ffp.nextData();
			try
			{
				JobParameter jp = new JobParameter(dataline.line().split("\t"));
				parameters.addContent(jp.toXml());
			}
			catch(Exception e)
			{
				throw new BaseException("Error on line "+dataline.lineNo()+" in the configuration file: " + f.getName(), e);
			}
		}
		XMLOutputter out = new XMLOutputter(Format.getPrettyFormat());
		headers.put(jobParametersParameter.getName(), out.outputString(parameters));
		return headers;
	}


	@SuppressWarnings({ "unchecked" })
	private Map<String, JobParameter> getJobParameterObjectsFromXML(String xml, boolean hideRemoved)
	{
		try
		{
			String doctype = "<!DOCTYPE jobparameters SYSTEM \"base1-plugin-configuration-file.dtd\">";
			Document doc = XMLUtil.getValidatedXml(doctype+xml, dtdFile);
			Iterator<Element> it = doc.getDescendants(new ElementFilter("parameter"));
			Map<String, JobParameter> pars = new TreeMap<String, JobParameter>();
			while (it.hasNext())
			{
				Element e = it.next();
				JobParameter jp = new JobParameter(e);
				if (!jp.isRemoved())
				{
					if (pars.containsKey(jp.getName()))
					{
						throw new BaseException("Duplicate parameter '" + jp.getName() + 
								"' at position "+jp.getPosition());
					}
					pars.put(jp.getName(), jp);
				}
			}
			return pars;
		}
		catch (JDOMException e1)
		{
			throw new BaseException(e1.getCause());
		}
		catch (IOException e1)
		{
			throw new BaseException(e1);
		}
	}


	@SuppressWarnings({ "unchecked" })
	private List<PluginParameter<?>> getJobParametersFromXML(String xml, boolean hideRemoved)
	{
		try
		{
			String doctype = "<!DOCTYPE jobparameters SYSTEM \"base1-plugin-configuration-file.dtd\">";
			Document doc = XMLUtil.getValidatedXml(doctype+xml, dtdFile);
			Iterator<Element> it = doc.getDescendants(new ElementFilter("parameter"));
			Map<Integer, PluginParameter<?>> positionParameter = 
				new TreeMap<Integer, PluginParameter<?>>();
			while (it.hasNext())
			{
				Element e = it.next();
				JobParameter jp = new JobParameter(e);
				if (!jp.isRemoved())
				{
					if (positionParameter.containsKey(jp.getPosition()))
					{
						throw new BaseException("Duplicate parameter '" + jp.getName() + 
								"' at position "+jp.getPosition());
					}
					positionParameter.put(jp.getPosition(), jp.getPluginParameter());
				}
			}
			return new ArrayList<PluginParameter<?>>(positionParameter.values());
		}
		catch (JDOMException e1)
		{
			throw new BaseException(e1.getCause());
		}
		catch (IOException e1)
		{
			throw new BaseException(e1);
		}
	}


	/**
		Creates a boolean from a string.
		
		@param s The string.
		@return false if s.equals("0") else true.
	*/
	private Boolean getBoolean(String s)
	{
		return s.equals("0") ? false : true;
	}
	
	/**
		Creates the information needed for manual configuration and initiates the 
		parameters. The parameters added are: versionNumber, usedColumns, usedFields, execName,
		geneAverages, SerialFormat, leaveStdin, leaveStdout, minChannel and maxCannel.
		
		@return The RequestInformation.
	*/
	private RequestInformation getManualConfigureParameters()
	{
		if (configureManual == null)
		{
			List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();

			versionNumberParameter = new PluginParameter<String>(
				"versionNumber",
				"Version",
				"The version of this plug-in",
				new StringParameterType(255, null, false));
			parameters.add(versionNumberParameter);
			
			usedColumnsParameter = new PluginParameter<String>(
				"usedColumns",
				"Columns",
				"A tab separeted sting of field namnes. A field is a value from the bioassay, the rawdata or feature",
				new StringParameterType(null, null, true));
			parameters.add(usedColumnsParameter);
			
			usedFieldsParameter = new PluginParameter<String>(
				"usedFields",
				"Fields",
				"A tab separeted sting of field namnes. A field is a value from the bioassay or the rawdata",
				new StringParameterType(null, null, true));
			parameters.add(usedFieldsParameter);
			
			execNameParameter = new PluginParameter<String>(
				"execName",
				"Name of executable",
				"The name of the file to be executed.",
				new StringParameterType(255, null, true));
			parameters.add(execNameParameter);
			
			geneAveragesParameter = new PluginParameter<Boolean>(
				"geneAverages",
				"Average on reporters",
				"Average on reporters within each assay",
				new BooleanParameterType(false, true));
			parameters.add(geneAveragesParameter);

			Enumeration<Boolean, String> format = new Enumeration<Boolean, String>();
			format.add(false, "Normal BASEfile (one big matrix)");
			format.add(true, "Serial BASEfile (one section per assay)");
			serialFormatParameter = new PluginParameter<Boolean>(
				"serialFormat",
				"Data format",
				"The format of the datafile can either be in a serial or matrix format",
				new BooleanParameterType(null, true, 1, format));
			parameters.add(serialFormatParameter);
			
			leaveStdinParameter = new PluginParameter<Boolean>(
				"leaveStdin",
				"Leave stdin.txt",
				"If the file stdin.txt should be removed after the plug-in execution or not",
				new BooleanParameterType(null, false));
			parameters.add(leaveStdinParameter);
			
			leaveStdoutParameter = new PluginParameter<Boolean>(
				"leaveStdout",
				"Leave stdout.txt",
				"If the file stdout.txt should be removed after the plug-in execution or not",
				new BooleanParameterType(null, false));
			parameters.add(leaveStdoutParameter);
			
			PluginParameter<Boolean> overrideLeaveStdParameter =  new PluginParameter<Boolean>(
				"overrideLeaveStd",
				"Allow override",
				"Select this option if the user should be allowed to override the values " +
				"for 'Leave stdout/stdin' parameters when starting a job.",
				new BooleanParameterType(null, false));
			parameters.add(overrideLeaveStdParameter);
			
			minChannelsParameter = new PluginParameter<Integer>(
				"minChannels",
				"Min channels",
				"Minimum number of channels",
				new IntegerParameterType(1, null, 2, true));
			parameters.add(minChannelsParameter);
			
			maxChannelsParameter = new PluginParameter<Integer>(
				"maxChannels",
				"Max channels",
				"Maximum number of channels. 0 for infinity",
				new IntegerParameterType(0, null, 0, true));
			parameters.add(maxChannelsParameter);
			
			parameters.add(jobParametersParameter);
			
			configureManual = new RequestInformation(
				Base1PluginExecuter.COMMAND_MANUAL_CONFIG,
				"Parse configuration file",
				"These parameters are parsed from the configuration file",
				parameters);
		}

		return configureManual;
	}

	
	/**
		Creates the information needed for job configuration and initiates the 
		parameters. The parameters added are: bioAssaySet and the parameters 
		found in the configuration file.
		
		@return The RequestInformation
	*/
	private RequestInformation getConfigureJobParameters()
	{
		if (configureJob == null)
		{
			// Parameters for CONFIGURE_JOB
			List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();
			parameters.add(getSourceBioAssaySetParameter("Bioassay set", "The bioassay set that this plugin will run on"));
			
			DbControl dc = null;
			try
			{
				dc = sc.newDbControl();
				Experiment experiment = getCurrentExperiment(dc);
				BioAssaySet bas = getCurrentBioAssaySet(dc);
				String homeDirectoryPath = "/" + Path.makeSafeFilename(PluginConfiguration.getById(dc, configuration.getId()).getName(), "");
				Directory homeDirectory = experiment == null ? null : experiment.getDirectory();
				if (homeDirectory != null)
				{
					homeDirectoryPath = homeDirectory.getPath() + homeDirectoryPath;
				}
				else
				{
					homeDirectory = User.getById(dc, sc.getLoggedInUserId()).getHomeDirectory();
					if (homeDirectory != null)
					{
						homeDirectoryPath = homeDirectory.getPath() + homeDirectoryPath;
					}
				}
				pluginDirectoryParameter = new PluginParameter<String>(
					"pluginDirectory",
					"Output directory",
					"The directory where a plugin can store its files.",
					new PathParameterType(Path.Type.DIRECTORY, homeDirectoryPath, true));
				parameters.add(pluginDirectoryParameter);
				
				parameters.add(getChildNameParameter(null, null, bas == null ? null : bas.getName()));
				parameters.add(getChildDescriptionParameter(null, null, null));
				
				boolean overrideLeaveStd = Boolean.TRUE.equals(configuration.getValue("overrideLeaveStd"));
				if (overrideLeaveStd)
				{
					Boolean leaveStdin = (Boolean)configuration.getValue("leaveStdin");
					leaveStdinParameter = new PluginParameter<Boolean>(
							"leaveStdin",
							"Leave stdin.txt",
							"If the file stdin.txt should be removed after the plug-in execution or not. " +
							"If not specified the file is " + (leaveStdin ? "kept." : "removed."),
							new BooleanParameterType(leaveStdin, false));
					parameters.add(leaveStdinParameter);
					Boolean leaveStdout = (Boolean)configuration.getValue("leaveStdout");
					leaveStdoutParameter = new PluginParameter<Boolean>(
							"leaveStdout",
							"Leave stdout.txt",
							"If the file stdout.txt should be removed after the plug-in execution or not. " +
							"If not specified the file is " + (leaveStdout ? "kept." : "removed."),
							new BooleanParameterType(leaveStdout, false));
					parameters.add(leaveStdoutParameter);
				}

				parameters.add(new PluginParameter<String>("pluginsettings", "Plug-in parameters", null, null));
			}
			finally
			{
				if (dc != null) dc.close();
			}
			
			parameters.addAll(getJobParameters());
			configureJob = new RequestInformation(
				Request.COMMAND_CONFIGURE_JOB,
				"Configure job",
				null,
				parameters);
		}
		return configureJob;
	}
	
	
	/**
 		Creates the information needed for plugin configuration and initiates the 
		parameters. The parameters added are: file and execPath.
		
		@return The RequestInformation.
	*/
	private RequestInformation getConfigurePluginParameters()
	{
		if (configurePlugin == null)
		{
			// Parameters object for CONFIGURE_PLUGIN
			List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>();
			parameters.add(fileParameter);
			parameters.add(execPathParameter);
			
			Enumeration<String, String> transforms = new Enumeration<String, String>();
			for (IntensityTransform t : IntensityTransform.values())
			{
				transforms.add(t.name(), t.toString());
			}
			parameters.add(
				new PluginParameter<String>(
					"sourceTransform",
					"Source intensities",
					"The type of spot intensity values that the source bioassay set must have " +
					"for this plug-in to work correctly. If not specified the plug-in can work " +
					"with any kind of data.",
					new StringParameterType(255, null, false, 1, 0, 0, transforms)
				)
			);
			parameters.add(
				new PluginParameter<String>(
					"resultTransform",
					"Resulting intensities",
					"The type of spot intensity values this plug-in generates. If not " +
					"specified it is expected to generate the same type of values as are found " +
					"in the source bioassay set",
					new StringParameterType(255, null, false, 1, 0, 0, transforms)
				)
			);
			
			configurePlugin = new RequestInformation(
				Request.COMMAND_CONFIGURE_PLUGIN,
				"Configure plugin",
				"The configuration file from base1",
				parameters);
		}
		return configurePlugin;
	}

	
	/**
		Get the exec directory as a java.io.File. The exec directory is a sub directory in 
		"java.io.tmpdir".
		
		@return the directory.
	*/
	private java.io.File getExecDirectory()
		throws IOException
	{
		if (execDirectory == null)
		{
			String now = new DateFormatter("-yyyyMMdd-HHmmss-").format(new Date());
			execDirectory = FileUtil.createTempDirectory("base1executor"+now, null, null);
		}
		return execDirectory;
	}
	
	
	private String getExecLine()
	{
		String execPath = (String) configuration.getValue("execPath");
		String execName = (String) configuration.getValue("execName");
		String s = java.io.File.separator;
		if (execPath.endsWith(s) || execName.startsWith(s))
		{
			s = "";
		}
		String line = execPath + s + execName;
		return line;
	}

	
	/**
		Creates a {@link FlatFileParser} for a plugin configuration file.
		
		@param stream The stream that the parser should read from.
		@return The initialised parser
	*/
	private FlatFileParser getInitializedFlatFileParser(InputStream stream, String charset)
	{
		FlatFileParser ffp = new FlatFileParser();
		ffp.setSectionRegexp(Pattern.compile("section[\\t ](.*)"));
		ffp.setHeaderRegexp(Pattern.compile("(.*?)\\t(.*)"));
		ffp.setDataHeaderRegexp(Pattern.compile("%"));
		ffp.setDataFooterRegexp(Pattern.compile("^$"));
		ffp.setDataSplitterRegexp(Pattern.compile("\\t"));
		ffp.setUseNullIfEmpty(false);

		ffp.setInputStream(stream, charset != null ? charset : Config.getCharset());
		return ffp;
	}
	

	/**
		Reads the configuration file an creates the parameters found there.
		
		@return A list of {@link PluginParameter}:s
	*/
	private List<PluginParameter<?>> getJobParameters()
		throws BaseException
	{
		if (jobParameters != null)
		{
			return jobParameters;
		}

		Object xml = configuration.getValue(jobParametersParameter.getName());
		// To make the plugin compatible with older configurations
		if (xml == null)
		{
			throw new BaseException("Cant find the jobparameters specification. Reconfigure the plugin.");
		}
		jobParameters = getJobParametersFromXML(xml.toString(), true);
		return jobParameters;
	}
	
	/**
		Get a {@link Directory directory} 
	*/
	private Directory getPluginDirectory(DbControl dc, String pathName)
	{ 
		Path path = new Path(pathName, Path.Type.DIRECTORY);
		Directory jobDirectory = null;
		try
		{
			// Load existing...
			jobDirectory = Directory.getByPath(dc, path);
		}
		catch(ItemNotFoundException e)
		{
			// Create a new directory if no existing were found
			jobDirectory = Directory.getNew(dc, path);
			dc.saveItem(jobDirectory);
		}
		return jobDirectory;
	}
	
	
	/**
		Imports the data from the file stdout.txt in the execution directory.
		@throws IOException if there is any error reading from stdout.txt
	*/
	private BioAssaySet importData(DbControl dc, File stdout, Transformation t, 
			IntensityTransform resultTransform, ProgressReporter progress)
		throws SQLException, IOException
	{
		BaseFileImporter importer = new BaseFileImporter();
		importer.setDbControl(dc);
		importer.setSourceFileWrapper(new BaseFileWrapper(stdout));
		importer.setTransformation(t);
		importer.setIntensityTransform(resultTransform);
		importer.setProgressReporter(progress);
		return importer.doImport();
	}
	
	// Associate files in a directory with transformation
	private void associateFiles(DbControl dc, Transformation trans, Directory homeDirectory)
	{
		ThreadSignalHandler.checkInterrupted();
		String root = homeDirectory.getPath().toString();
		AnyToAny ata = AnyToAny.getNew(dc, trans, homeDirectory, "./", false);
		dc.saveItem(ata);
		for(File f : homeDirectory.getFiles().list(dc))
		{
			ata = AnyToAny.getNew(dc, trans, f, f.getPath().toString().replace(root, "."), false);
			dc.saveItem(ata);
		}
		for(Directory d : homeDirectory.getSubDirectories().list(dc))
		{
			ata = AnyToAny.getNew(dc, trans, d, d.getPath().toString().replace(root, "."), false);
			dc.saveItem(ata);
		}
	}

	// Import external files to a directory in BASE
	private void importTempFiles(DbControl dc, java.io.File[] files, Directory d)
	{
		for (java.io.File f : files)
		{
			if (f.isFile())
			{
				if (!File.exists(dc, d, f.getName()))
				{
					File newFile = File.getNew(dc, d);
					newFile.setName(f.getName());
					newFile.setMimeTypeAuto(null, null);
					dc.saveItem(newFile);
					try
					{
						ThreadSignalHandler.checkInterrupted();
						newFile.upload(new FileInputStream(f), true);
					}
					catch (FileNotFoundException e)
					{
						continue;
					}
				}
			}
			else if (f.isDirectory())
			{
				Directory newDirectory = Directory.getNew(dc, d);
				newDirectory.setName(f.getName());
				dc.saveItem(newDirectory);
				importTempFiles(dc, f.listFiles(), newDirectory);
			}
		}
	}
	
	
	private class JobParameter
	{
		private int position;
		
		private Base1JobParameterType type;
		
		private String name;
		
		private String commonName;
		
		private String options;
		
		private String defaultValue;
		
		private Enumeration<String, String> enumOptions = new Enumeration<String, String>();
		
		private boolean removed = false;

		JobParameter(String... base1ConfigLine)
		{
			if (base1ConfigLine.length != 8)
			{
				throw new BaseException("Must have 8 parameters.");
			}
			if (base1ConfigLine[0].equals(""))
			{
				throw new BaseException("Position cannot be null");
			}
			position = Integer.parseInt(base1ConfigLine[0]);
			if (base1ConfigLine[1].length() != 1)
			{
				throw new BaseException("Type must one character");
			}
			type = Base1JobParameterType.get(base1ConfigLine[1].charAt(0));
			name = base1ConfigLine[2];
			commonName = base1ConfigLine[3];
			options = base1ConfigLine[4];
			defaultValue = base1ConfigLine[5];
			
			if (type == Base1JobParameterType.ENUMERATION)
			{
				String[] enums = base1ConfigLine[6].split("\\\\t");
				if (enums.length % 2 != 0) throw new BaseException("name/value pairs in enum isnt even: "+base1ConfigLine[6]);
				Set<String> values = new HashSet<String>();
				for (int i = 0; i < enums.length; i += 2)
				{
					enumOptions.add(enums[i], enums[i+1]);
					if (!values.add(enums[i+1]))
					{
						throw new BaseException("Enumeration value for parameter '" + name + 
								"' is not unique: " + enums[i+1]);
					}
				}
			}
			removed = Values.getBoolean(base1ConfigLine[7]);
		}

		@SuppressWarnings({ "unchecked" })
		JobParameter(Element e)
		{
			try
			{
				position = e.getAttribute("position").getIntValue();
				type = Base1JobParameterType.get(e.getAttribute("type").getValue().charAt(0));
				Iterator<Element> it = e.getDescendants(new ElementFilter());
				while (it.hasNext())
				{
					Element child = it.next();
					if (child.getName().equals("name"))
					{
						name = child.getValue();
					}
					else if (child.getName().equals("commonname"))
					{
						commonName = child.getValue();
					}
					else if (child.getName().equals("options"))
					{
						options = child.getValue();
					}
					else if (child.getName().equals("defaultvalue"))
					{
						defaultValue = child.getValue();
					}
					else if (child.getName().equals("removed"))
					{
						removed = true;
					}
					else if (child.getName().equals("enumoptions"))
					{
						Set<String> values = new HashSet<String>();
						Iterator<Element> it2 = child.getDescendants(new ElementFilter());
						while (it2.hasNext())
						{
							Element value = it2.next();
							enumOptions.add(value.getAttributeValue("key"), value.getValue());
							if (!values.add(value.getValue()))
							{
								throw new BaseException("Enumeration value for parameter '" + name + 
										"' is not unique: " + value.getValue());
							}
						}
					}
				}
			}
			catch (JDOMException ex)
			{
				throw new BaseException(ex);
			}
		}

		public PluginParameter<?> getPluginParameter()
		{
			PluginParameter<?> parameter;
			switch (type)
			{
				case STRING:
				{
					StringParameterType t = new StringParameterType(65535, defaultValue, false, 1, new Integer(options), 1);
					parameter = new PluginParameter<String>(name, commonName, "", t);
					break;
				}
				case INTEGER:
				{
					Integer defValue = Values.getInteger(defaultValue, null);
					IntegerParameterType t = new IntegerParameterType(
						null,
						null,
						defValue,
						false,
						1,
						Values.getInt(options, 30),
						1,
						null);
					parameter = new PluginParameter<Integer>(name, commonName, "", t);
					break;
				}
				case HIDDEN_STRING:
				{
					StringParameterType t = new StringParameterType(65535, defaultValue, false, 1, 0, 1);
					parameter = new PluginParameter<String>(name, null, "", t);
					break;
				}
				case FLOAT:
				{
					Float defValue = Values.getFloat(defaultValue, null);
					FloatParameterType t = new FloatParameterType(null, null, defValue, false, 1, Values.getInt(options, 30), 1, null);
					parameter = new PluginParameter<Float>(name, commonName, "", t);
					break;
				}
				case TEXT:
				{
					String[] wh = options.split(",");
					int width = Values.getInt(wh[0], 30);
					int height = wh.length > 1 ? Values.getInt(wh[1], 1) : 1;
					StringParameterType t = new StringParameterType(null, defaultValue, false, 1, width, height);
					parameter = new PluginParameter<String>(name, commonName, "", t);
					break;
				}
				case ANNOTATION:
				{
					List<AnnotationType> annotations = getAnnotationTypes();
					Enumeration<String, String> enums = new Enumeration<String, String>();
					for (AnnotationType at : annotations)
					{
						enums.add(at.getName(), at.getName());
					}
					StringParameterType t = new StringParameterType(65535, null, false, 1, Values.getInt(options, 30), 1, enums);
					parameter = new PluginParameter<String>(name, commonName, "", t);
					break;
				}
				case ENUMERATION:
				{
					StringParameterType t = new StringParameterType(65535, defaultValue, false, 1, Values.getInt(options, 30), 1, enumOptions);
					parameter = new PluginParameter<String>(name, commonName, "", t);
					break;
				}
				default:
					parameter = null;
			}
			return parameter;
		}
		
		public final String getCommonName()
		{
			return commonName;
		}

		public final String getDefaultValue()
		{
			return defaultValue;
		}

		public final Enumeration<String, String> getEnumOptions()
		{
			return enumOptions;
		}

		public final String getName()
		{
			return name;
		}

		public final String getOptions()
		{
			return options;
		}

		public final int getPosition()
		{
			return position;
		}

		public final boolean isRemoved()
		{
			return removed;
		}

		public final Base1JobParameterType getType()
		{
			return type;
		}

		public Element toXml()
		{
			Element parameter = new Element("parameter");
			parameter.setAttribute("position", String.valueOf(getPosition()));
			parameter.setAttribute("type", getType().toString());
			
			parameter.addContent(new Element("name").setText(getName()));
			parameter.addContent(new Element("commonname").setText(getCommonName()));
			parameter.addContent(new Element("options").setText(getOptions()));
			parameter.addContent(new Element("defaultvalue").setText(getDefaultValue()));
			
			if (enumOptions.size() > 0)
			{
				Element enums = new Element("enumoptions");
				for (int i = 0; i < enumOptions.size(); ++i)
				{
					Element value = new Element("value");
					value.setAttribute("key", enumOptions.getKey(i));
					value.setText(enumOptions.getValue(i));
					enums.addContent(value);
				}
				parameter.addContent(enums);
			}
			if (removed) parameter.addContent(new Element("removed"));
			return parameter;
		}

		@Override
		public String toString()
		{
			XMLOutputter out = new XMLOutputter(Format.getPrettyFormat());
			return out.outputString(toXml());
		}
	}

	private enum Base1JobParameterType
	{
		HIDDEN_STRING('h'), STRING('t'), INTEGER('i'), FLOAT('f'), TEXT('a'), ANNOTATION('n'), ENUMERATION('e');
		
		private final char c;
		
		private Base1JobParameterType(char c)
		{
			this.c = c;
		}
		
		public static Base1JobParameterType get(char c)
		{
			switch (c)
			{
				case 'h' : return HIDDEN_STRING;
				case 't' : return STRING;
				case 'i' : return INTEGER;
				case 'f' : return FLOAT;
				case 'a' : return TEXT;
				case 'n' : return ANNOTATION;
				case 'e' : return ENUMERATION;
				default:
					throw new BaseException("Unknown type: "+c);
			}
		}

		@Override
		public String toString()
		{
			return String.valueOf(c);
		}
	}
}
