/**
	$Id: ArrayDesignImporter.java 5371 2010-06-24 08:46:52Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.plugins.batchimport;

import java.util.Collections;
import java.util.List;
import java.util.Set;

import net.sf.basedb.core.ArrayDesign;
import net.sf.basedb.core.DataFileType;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.File;
import net.sf.basedb.core.Hybridization;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.Path;
import net.sf.basedb.core.PathParameterType;
import net.sf.basedb.core.Platform;
import net.sf.basedb.core.PlatformVariant;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.Version;
import net.sf.basedb.core.plugin.About;
import net.sf.basedb.core.plugin.AboutImpl;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.util.parser.FlatFileParser;
import net.sf.basedb.util.parser.Mapper;

/**
	Plug-in for importing array design items in a batch. The plug-in can create new 
	items and updated existing items.

	@author nicklas
	@version 2.8
	@base.modified $Date: 2010-06-24 10:46:52 +0200 (Thu, 24 Jun 2010) $
*/
public class ArrayDesignImporter
	extends AbstractItemImporter<ArrayDesign>
{

	private static final Set<GuiContext> guiContexts =  
		Collections.singleton(new GuiContext(Item.ARRAYDESIGN, GuiContext.Type.LIST));

	private static final About about =
		new AboutImpl
	(
			"Array design importer",
			"Imports and updates array designs in a batch.",
			Version.getMajor() + "." + Version.getMinor() + "." + Version.getMaintenance(),
			"2008, Base 2 development team",
			null,
			null,
			"http://base.thep.lu.se"
	);

	protected static final PluginParameter<String> numArraysColumnMapping = 
		new PluginParameter<String>(
				"numArraysColumnMapping",
				"Number of arrays",
				"Mapping that picks the number of arrays that fits on a single slide. The " +
				"values must be in the numeric format specified by the 'Number format' parameter." +
				"Example: \\Num arrays\\",
				optionalColumnMapping
		);
	
	protected static final PluginParameter<String> platformColumnMapping = new PluginParameter<String>(
			"platformColumnMapping",
			"Platform",
			"Mapping that picks the name or ID of the experimental platform from the data columns. " +
			"The plug-in will first try to find a platform with the given name. If none is found and " +
			"the value is numeric it will try to load by internal ID. " +
			"Example: \\Platform\\",
			optionalColumnMapping
			);
	
	protected static final PluginParameter<String> variantColumnMapping = new PluginParameter<String>(
			"variantColumnMapping",
			"Variant",
			"Mapping that picks the name or ID of the experimental platform variant from the data columns. " +
			"The plug-in will first try to find a variant with the given name. If none is found and " +
			"the value is numeric it will try to load by internal ID. " +
			"Example: \\Variant\\",
			optionalColumnMapping
			);

	protected static final PluginParameter<String> fileTypeColumnMapping = 
		new PluginParameter<String>(
				"fileTypeColumnMapping",
				"Data file type",
				"Mapping that picks the name or ID of a data file type from the data columns. " +
				"The plug-in will first try to find a file type with the given name. If none is " +
				"found and the value is numeric it will try to load by internal ID. " +
				"Example: \\File type\\",
				optionalColumnMapping
		);
	
	protected static final PluginParameter<String> fileColumnMapping = 
		new PluginParameter<String>(
				"fileColumnMapping",
				"File",
				"Mapping that picks the path of a data file from the columns. " +
				"The path can be an absolute path (starting with /) or relative " +
				"to the 'Data directory' parameter." +
				"Example: \\File\\",
				optionalColumnMapping
		);

	
	public ArrayDesignImporter()
	{}

	/*
		From the Plugin interface
		------------------------------------
	*/
	@Override
	public About getAbout()
	{
		return about;
	}
	// ------------------------------------

	/*
		From the InteractivePlugin interface
		------------------------------------
	*/
	@Override
	public Set<GuiContext> getGuiContexts()
	{
		return guiContexts;
	}
	// ------------------------------------

	/*
		From the AbstractItemImporter class
		------------------------------------
	*/
	
	private Mapper nameMapper;
	private Mapper descriptionMapper;
	private Mapper numArraysMapper;
	private Mapper platformMapper;
	private Mapper variantMapper;
	private Mapper fileTypeMapper;
	private Mapper fileMapper;
	private String dataDirectory;
	
	/**
		Useable methods are:
		<ul>
		<li>{@link PropertyIdMethod#NAME}
		<li>{@link InternalIdMethod#INTERNAL_ID}
		<li>{@link FallbackIdMethod#NAME_OR_ID}
		</ul>
	*/
	@Override
	protected IdMethod[] getIdMethods()
	{
		return new IdMethod[] { PropertyIdMethod.NAME,  
				InternalIdMethod.INTERNAL_ID, FallbackIdMethod.NAME_OR_ID};
	}
	
	/**
		Add parameter for "Data directory" directory where data files may be located. 
	*/
	@Override
	protected void addMoreJobParameters(List<PluginParameter<?>> parameters)
	{
		parameters.add(new PluginParameter<String>(
				"dataDirectory",
				"Data directory",
				"A base directory were data files are located. Relative file paths are " +
				"resolved against this directory. If no directory is given, only absolute " +
				"paths (=paths starting with /) can be used.",
				new PathParameterType(Path.Type.DIRECTORY))
			);
	}

	
	/**
		Adds column mappings for name, externalId, description,
		protocol, creation date, original quantity, pooled, parent
		and used quantity from parent.
	*/
	@Override
	protected void addMoreColumnMappingParameters(List<PluginParameter<?>> parameters)
	{
		parameters.add(internalIdColumnMapping);
		parameters.add(nameColumnMapping);
		parameters.add(descriptionColumnMapping);
		parameters.add(numArraysColumnMapping);
		parameters.add(platformColumnMapping);
		parameters.add(variantColumnMapping);
		parameters.add(fileTypeColumnMapping);
		parameters.add(fileColumnMapping);
		parameters.add(permissionTemplateColumnMapping);
	}
	
	/**
		Calls {@link Hybridization#getQuery()}.
	*/
	@Override
	protected ItemQuery<ArrayDesign> createItemQuery()
	{
		return ArrayDesign.getQuery();
	}
	
	@Override
	protected void createColumnMappers(FlatFileParser ffp, boolean cropStrings)
	{
		nameMapper = getMapper(ffp, (String)job.getValue("nameColumnMapping"), 
				cropStrings ? ArrayDesign.MAX_NAME_LENGTH : null, null);
		descriptionMapper = getMapper(ffp, (String)job.getValue("descriptionColumnMapping"), 
				cropStrings ? ArrayDesign.MAX_DESCRIPTION_LENGTH : null, null);
		numArraysMapper = getMapper(ffp, (String)job.getValue("numArraysColumnMapping"), null, null);
		platformMapper = getMapper(ffp, (String)job.getValue("platformColumnMapping"), null, null);
		variantMapper = getMapper(ffp, (String)job.getValue("variantColumnMapping"), null, null);
		fileTypeMapper = getMapper(ffp, (String)job.getValue("fileTypeColumnMapping"), null, null);
		fileMapper = getMapper(ffp, (String)job.getValue("fileColumnMapping"), null, null);
		dataDirectory = (String)job.getValue("dataDirectory");
		if (dataDirectory != null && !dataDirectory.endsWith("/"))
		{
			dataDirectory += "/";
		}
	}
	
	@Override
	protected ArrayDesign createItem(DbControl dc, FlatFileParser.Data data)
	{
		Platform platform = null;
		PlatformVariant variant = null;
		
		if (platformMapper != null) 
		{
			platform = findPlatform(dc, FallbackIdMethod.NAME_OR_EXTERNALID_OR_ID, platformMapper.getValue(data));
		}
		else 
		{
			platform = getProjectDefaultPlatform(dc);
		}
		if (variantMapper != null) 
		{
			variant = findVariant(dc, platform, FallbackIdMethod.NAME_OR_EXTERNALID_OR_ID, variantMapper.getValue(data));
		}
		else
		{
			variant = getProjectDefaultVariant(dc);
		}
		
		ArrayDesign design = null;
		if (variant != null)
		{
			design = ArrayDesign.getNew(dc, variant);
		}
		else
		{
			if (platform == null) platform = Platform.getByExternalId(dc, Platform.GENERIC);
			design = ArrayDesign.getNew(dc, platform);
		}
		updateItem(dc, design, data);
		return design;
	}
	
	@Override
	protected void updateItem(DbControl dc, ArrayDesign design, FlatFileParser.Data data)
	{
		if (nameMapper != null) design.setName(nameMapper.getValue(data));
		if (descriptionMapper != null) design.setDescription(descriptionMapper.getValue(data));
		if (numArraysMapper != null)
		{
			Integer numArrays = numArraysMapper.getInt(data);
			if (numArrays != null) design.setNumArrays(numArrays);
		}
		if (design.isInDatabase()) 
		{
			Platform platform = null;
			PlatformVariant variant = null;
			if (platformMapper != null) 
			{
				String nameOrId = platformMapper.getValue(data);
				platform = findPlatform(dc, FallbackIdMethod.NAME_OR_EXTERNALID_OR_ID, nameOrId);
			}
			if (variantMapper != null) 
			{
				String nameOrId = variantMapper.getValue(data);
				variant = findVariant(dc, platform, FallbackIdMethod.NAME_OR_EXTERNALID_OR_ID, nameOrId);
				if (nameOrId == null || variant != null) design.setVariant(variant);
			}
			if (variant != null)
			{
				design.setVariant(variant);
			}
			else if (platform != null)
			{
				design.setPlatform(platform);
			}
		}
		updateMultiLineItem(dc, design, data, 0);
	}
	/**
		Add data files
	*/
	@Override
	protected void updateMultiLineItem(DbControl dc, ArrayDesign design, 
		FlatFileParser.Data data, int multiLineNum)
	{
		if (fileTypeMapper != null && fileMapper != null)
		{
			String nameOrId = fileTypeMapper.getValue(data);
			String filePath = fileMapper.getValue(data);
			DataFileType fileType = findDataFileType(dc, FallbackIdMethod.NAME_OR_EXTERNALID_OR_ID, nameOrId);
			if (fileType != null)
			{
				File file = null;
				if (filePath != null)
				{
					if (!filePath.startsWith("/") && dataDirectory != null) 
					{
						filePath = dataDirectory + filePath;
					}
					file = File.getByPath(dc, new Path(filePath, Path.Type.FILE), false);
				}
				if (file != null)
				{
					design.getFileSet().setMember(file, fileType);
				}
				else
				{
					design.getFileSet().removeMember(fileType);
				}
			}
		}
	}
	/**
		Validate data files if they have been set.
	*/
	@Override
	protected void doneWithItem(DbControl dc, ArrayDesign design, int numMultiLines)
	{
		if (design.hasFileSet()) design.getFileSet().validate(dc, true);
	}
	// ------------------------------------
}
