/**
	$Id: PropertyIdMethod.java 5384 2010-08-13 08:50:27Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.plugins.batchimport;

import java.util.List;

import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;

/**
	Identification method implementation that can use a single property
	(eg, name, id, externalId, etc.) for identifying items.

	@author nicklas
	@version 2.8
	@base.modified $Date: 2010-08-13 10:50:27 +0200 (Fri, 13 Aug 2010) $
*/
public class PropertyIdMethod
	implements IdMethod
{

	/**
		Use the name for finding items.
	*/
	public static final PropertyIdMethod NAME = 
		new PropertyIdMethod("name", "Name", "name", 
				AbstractItemImporter.nameColumnMapping.getName(), false, Type.STRING);
	
	/**
		Use the external ID for finding items.
	*/
	public static final PropertyIdMethod EXTERNAL_ID = 
		new PropertyIdMethod("externalId", "External ID", "externalId", 
				AbstractItemImporter.externalIdColumnMapping.getName(), false, Type.STRING);

	/**
		Use the barcode for finding items.
	*/
	public static final PropertyIdMethod BARCODE = 
		new PropertyIdMethod("barcode", "Barcode", "barcode", 
			ArraySlideImporter.barcodeColumnMapping.getName(), false, Type.STRING);

	/**
		@deprecated In 2.16, due to possible problems with circular class initialization
			with the {@link FallbackIdMethod} class. This array is now empty.
	*/
	@Deprecated
	public static final IdMethod[] DEFAULT_METHODS = {};

	private final String method;
	private final String property;
	private final String columnMappingParameterName;
	private final String title;
	private final boolean unique;
	private final Type valueType;
	
	/**
		Create a new id method.

		@param method The method name, which should be unique among all methods
		@param title The title to use for display purposes
		@param property The name of the property to filter on when using this
			id method
		@param unique If this property is guaranteed to be unique or not
		@param valueType The type of values
	*/
	public PropertyIdMethod(String method, String title, String property, 
			String columnMappingParameterName,
			boolean unique, Type valueType)
	{
		this.method = method;
		this.title = title;
		this.columnMappingParameterName = columnMappingParameterName;
		this.unique = unique;
		this.property = property;
		this.valueType = valueType;
	}
	
	/*
		From the IdMethod interface
		---------------------------
	*/
	@Override
	public String getMethod()
	{
		return method;
	}
	
	@Override
	public String getTitle()
	{
		return title;
	}
	
	@Override
	public String getColumnMappingParameterName()
	{
		return columnMappingParameterName;
	}

	@Override
	public boolean isTrulyUnique()
	{
		return unique;
	}

	/**
		Adds a restriction to the query:
		<code>property = :identifier</code>
	*/
	@Override
	public <I extends BasicItem> ItemQuery<I> prepareQuery(DbControl dc, ItemQuery<I> query)
	{
		query.restrictPermanent(
				Restrictions.eq(
					Hql.property(getProperty()),
					Expressions.parameter("identifier")
				)
			);
		return query;
	}

	/**
		Find all items with <code>property = identifier</code>.
	*/
	@Override
	public <I extends BasicItem> List<I> find(DbControl dc, ItemQuery<I> query, String identifier)
	{
		Type idType = getValueType();
		query.setParameter("identifier", idType.parseString(identifier), idType);
		return query.list(dc);
	}
	// ---------------------------
	
	/*
		From the Object class
		---------------------
	*/
	@Override
	public String toString()
	{
		return title;
	}
	// --------------------------

	public String getProperty()
	{
		return property;
	}
	
	public Type getValueType()
	{
		return valueType;
	}
	
}
