/*
	$Id: TestPermissionTemplate.java 5368 2010-06-22 12:51:41Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
import net.sf.basedb.core.*;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;

import java.util.Date;
import java.util.EnumSet;
import java.util.Set;

public class TestPermissionTemplate
{

	static boolean ok = true;
	public static void main(String[] args)
	{
		TestUtil.checkArgs(args);
		TestUtil.begin();
		ok = test_all();
		TestUtil.stop();
	}

	static boolean test_all()
	{
		write("++Testing permission templates");
		write_header();
		// Standard tests: create, load, list
		
		// Create a user, group and project and keys for sharing
		int userId = TestUser.test_create(false);
		int groupId = TestGroup.test_create(false);
		int projectId = TestProject.test_create("Main project", false);
		int itemKeyId = TestItemKey.test_create(userId, EnumSet.of(Permission.WRITE), groupId, EnumSet.of(Permission.READ));
		int projectKeyId = TestProjectKey.test_create(projectId, EnumSet.of(Permission.DELETE));
		
		int id = test_create("Test template", itemKeyId, projectKeyId);
		test_load(id);
		test_list(-1);

		int projectId2 = TestProject.test_create("Template project", false);
		test_assign_to_project(id, projectId2);
		TestProject.test_activate_project(projectId2, null);
		int sampleId = TestSample.test_create(0, "Sample (shared via template)", false);
		test_verify_share(Item.SAMPLE, sampleId, itemKeyId, projectKeyId);
		TestProject.test_noactive_project();
		
		// Standard test: Delete
		if (TestUtil.waitBeforeDelete()) TestUtil.waitForEnter();
		TestSample.test_delete(sampleId);
		TestProject.test_delete(projectId2);
		test_delete(id);
		
		TestItemKey.test_delete_unused(-1);
		TestProjectKey.test_delete_unused(-1);
		TestProject.test_delete(projectId);
		TestGroup.test_delete(groupId);
		TestUser.test_delete(userId);
		
		write("++Testing permission templates "+(ok ? "OK" : "Failed")+"\n");
		return ok;
	}

	static int test_create(String name, int itemKeyId, int projectKeyId)
	{
		if (!TestUtil.hasPermission(Permission.CREATE, Item.LABEL)) return 0;
		int id = 0;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			PermissionTemplate pt = PermissionTemplate.getNew(dc);
			pt.setName(name);
			if (itemKeyId != 0) pt.setItemKey(ItemKey.getById(dc, itemKeyId));
			if (projectKeyId != 0) pt.setProjectKey(ProjectKey.getById(dc, projectKeyId));
			pt.setDescription("Added at "+new Date());
			dc.saveItem(pt);
			dc.commit();
			id = pt.getId();
			dc = TestUtil.getDbControl();
			dc.reattachItem(pt, false);
			write_item(0, pt);
			write("--Create permission template OK");
		}
		catch (Throwable ex)
		{
			write("--Create permission template FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return id;
	}

	static void test_load(int id)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			PermissionTemplate pt = PermissionTemplate.getById(dc, id);
			write_item(0, pt);
			write("--Load permission template OK");
		}
		catch (Throwable ex)
		{
			write("--Load permission template FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static int test_load(String name)
	{
		if (name == null) return 0;
		DbControl dc = null;
		int id = 0;
		try
		{
			dc = TestUtil.getDbControl();
			ItemQuery<PermissionTemplate> query = PermissionTemplate.getQuery();
			query.restrict(Restrictions.eq(Hql.property("name"), Expressions.string(name)));
			PermissionTemplate pt = query.list(dc).get(0);
			write_item(0, pt);
			write("--Load permission template by name OK");
			id = pt.getId();
		}
		catch (Throwable ex)
		{
			write("--Load permission template by name FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return id;
	}

	
	static void test_list(int expectedResults)
	{
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			ItemResultList<PermissionTemplate> l = PermissionTemplate.getQuery().list(dc);
			for (int i = 0; i<l.size(); i++)
			{
				write_item(i, l.get(i));
			}
			if (expectedResults >= 0 && expectedResults != l.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+l.size());
			}
			write("--List permission templates OK ("+l.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List permission templates FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_delete(int id)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			PermissionTemplate pt = PermissionTemplate.getById(dc, id);
			dc.deleteItem(pt);
			Set<ItemProxy> using = pt.getUsingItems();
			if (using.size() > 0) 
			{
			   throw new BaseException(pt + " is used by " + using);
			}
			dc.commit();
			write("--Delete permission template OK");
		}
		catch (Throwable ex)
		{
			write("--Delete permission template FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void write_header()
	{
		if (!TestUtil.getSilent())
		{
			write("   \tID \tName      \tDescription\tItem key\tProject key");
			write("-- \t-- \t--------- \t-----------\t--------\t-----------");
		}
	}
	static void write_item(int i, PermissionTemplate pt)
		throws BaseException
	{
		if (!TestUtil.getSilent()) System.out.println(i+":\t"+pt.getId()+"\t"+pt.getName()+"\t"+pt.getDescription()+
				"\t"+pt.getItemKey() + "\t" + pt.getProjectKey());
	}
	static void write(String message)
	{
		System.out.println(message);
	}
	
	static void test_assign_to_project(int templateId, int projectId)
	{
		if (templateId == 0 || projectId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			PermissionTemplate pt = PermissionTemplate.getById(dc, templateId);
			Project p = Project.getById(dc, projectId);
			p.setPermissionTemplate(pt);
			dc.commit();
			write("--Assign permission template to project OK");
		}
		catch (Throwable ex)
		{
			write("--Assign permission template to project FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_verify_share(Item itemType, int itemId, int expectedItemKeyId, int expectedProjectKeyId)
	{
		if (itemId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Shareable item = (Shareable)itemType.getById(dc, itemId);
			
			if (expectedItemKeyId >= 0)
			{
				ItemKey itemKey = item.getItemKey();
				if (expectedItemKeyId != (itemKey == null ? 0 : itemKey.getId()))
				{
					throw new BaseException("Item " + item + " is not shared with the expected item key: " + 
						itemKey + "; expected id = " + expectedItemKeyId);
				}
			}
			if (expectedProjectKeyId >= 0)
			{
				ProjectKey projectKey = item.getProjectKey();
				if (expectedProjectKeyId != (projectKey == null ? 0 : projectKey.getId()))
				{
					throw new BaseException("Item " + item + " is not shared with the expected project key: " + 
						projectKey + "; expected id = " + expectedProjectKeyId);
				}
			}
			
			dc.commit();
			write("--Verify item share " + item + " OK");
		}
		catch (Throwable ex)
		{
			write("--Verify item share " + itemType + "[id=" + itemId + "] FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

}
